kAlert = {}

kAlert.context = UI.CreateContext("kAlert")

kAlert.version = 1.12

kAlert.alertSet = {}
kAlert.alertSet.count = 0
kAlert.alertSet.alerts = {}

kAlert.screenObjects = {}
kAlert.screenObjects.object = nil
kAlert.screenObjects.max = 20

kAlert.units = {"player","player.target","focus","player.pet"}
kAlert.unitList = {}
kAlert.combat = false

kAlert.changeHandler = {}
kAlert.changeHandler.buffChanged = false
kAlert.changeHandler.abilityChanged = false
kAlert.changeHandler.resourceChanged = false
kAlert.changeHandler.castingChanged = false

kAlert.systemScanner = {}
kAlert.systemScanner.activeScanning = false
kAlert.systemScanner.nextScan = math.floor(Inspect.Time.Frame())

function kAlert.main()

	table.insert(Event.System.Secure.Enter, {function () kAlert.combat = true end, "kAlert", "enterCombat"})
	table.insert(Event.System.Secure.Leave, {function () kAlert.combat = false end, "kAlert", "exitCombat"})

	table.insert(Event.Unit.Add, {kAlert.changeHandler.checkUnitChanged, "kAlert", "unitChangedAdd"})
	table.insert(Event.Unit.Remove, {kAlert.changeHandler.checkUnitChanged, "kAlert", "unitChangedRemove"})

	table.insert(Event.Unit.Castbar, {kAlert.changeHandler.checkCasting, "kAlert", "castDetected"})

	table.insert(Event.Buff.Add, {kAlert.changeHandler.checkBuffChanged, "kAlert", "buffAdded"})
	table.insert(Event.Buff.Remove, {kAlert.changeHandler.checkBuffChanged, "kAlert", "buffRemoved"})

	table.insert(Event.Ability.Cooldown.Begin, {kAlert.changeHandler.checkAbilityChanged, "kAlert", "cooldownBegin"})
	table.insert(Event.Ability.Cooldown.End, {kAlert.changeHandler.checkAbilityChanged, "kAlert", "cooldownEnd"})
	table.insert(Event.Ability.Usable.False, {kAlert.changeHandler.checkAbilityChanged, "kAlert", "cooldownEnd"})
	table.insert(Event.Ability.Usable.True, {kAlert.changeHandler.checkAbilityChanged, "kAlert", "cooldownEnd"})
	
	table.insert(Event.Unit.Detail.Health, {kAlert.changeHandler.checkResourceChanged, "kAlert", "healthChanged"})
	table.insert(Event.Unit.Detail.Mana, {kAlert.changeHandler.checkResourceChanged, "kAlert", "manaChanged"})
	table.insert(Event.Unit.Detail.Energy, {kAlert.changeHandler.checkResourceChanged, "kAlert", "energyChanged"})
	table.insert(Event.Unit.Detail.Power, {kAlert.changeHandler.checkResourceChanged, "kAlert", "powerChanged"})
	table.insert(Event.Unit.Detail.Charge, {kAlert.changeHandler.checkResourceChanged, "kAlert", "chargeChanged"})
	table.insert(Event.Unit.Detail.Combo, {kAlert.changeHandler.checkResourceChanged, "kAlert", "comboChanged"})

	table.insert(Event.System.Update.Begin, {kAlert.eventHandler, "kAlert", "eventHandler"})

end

function kAlert.alertSet.load()

	kAlert.alertSet.count = 0
	kAlert.alertSet.alerts = {}
	for id, details in pairs (kAlertAlerts) do
		if details.set == kAlertSet and kAlert.alertSet.count < kAlert.screenObjects.max then
			kAlert.alertSet.alerts[details.name] = details
			kAlert.alertSet.count = kAlert.alertSet.count + 1
		end
	end

end

function kAlert.alertSet.save()

	for id, details in pairs (kAlertAlerts) do
		if details.set == kAlertSet then
			kAlertAlerts[id] = nil
		end
	end
	
	for id, details in pairs (kAlert.alertSet.alerts) do
		kAlertAlerts[tostring(kAlertSet) .. details.name] = details
	end
	
end

function kAlert.alertSet.add(alert)

	if kAlert.alertSet.count < kAlert.screenObjects.max then
		kAlert.alertSet.alerts[alert.name] = alert
		kAlert.alertSet.count = kAlert.alertSet.count + 1
		kAlert.alertSet.save()
		return true
	end
	
	return false
	
end

function kAlert.alertSet.delete(name)
	
	if kAlert.alertSet.alerts[name] ~= nil then
		kAlert.alertSet.alerts[name] = nil
		kAlert.alertSet.count = kAlert.alertSet.count - 1
		kAlert.alertSet.save()
		return true
	end
	
	return false
	
end

function kAlert.commandHandler(commandline)
	
	kAlert.screenObjects.init()
	
	if string.len(commandline) == 0 then
		kAlert.screenObjects.hide()
		kAlert.config.main()
	else
		local command = string.match(commandline,"set=%d+")
		if command then
			local setNumber = tonumber(string.match(command,"%d+"))
			if setNumber > 0 and setNumber < 6 then
				kAlertSet = setNumber
				kAlert.screenObjects.refresh()
			end
		else
			print("Karuul Alert Commands")
			print("Usage: /KaruulAlert {command}")
			print("Commands:")
			print("set={number} - to change current alert set")
		end
	end
	
end

table.insert(Command.Slash.Register("KAlert"), {kAlert.commandHandler, "kAlert", "config"})
table.insert(Command.Slash.Register("KaruulAlert"), {kAlert.commandHandler, "kAlert", "config"})

function kAlert.settingsHandler() 

	if kAlertVersion == nil then
		kAlertVersion = 1.00
	end
	
	if kAlertGlobalVersion == nil then
		kAlertGlobalVersion = 1.00
	end
	
	if kAlertSet == nil then
		kAlertSet = 1
	end
	
	if kAlertAlerts == nil then
		kAlertAlerts = {}
	end
	
	if kAlertGlobalItems == nil then
		kAlertGlobalItems = {}
	end
	
end

table.insert(Event.Addon.SavedVariables.Load.End, {kAlert.settingsHandler, "kAlert", "vaiablesLoaded"})

function kAlert.eventHandler()
	
	kAlert.screenObjects.init()
	
	local scanTime = Inspect.Time.Frame()
	if kAlert.systemScanner.nextScan < scanTime then
		kAlert.systemScanner.nextScan = scanTime + 0.2

		if not kAlert.config.active then
			for id, details in pairs (kAlert.screenObjects.object) do
				if details.duration > 0 then
					local remaining = details.duration - (scanTime - details.begin)
					details.setTimer(remaining)
					if details.type == 1 then -- Buff Timer
						if remaining < details.timerLength and details.timer and not details.typeToggle then details:SetVisible(true) end
					elseif details.type == 2 then -- Ability Timer
						if remaining < details.timerLength and details.timer and details.typeToggle then details:SetVisible(true) end
					end
				end
			end
			if kAlert.changeHandler.buffChanged then kAlert.processBuffs() end
			if kAlert.changeHandler.abilityChanged then kAlert.processAbilities() end
			if kAlert.changeHandler.resourceChanged then kAlert.processResources() end
			if kAlert.changeHandler.castingChanged then kAlert.processCasting() end
		end
		
		if kAlert.systemScanner.activeScanning then
			kAlert.systemScanner.scan()
		end
		
	end
	
end

function kAlert.updateAlerts()
	for id, details in pairs(kAlertAlerts) do
		if details.combatOnly == nil then details.combatOnly = false end
		if details.itemValue == nil then details.itemValue = 1 end
		if details.itemLength == nil then details.itemLength = 0 end
		if details.imageOpacity == nil then details.imageOpacity = 1 end
		if details.text == nil then details.text = "" end
		if details.textOpacity == nil then details.textOpacity = 1 end
		if details.textFont == nil then
			if details.timerFont ~= nil then
				details.textFont = details.timerFont
				details.timerFont = nil
			else
				details.textFont = ""
			end
		end
		if details.textRed == nil then details.textRed = 1 end
		if details.textGreen == nil then details.textGreen = 1 end
		if details.textBlue == nil then details.textBlue = 1 end
		if details.textSize == nil then details.textSize = 30 end
		if details.textLocation == nil then details.textLocation = "CENTER" end
		if details.textInside == nil then details.textInside = true end
		if details.timerLength == nil then details.timerLength = 5 end
		if details.selfCast == nil then details.selfCast = false end
		if details.range == nil then details.range = false end
		if details.rangeLow == nil then details.rangeLow = 0 end
		if details.rangeHigh == nil then details.rangeHigh = 0 end
	end
end

function kAlert.updateGlobalItems()
	for id, details in pairs(kAlertGlobalItems) do
		if id ~= details.name then
			kAlertGlobalItems[details.name] = details
			kAlertGlobalItems[id] = nil
		end
	end
end

function kAlert.screenObjects.init()

	if kAlert.screenObjects.object == nil then
		if kAlertVersion < kAlert.version then
			print("v" .. tostring(kAlertVersion) .. " -> v" .. tostring(kAlert.version))
			kAlertVersion = kAlert.version
		end
		kAlert.updateAlerts()
		kAlert.updateGlobalItems()
		print("v" .. tostring(kAlert.version))
		kAlert.screenObjects.object = {}
		for i = 1, kAlert.screenObjects.max, 1 do
			kAlert.screenObjects.object[i] = kAlert.screenObjects.addObject(i)
		end
		kAlert.screenObjects.refresh()
	end
	
end

function kAlert.screenObjects.addObject(id)

	local iID = tostring(id)
	local object = UI.CreateFrame("Frame", "objectFrame" .. iID, kAlert.context)
	object:SetWidth(75)
	object:SetHeight(75)
	object:SetPoint ("TOPLEFT", kAlert.context, "TOPLEFT")
	object:SetVisible(false)
	object:SetBackgroundColor(0, 0, 0, 0)
	object:SetLayer(-1)

	object.name = nil
	object.unit = nil
	object.type = nil
	object.typeToggle = nil
	object.range = false
	object.rangeLow = 0
	object.rangeHigh = 0
	object.itemId = nil
	object.itemName = nil
	object.itemValue = nil
	object.itemValuePercent = nil
	object.itemLength = 0
	object.timer = nil
	object.timerLength = 0
	object.combatOnly = nil
	object.begin = 0
	object.duration = 0
	object.selfCast = false
		
	object.image = UI.CreateFrame("Texture", "objectIcon" .. iID, object)
	object.image:SetPoint("CENTER", object, "CENTER", 0, 2)
	object.image:SetLayer(0)
	
	object.text = UI.CreateFrame("Text", "objectText" .. iID, object)
	object.text:SetText('')
	object.text:SetFontSize(object:GetWidth()/2)
	object.text:SetFontColor(1, 1, 1, 1)
	object.text:SetLayer(1)
	object.text:SetPoint("CENTER", object.image, "CENTER")
	
	object.counter = UI.CreateFrame("Text", "objectDuration" .. iID, object)
	object.counter:SetText('')
	object.counter:SetFontSize(object:GetWidth()/2)	
	object.counter:SetFontColor(1, 1, 1, 1)
	object.counter:SetLayer(2)
	object.counter:SetPoint("CENTER", object.image, "CENTER")
	object.counter:SetVisible(false)
	
	function object.setDims(width, height)
		object:SetWidth(width)
		object:SetHeight(height)
		object.image:SetWidth(width)
		object.image:SetHeight(height)
	end
	
	function object.setTimer(value)
		if value >= 0 then
			object.counter:SetText(tostring(math.floor(value)))
		else
			object.counter:SetText("")
		end
		object.counter:SetWidth(object.counter:GetFullWidth())
	end
	
	function object.setText(value)
		object.text:SetText(value)
		object.text:SetWidth(object.text:GetFullWidth())
	end
	
	return object
	
end

function kAlert.screenObjects.clear()
	for i = 1, kAlert.screenObjects.max, 1 do
		kAlert.screenObjects.object[i]:SetVisible(false)
		kAlert.screenObjects.object[i].name = nil
		kAlert.screenObjects.object[i].unit = nil
		kAlert.screenObjects.object[i].type = nil
		kAlert.screenObjects.object[i].typeToggle = nil
		kAlert.screenObjects.object[i].itemId = nil
		kAlert.screenObjects.object[i].itemName = nil
		kAlert.screenObjects.object[i].itemValue = nil
		kAlert.screenObjects.object[i].itemValuePercent = nil
		kAlert.screenObjects.object[i].itemLength = 0
		kAlert.screenObjects.object[i].setDims(10,10)
		kAlert.screenObjects.object[i].timer = nil
		kAlert.screenObjects.object[i].timerLength = 0
		kAlert.screenObjects.object[i].combatOnly = nil
		kAlert.screenObjects.object[i].begin = 0
		kAlert.screenObjects.object[i].duration = 0
		kAlert.screenObjects.object[i].setTimer(-1)
		kAlert.screenObjects.object[i].setText("")
		kAlert.screenObjects.object[i].selfCast = false
	end
end

function kAlert.screenObjects.refresh()

	local objectCount = 0
	kAlert.screenObjects.clear()
	kAlert.alertSet.load()
	
	for id, details in pairs (kAlert.alertSet.alerts) do
		if details.active == true and objectCount < kAlert.screenObjects.max then
			objectCount = objectCount + 1
			local screenObject = kAlert.screenObjects.object[objectCount]
			-- Check Data
			screenObject.name = details.name
			screenObject.unit = details.unit
			screenObject.type = details.type
			screenObject.typeToggle = details.typeToggle
			screenObject.range = details.range
			screenObject.rangeLow = details.rangeLow
			screenObject.rangeHigh = details.rangeHigh
			screenObject.itemId = details.itemId
			screenObject.itemLength = details.itemLength
			screenObject.selfCast = details.selfCast
			if details.type == 1 then
				screenObject.itemName = kUtils.split(",",details.itemName)
				if screenObject.itemLength ~= 0 then
					for id, itemName in pairs(screenObject.itemName) do
						screenObject.itemName[id] = itemName .. tostring(screenObject.itemLength)
					end
				end
				if screenObject.selfCast then
					for id, itemName in pairs(screenObject.itemName) do
						screenObject.itemName[id] = itemName .. "S"
					end
				end
			else
				screenObject.itemName = details.itemName
			end
			screenObject.itemValue = details.itemValue
			screenObject.itemValuePercent = details.itemValuePercent
			screenObject.combatOnly = details.combatOnly
			screenObject.timerLength = details.timerLength
			-- Image
			screenObject.setDims(details.imageWidth,details.imageHeight)
			screenObject.image:SetTexture(details.imageSource, details.image)
			screenObject.image:SetAlpha(details.imageOpacity)
			screenObject:SetPoint("TOPLEFT", kAlert.context, "TOPLEFT", details.imageX, details.imageY)
			-- Text
			screenObject.text:SetText(details.text)
			screenObject.text:ClearAll()
			screenObject.text:SetFontColor(details.textRed,details.textGreen,details.textBlue)
			screenObject.text:SetAlpha(details.textOpacity)
			if string.len(details.textFont) > 0 then
				screenObject.text:SetFont("kAlert",details.textFont)
			else
				screenObject.text:SetFont("Rift","$Flareserif_medium")
			end
			screenObject.text:SetFontSize(details.textSize)
			screenObject.text:SetHeight(details.textSize + 5)
			screenObject.text:SetWidth(screenObject.text:GetFullWidth())
			if details.textLocation == "CENTER" then
				screenObject.text:SetPoint("CENTER", screenObject.image, "CENTER")
			elseif details.textLocation == "TOPCENTER" then
				if details.textInside then
					screenObject.text:SetPoint("TOPCENTER", screenObject.image, "TOPCENTER")
				else
					screenObject.text:SetPoint("BOTTOMCENTER", screenObject.image, "TOPCENTER")
				end
			elseif details.textLocation == "BOTTOMCENTER" then
				if details.textInside then
					screenObject.text:SetPoint("BOTTOMCENTER", screenObject.image, "BOTTOMCENTER")
				else
					screenObject.text:SetPoint("TOPCENTER", screenObject.image, "BOTTOMCENTER")
				end
			elseif details.textLocation == "LEFTCENTER" then
				if details.textInside then
					screenObject.text:SetPoint("LEFTCENTER", screenObject.image, "LEFTCENTER")
				else
					screenObject.text:SetPoint("RIGHTCENTER", screenObject.image, "LEFTCENTER")
				end
			elseif details.textLocation == "RIGHTCENTER" then
				if details.textInside then
					screenObject.text:SetPoint("RIGHTCENTER", screenObject.image, "RIGHTCENTER")
				else
					screenObject.text:SetPoint("LEFTCENTER", screenObject.image, "RIGHTCENTER")
				end
			end
			-- Timer
			screenObject.timer = details.timer
			screenObject.counter:SetVisible(details.timer)
			screenObject.counter:ClearAll()
			if string.len(details.textFont) > 0 then
				screenObject.counter:SetFont("kAlert",details.textFont)
			else
				screenObject.counter:SetFont("Rift","$Flareserif_medium")
			end
			screenObject.counter:SetFontSize(details.timerSize)
			screenObject.counter:SetHeight(details.timerSize + 5)
			screenObject.counter:SetWidth(screenObject.counter:GetFullWidth())
			screenObject.counter:SetFontColor(details.textRed,details.textGreen,details.textBlue)
			screenObject.counter:SetAlpha(details.textOpacity)
			if details.timerLocation == "CENTER" then
				screenObject.counter:SetPoint("CENTER", screenObject.image, "CENTER")
			elseif details.timerLocation == "TOPCENTER" then
				if details.timerInside then
					screenObject.counter:SetPoint("TOPCENTER", screenObject.image, "TOPCENTER")
				else
					screenObject.counter:SetPoint("BOTTOMCENTER", screenObject.image, "TOPCENTER")
				end
			elseif details.timerLocation == "BOTTOMCENTER" then
				if details.timerInside then
					screenObject.counter:SetPoint("BOTTOMCENTER", screenObject.image, "BOTTOMCENTER")
				else
					screenObject.counter:SetPoint("TOPCENTER", screenObject.image, "BOTTOMCENTER")
				end
			elseif details.timerLocation == "LEFTCENTER" then
				if details.timerInside then
					screenObject.counter:SetPoint("LEFTCENTER", screenObject.image, "LEFTCENTER")
				else
					screenObject.counter:SetPoint("RIGHTCENTER", screenObject.image, "LEFTCENTER")
				end
			elseif details.timerLocation == "RIGHTCENTER" then
				if details.timerInside then
					screenObject.counter:SetPoint("RIGHTCENTER", screenObject.image, "RIGHTCENTER")
				else
					screenObject.counter:SetPoint("LEFTCENTER", screenObject.image, "RIGHTCENTER")
				end
			end
			screenObject:SetVisible(false)
		end
	end
	kAlert.processBuffs()
	kAlert.processAbilities()
	kAlert.processResources()
end

function kAlert.screenObjects.hide()
	for id, details in pairs (kAlert.screenObjects.object) do
		details:SetVisible(false)
	end
end

function kAlert.changeHandler.checkCasting(units)

	if not kAlert.config.active and not kAlert.changeHandler.castingChanged then
		local unitList = Inspect.Unit.List()
		for id, details in pairs(units) do
			for i = 1, table.getn(kAlert.units), 1 do
				if unitList[id] == kAlert.units[i] then kAlert.changeHandler.castingChanged = true end
			end
		end
	end

end

function kAlert.processCasting()

	local unitsResources = {}
	for i = 1, table.getn(kAlert.units), 1 do
		local spellCast = Inspect.Unit.Castbar(kAlert.units[i])
		if spellCast ~= nil then
			unitsResources[kAlert.units[i]] = spellCast.abilityName
		end
	end
	for id, details in pairs (kAlert.screenObjects.object) do
		if details.type ~= nil then
			if details.type == 4 then -- Casting
				local showObject = false
				if not details.combatOnly or (details.combatOnly and kAlert.combat) then
					if unitsResources[details.unit] ~= nil then
						if string.len(details.itemName) == 0 or unitsResources[details.unit] == details.itemName then
							showObject = true
						end
					end
				end
				details:SetVisible(showObject)
			end
		end
	end
	kAlert.changeHandler.castingChanged = false

end

function kAlert.changeHandler.checkUnitChanged(units)
	for id, details in pairs(units) do
		for i = 1, table.getn(kAlert.units), 1 do
			if details == kAlert.units[i] or (kAlert.unitList[id] ~= nil and not details) then
				if details then 
					kAlert.unitList[id] = details
				else
					kAlert.unitList[id] = nil
				end
				kAlert.changeHandler.buffChanged = true
				kAlert.changeHandler.resourceChanged = true
			end
		end
	end
end

function kAlert.changeHandler.checkBuffChanged(unit)

	if not kAlert.config.active and not kAlert.changeHandler.buffChanged then
		local unitList = Inspect.Unit.List()
		for i = 1, table.getn(kAlert.units), 1 do
			if unitList[unit] == kAlert.units[i] then kAlert.changeHandler.buffChanged = true end
		end
	end
	
end

function kAlert.processBuffs()

	kAlert.screenObjects.init()
	local unitsBuffs = {}
	for i = 1, table.getn(kAlert.units), 1 do
		local unitBuffs = Inspect.Buff.List(kAlert.units[i])
		if unitBuffs ~= nil then
			local buffList = {}
			if unitBuffs ~= nil then
				local buffDetails = Inspect.Buff.Detail(kAlert.units[i], unitBuffs)
				for id, details in pairs(buffDetails) do
					if details.name ~= nil then
						local buffStacks = 1
						if details.stack ~= nil then buffStacks = details.stack end
						if details.remaining == nil then
							buffList[details.name] = {remaining = 0, begin = 0, duration = 0, stacks = buffStacks}
							buffList[details.name .. "0"] = {remaining = 0, begin = 0, duration = 0, stacks = buffStacks}
							if kAlert.unitList[details.caster] == "player" then
								buffList[details.name .. "S"] = {remaining = 0, begin = 0, duration = 0, stacks = buffStacks}
								buffList[details.name .. "0S"] = {remaining = 0, begin = 0, duration = 0, stacks = buffStacks}
							end
						else
							buffList[details.name] = {remaining = details.remaining, begin = details.begin, duration = details.duration, stacks = buffStacks}
							buffList[details.name .. tostring(math.floor(details.duration + 0.5))] = {remaining = details.remaining, begin = details.begin, duration = details.duration, stacks = buffStacks}
							if kAlert.unitList[details.caster] == "player" then
								buffList[details.name .. "S"] = {remaining = details.remaining, begin = details.begin, duration = details.duration, stacks = buffStacks}
								buffList[details.name .. tostring(math.floor(details.duration + 0.5)) .. "S"] = {remaining = details.remaining, begin = details.begin, duration = details.duration, stacks = buffStacks}
							end
						end
					end
				end
			end
			unitsBuffs[kAlert.units[i]] = buffList
		end
	end
	
	for id, details in pairs (kAlert.screenObjects.object) do
		if details.type ~= nil then
			if details.type == 1 then -- Buff
				local showObject = false
				details.setTimer(-1)
				details.begin = 0
				details.duration = 0
				if not details.combatOnly or (details.combatOnly and kAlert.combat) then
					if unitsBuffs[details.unit] ~= nil then
						local buffDetails = nil
						for id, itemName in pairs(details.itemName) do
							if unitsBuffs[details.unit][itemName] ~= nil then
								buffDetails = unitsBuffs[details.unit][itemName]
							end
						end
						local timeLeft = details.timerLength
						if buffDetails ~= nil and buffDetails.stacks >= details.itemValue then
							timeLeft = buffDetails.remaining
							details.setTimer(timeLeft)
							details.begin = buffDetails.begin
							details.duration = buffDetails.duration
						end
						if timeLeft < details.timerLength and details.timer and not details.typeToggle then
							showObject = true
						else
							showObject = (((buffDetails ~= nil) and (buffDetails.stacks >= details.itemValue)) == details.typeToggle)
						end
					end
				end
				details:SetVisible(showObject)
			end
		end
	end

	kAlert.changeHandler.buffChanged = false

end

function kAlert.changeHandler.checkAbilityChanged()

	if not kAlert.config.active and not kAlert.changeHandler.abilityChanged then
		kAlert.changeHandler.abilityChanged = true
	end
	
end

function kAlert.processAbilities()

	if kAlert.config.active then return end
	
	kAlert.screenObjects.init()
	local activeAbilities = {}
	local abilityList = Inspect.Ability.List()
	if abilityList ~= nil then
		local abilityDetails = Inspect.Ability.Detail(abilityList)
		for id, details in pairs(abilityDetails) do
			if details.name ~= nil then
				if details.currentCooldownDuration ~= nil then
					if details.currentCooldownDuration > 2 and details.currentCooldownRemaining ~= nil then
						activeAbilities[details.name] =
						{
							remaining = details.currentCooldownRemaining, 
							begin = details.currentCooldownBegin, 
							duration = details.currentCooldownDuration
						}
					elseif details.unusable then
						activeAbilities[details.name] =
						{
							remaining = -1, 
							begin = -1, 
							duration = -1
						}
					end
				elseif details.unusable then
					activeAbilities[details.name] =
					{
						remaining = -1, 
						begin = -1, 
						duration = -1
					}
				end
			end
		end
	end

	for id, details in pairs (kAlert.screenObjects.object) do
		if details.type ~= nil then
			if details.type == 2 then -- Ability
				local showObject = false
				details.setTimer(-1)
				details.begin = 0
				details.duration = 0
				if not details.combatOnly or (details.combatOnly and kAlert.combat) then
					local abilityDetails = activeAbilities[details.itemName]
					local timeLeft = details.timerLength
					if abilityDetails ~= nil then
						timeLeft = abilityDetails.remaining
						details.setTimer(timeLeft)
						details.begin = abilityDetails.begin
						details.duration = abilityDetails.duration
					end
					if timeLeft < details.timerLength and details.timer and details.typeToggle then
						showObject = true
					else
						showObject = ((abilityDetails == nil) == details.typeToggle)
					end
				end
				details:SetVisible(showObject)
			end
		end
	end
	
	kAlert.changeHandler.abilityChanged = false
	
end

function kAlert.changeHandler.checkResourceChanged(units)

	if not kAlert.config.active and not kAlert.changeHandler.resourceChanged then
		local unitList = Inspect.Unit.List()
		for id, details in pairs(units) do
			for i = 1, table.getn(kAlert.units), 1 do
				if unitList[id] == kAlert.units[i] then kAlert.changeHandler.resourceChanged = true end
			end
		end
	end

end

function kAlert.processResources()

	if kAlert.config.active then return end
	
	kAlert.screenObjects.init()
	local unitsResources = {}
	for i = 1, table.getn(kAlert.units), 1 do
		local unitResources =  Inspect.Unit.Detail(kAlert.units[i])
		local resourceList = {}
		if unitResources ~= nil then
			if unitResources.health then
				resourceList[1] = {value = unitResources.health, percent = (unitResources.health/unitResources.healthMax)*100}
			end
			if unitResources.mana then
				resourceList[2] = {value = unitResources.mana, percent = (unitResources.mana/unitResources.manaMax)*100}
			end
			if unitResources.energy then
				resourceList[3] = {value = unitResources.energy, percent = (unitResources.energy/unitResources.energyMax)*100}
			end
			if unitResources.power then
				resourceList[4] = {value = unitResources.power, percent = unitResources.power}
			end
			if unitResources.charge then
				resourceList[5] = {value = unitResources.charge, percent = unitResources.charge}
			end
			if unitResources.combo then
				resourceList[6] = {value = unitResources.combo, percent = unitResources.combo}
			end
		end
		unitsResources[kAlert.units[i]] = resourceList
	end
	
	for id, details in pairs (kAlert.screenObjects.object) do
		if details.type ~= nil then
			if details.type == 3 then -- Resource
				local showObject = false
				if not details.combatOnly or (details.combatOnly and kAlert.combat) then
					if unitsResources[details.unit] ~= nil then
						if unitsResources[details.unit][details.itemId] ~= nil then
							local checkValue = 0
							if details.itemValuePercent then
								checkValue = unitsResources[details.unit][details.itemId].percent
							else
								checkValue = unitsResources[details.unit][details.itemId].value
							end
							if details.range then
								showObject = (checkValue >= details.rangeLow and checkValue <= details.rangeHigh)
							else
								showObject = ((checkValue >= details.itemValue) == details.typeToggle)
							end
						end
					end
				end
				details:SetVisible(showObject)
			end
		end
	end
	
	kAlert.changeHandler.resourceChanged = false

end

function kAlert.systemScanner.scan()

	local unitsBuffs = {}
	for i = 1, table.getn(kAlert.units), 1 do
		local unitBuffs = Inspect.Buff.List(kAlert.units[i])
		if unitBuffs ~= nil then
			local buffList = {}
			if unitBuffs ~= nil then
				local buffDetails = Inspect.Buff.Detail(kAlert.units[i], unitBuffs)
				for id, details in pairs(buffDetails) do
					if details.name ~= nil then
						if kAlertGlobalItems[details.name] == nil then
							kAlertGlobalItems[details.name] = {ability = details.ability, name = details.name, icon = details.icon}
							print(kAlertTexts.msgNewBuff .. ': ' .. details.name) -- .. '-' .. details.ability)
						end
					end
				end
			end
			unitsBuffs[kAlert.units[i]] = buffList
		end
	end
	
	local abilityList = Inspect.Ability.List()
	if abilityList then
		local abilityDetails = Inspect.Ability.Detail(abilityList)
		for id, details in pairs(abilityDetails) do
			if details.name ~= nil then
				if kAlertGlobalItems[details.name] == nil then
					kAlertGlobalItems[details.name] = {ability = id, name = details.name, icon = details.icon}
					print(kAlertTexts.msgNewAbility .. ': ' .. details.name) -- .. '-' .. id)
				end
			end
		end	
	end

end

kAlert.main()
