kUtils = {}

function kUtils.pairsByKeys(t, f)
	local a = {}
	for n in pairs(t) do table.insert(a, n) end
	table.sort(a, f)
	local i = 0      -- iterator variable
	local iter = function ()   -- iterator function
		i = i + 1
		if a[i] == nil then return nil
		else return a[i], t[a[i]]
		end
	end
	return iter
end

function kUtils.split(delimiter, text)
  local list = {}
  local pos = 1
  if string.find("", delimiter, 1) then -- this would result in endless loops
    return nil
  end
  while 1 do
    local first, last = string.find(text, delimiter, pos)
    if first then -- found?
      table.insert(list, string.sub(text, pos, first-1))
      pos = last+1
    else
      table.insert(list, string.sub(text, pos))
      break
    end
  end
  return list
end

function kUtils.join(delimiter, list)
  local len = getn(list)
  if len == 0 then 
    return "" 
  end
  local text = list[1]
  for i = 2, len do 
    text = text .. delimiter .. list[i] 
  end
  return text
end

function kUtils.buildFrame(name, parent, height, width)

	local frame = UI.CreateFrame("Frame", name, parent)
	frame:SetHeight(height)
	frame:SetWidth(width)
	
	return frame
	
end

function kUtils.buildLabel(name, parent, height, width, text)

	local label = UI.CreateFrame("Text", name, parent)
	label:SetText(text)
	label:SetFontSize(height)
	label:SetHeight(label:GetFullHeight())
	label:SetWidth(width)
	label:SetFontColor(1, 1, 1, 1)
	
	return label
	
end

function kUtils.buildSlider(name, parent, height, width, minVal, maxVal)

	local frame = UI.CreateFrame("RiftSlider", name, parent)
	frame:SetHeight(height)
	frame:SetWidth(width)
	frame:SetRange(minVal, maxVal)
	
	return frame
	
end

function kUtils.buildCheckBox(name, parent, height, width, label, reverse)

	local checkBox = UI.CreateFrame("RiftCheckbox", name, parent)
	checkBox:SetHeight(height)
	checkBox:SetWidth(height)
	checkBox.label = kUtils.buildLabel(name .. "Label", checkBox, height-5, width - height - 3, label)
	if reverse then
		checkBox.label:SetPoint("CENTERRIGHT", checkBox, "CENTERLEFT", 3, 0)
	else
		checkBox.label:SetPoint("CENTERLEFT", checkBox, "CENTERRIGHT", 3, 0)
	end
	
	return checkBox
	
end

function kUtils.buildButton(name, parent, height, width, label)

	local button = UI.CreateFrame("RiftButton", name, parent)
	button:SetText(label)
	button:SetHeight(height)
	button:SetWidth(width)
	
	return button
	
end

function kUtils.buildEditBox(name, parent, height, width, label)

	local editBox = UI.CreateFrame("Frame", name, parent)
	editBox:SetPoint("TOPLEFT", parent, "TOPLEFT")
	editBox:SetHeight(height)
	editBox:SetWidth(width)
	editBox.label = kUtils.buildLabel(name .. "Label", editBox, height - 5, width, label)
	editBox.label:SetWidth(editBox.label:GetFullWidth())
	editBox.label:SetPoint("CENTERLEFT", editBox, "CENTERLEFT")
	editBox.text = UI.CreateFrame("RiftTextfield", name .. "Text", editBox)
	editBox.text:SetPoint("CENTERLEFT", editBox.label, "CENTERRIGHT")
	editBox.text:SetHeight (height)
	editBox.text:SetWidth (width - editBox.label:GetWidth())
	editBox.text:SetBackgroundColor(0, 0, 0)
	
	return editBox
	
end

function kUtils.buildImageBox(name, parent, height, width, source, image)

	local imageBox = UI.CreateFrame("Texture", name, parent)
	imageBox:SetWidth(width)
	imageBox:SetHeight(height)
	imageBox:SetBackgroundColor(0, 0, 0, 0.5)
	imageBox:SetTexture(source, image)
	imageBox:SetLayer(0)
	
	function imageBox.setImage(source, image)
		imageBox.source = source
		imageBox.image = image
		imageBox:SetTexture(source, image)
	end
	
	return imageBox

end

function kUtils.buildToggleBox(name, parent, height, width, options, columns, allowNone)

	local toggleBox = UI.CreateFrame("Frame", name, parent)
	toggleBox:SetPoint("TOPLEFT", parent, "TOPLEFT", 30, 65)
	if columns == nil then columns = 1 end
	toggleBox:SetHeight(height * math.ceil(table.getn(options)/columns))
	toggleBox:SetWidth(width * columns)
	toggleBox:SetBackgroundColor(0, 0, 0, 0.5)
	toggleBox:SetLayer(-1)
	toggleBox.updateOptions = true
	toggleBox.cbOption = {}
	
	local row = 1
	for id, details in pairs (options) do
		option = kUtils.buildCheckBox(name .. "Option" .. tostring(id),toggleBox,height,width,details)
		if id == 1 then
			option:SetPoint("TOPLEFT", toggleBox, "TOPLEFT")
			option:SetChecked(true)
		elseif math.ceil(id/columns)-row == 0 then
			option:SetPoint("TOPLEFT", toggleBox.cbOption[id-1].label, "TOPRIGHT")
		else
			option:SetPoint("TOPLEFT", toggleBox.cbOption[id-columns], "BOTTOMLEFT")
			row = row + 1
		end
		if allowNone then
			function option.Event:CheckboxChange()
				if toggleBox.updateOptions then
					toggleBox.updateOptions = false
					local setValue = self:GetChecked()
					for i = 1, table.getn(options), 1 do
						toggleBox.cbOption[i]:SetChecked(false)
					end
					self:SetChecked(setValue)
					toggleBox.updateOptions = true
				end
				if toggleBox.change ~= nil then toggleBox.change() end
			end
		else
			function option.Event:CheckboxChange()
				if toggleBox.updateOptions then
					toggleBox.updateOptions = false
					for i = 1, table.getn(options), 1 do
						toggleBox.cbOption[i]:SetChecked(false)
					end
					self:SetChecked(true)
					toggleBox.updateOptions = true
				end
				if toggleBox.change ~= nil then toggleBox.change() end
			end
		end
		toggleBox.cbOption[id] = option
	end
	
	function toggleBox.setOptions(options)
		for i = 1, table.getn(toggleBox.cbOption), 1 do
			if options[i] ~= nil then toggleBox.cbOption[i].label:SetText(options[i]) end
		end
	end
	
	function toggleBox.getSelected()
		for i = 1, table.getn(options), 1 do
			if toggleBox.cbOption[i]:GetChecked() then return i end
		end
	end
	
	function toggleBox.setSelected(id)
		toggleBox.cbOption[id]:SetChecked(true)
	end
	
	return toggleBox
	
end

function kUtils.buildListPane(name, parent, height, width, size)

	local listPane = UI.CreateFrame("Frame", name, parent)
	listPane:SetHeight (height)
	listPane:SetWidth (width)
	listPane:SetBackgroundColor(0, 0, 0, 0.5)
	listPane:SetLayer(-1)
	local itemHight = listPane:GetHeight() / size
	listPane.itemMax = size
	listPane.lpItem = {}
	
	for i = 1, size, 1 do
		item = UI.CreateFrame("Text", name .. "Item" .. tostring(i), listPane)
		item:SetText("")
		item:SetFontSize(itemHight - 5)
		item:SetHeight(itemHight)
		item:SetWidth(listPane:GetWidth())
		item:SetFontColor(1, 1, 1, 1)
		item.selected = false
		
		if i == 1 then
			item:SetPoint("TOPLEFT", listPane, "TOPLEFT")
		else
			item:SetPoint("TOPLEFT", listPane.lpItem[i-1], "BOTTOMLEFT")
		end
		
		function item.Event:MouseIn()
			if listPane.onMouseOver ~= nil then listPane.onMouseOver() end
		end
		
		function item.Event:LeftClick()
			if string.len(self:GetText()) > 0 then
				for i = 1, size, 1 do
					listPane.lpItem[i]:SetBackgroundColor(1, 1, 1, 0)
					listPane.lpItem[i].selected = false
				end
				self:SetBackgroundColor(1, 1, 1, 0.5)
				self.selected = true
			end
			if listPane.click ~= nil then listPane.click("LEFT") end
		end
		
		listPane.lpItem[i] = item
	end

	function listPane.full()
		return not(listPane.itemCount < listPane.itemMax)
	end

	function listPane.clearSelected()
		for i = 1, listPane.itemMax, 1 do
			listPane.lpItem[i]:SetBackgroundColor(1, 1, 1, 0)
			listPane.lpItem[i].selected = false
		end
	end

	function listPane.getSelected()
		for i = 1, listPane.itemMax, 1 do
			if listPane.lpItem[i].selected then return listPane.lpItem[i]:GetText() end
		end
		return ""
	end

	function listPane.clearList()
		for i = 1, listPane.itemMax, 1 do
			listPane.lpItem[i]:SetText("")
			listPane.lpItem[i]:SetBackgroundColor(1, 1, 1, 0)
			listPane.lpItem[i].selected = false
		end
	end
	
	function listPane.fillList(items)
		local selectedItem = listPane.getSelected()
		listPane.clearList()
		listPane.itemCount = 0
		for id, details in kUtils.pairsByKeys(items) do
			if not listPane.full() then	
				listPane.itemCount = listPane.itemCount + 1
				listPane.lpItem[listPane.itemCount]:SetText(id)
			end
		end
	end
	
	function listPane.fillListSubSet(items,set)
		local selectedItem = listPane.getSelected()
		listPane.clearList()
		listPane.itemCount = 0
		for id, details in kUtils.pairsByKeys(items) do
			if not listPane.full() and set == string.sub(id,1,string.len(set)) then	
				listPane.itemCount = listPane.itemCount + 1
				listPane.lpItem[listPane.itemCount]:SetText(id)
			end
		end
		return listPane.itemCount
	end

	return listPane
	
end

function kUtils.buildWindowMover(name, parent)

	local windowMover = UI.CreateFrame("Frame", name, parent)
	windowMover:SetPoint("TOPLEFT", parent, "TOPLEFT")
	windowMover:SetHeight (50)
	windowMover:SetWidth (parent:GetWidth())
	windowMover:SetBackgroundColor(0, 0, 0, 0)

	function windowMover.Event:LeftDown()
		self.MouseDown = true
		mouseData = Inspect.Mouse()
		self.MyStartX = parent:GetLeft()
		self.MyStartY = parent:GetTop()
		self.StartX = mouseData.x - self.MyStartX
		self.StartY = mouseData.y - self.MyStartY
	end
	
	function windowMover.Event:MouseMove(mouseX, mouseY)
		if self.MouseDown then
			parent:SetPoint("TOPLEFT", UIParent, "TOPLEFT", (mouseX - self.StartX), (mouseY - self.StartY))
		end
	end
	
	function windowMover.Event:LeftUp()
		self.MouseDown = false
	end

	return windowMover
	
end

function kUtils.makeMovable(object) -- Note: only supports TOPLEFT at this point

	object.movable = true
	
	function object.Event:LeftDown()
		self.MouseDown = self.movable
		mouseData = Inspect.Mouse()
		self.MyStartX = self:GetLeft()
		self.MyStartY = self:GetTop()
		self.StartX = mouseData.x - self.MyStartX
		self.StartY = mouseData.y - self.MyStartY
	end
	
	function object.Event:MouseMove(mouseX, mouseY)
		if self.MouseDown then
			self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", (mouseX - self.StartX), (mouseY - self.StartY))
		end
	end
	
	function object.Event:LeftUp()
		self.MouseDown = false
	end
end