local info, data = ...
local me = {Name = "myUI Buff Boxes", Addon = info.identifier, SettingsKey = "BuffBoxes", ModuleVersion = "0.2.0", ReleaseType = "(public beta)"};
table.insert(Event.Addon.Load.End, {function(addon) if (addon == me.Addon) then myUI:RegisterModule(me)  end end, me.Addon, "_startup"})

local BuffBoxes = me
local L
local BuffBoxset
local BuffBox
local isnil, ifnil, notnil, iif, minmax, lpad, trim
local white, black, dodgerBlue, red, darkGoldenrod, green, blue, purple
local __setBoxsetDefaults
local q
local adaptiveUpdating 

local simdetail = {
	name = "IWin Buff!",
	icon = "/UI/ability_icons/headshot2.dds",
	duration = 20,
	remaining = 20
}

local _simulating = false

local function __updateSimDetails(casterid, isdebuff, now, elapsed)
	if (now ~= simdetail._now) then
		local newrem = simdetail.remaining - elapsed
		if (newrem < 0) then
			simdetail.remaining = simdetail.duration
		else
			simdetail.remaining = newrem
		end
		simdetail.debuff = isdebuff
		simdetail.begin = now + simdetail.remaining - simdetail.duration
		simdetail.caster = unitid
		simdetail._now = now
	end
	return myUI.global.shallowCopy(simdetail)
end

local function __getSimDetailList(barset, now, elapsed)
	local r = {}
	local unitid = Inspect.Unit.Lookup(barset.Unit)
	for i =1, 40, 1 do
		table.insert(r, __updateSimDetails(unitid, (i <= 20), now, elapsed))
	end
	return r
end

function BuffBoxes:GetDefaults()
	local defaults = {
		FirstRun = true,
		Enabled = true,
		Version = self.ModuleVersion,
		Boxsets = {
			["player"] = {x = 700, y = 200},
			["player.target"] = {x = 700, y = 400},
			["player.pet"] = {x = 700, y = 600}
		}
	}
	return defaults
end

function BuffBoxes:Initialize(profile)
	myUI.BuffBoxes = BuffBoxes
	L = self:LoadLanguage()
	
	profile.Version = self.ModuleVersion
	
	isnil = myUI.global.isnil
	ifnil = myUI.global.ifnil
	notnil = myUI.global.notnil
	iif = myUI.global.iif
	minmax = myUI.global.minmax
	lpad = myUI.string.lpad
	trim = myUI.string.trim
	
	local colors = myUI.colors
	
	white = colors.White
	black = colors.Black
	dodgerBlue = colors.DodgerBlue
	red = colors.Red
	darkGoldenrod = colors.DarkGoldenrod
	green = colors.Green
	blue = colors.Blue
	purple = colors.Purple
	
	self.Definition = profile.BuffBoxes
	self.BoxsetDefs = profile.BuffBoxes.Boxsets
	
	__setBoxsetDefaults(self)
	
	self.Boxsets = {}
	
	for unit, def in pairs(self.BoxsetDefs) do
		if (def.enabled) then
			if (unit == "group01") then
				for i = 1, 20, 1 do
					local u = "group" .. lpad(string.format("%d", i), 2, "0")
					self.Boxsets[u] = BuffBoxset:new(u, def)
				end
			elseif (unit == "group01.pet") then
				for i = 1, 20, 1 do
					local u = "group" .. lpad(string.format("%d", i), 2, "0") .. ".pet"
					self.Boxsets[u] = BuffBoxset:new(u, def)
				end
			else
				self.Boxsets[unit] = BuffBoxset:new(unit, def);
			end
		end
	end	
	
	myUI.Events.UpdateSlow:Subscribe("myUI.BuffBoxes.OnTimer", BuffBoxes, BuffBoxes.__onTimer);
	myUI.Events.UnitsChanged:Subscribe("myUI.BuffBoxes.OnUnitDetails", BuffBoxes, BuffBoxes.__onUnitDetails);
	myUI.Events.BuffsChanged:Subscribe("myUI.BuffBoxes.OnBuff", BuffBoxes, BuffBoxes.__onBuff);
	myUI.Events.AdaptiveUpdatingChanged:Subscribe("myUI.BuffBoxes.OnAdaptiveUpdatingChanged", BuffBoxes, BuffBoxes.__onAdaptiveUpdatingChanged);
	q = {}
	
	for _, boxset in pairs(self.Boxsets) do	
		table.insert(q, boxset.Unit)
		boxset:Layout() -- needed so group frames sync up properly
	end
	
	adaptiveUpdating = myUI:GetAdaptiveUpdating()
end

function BuffBoxes:ToggleMovers()
	for _, boxset in pairs(self.Boxsets) do
		boxset:ToggleMover();
	end
end
function BuffBoxes:ShowMovers()
	for _, boxset in pairs(self.Boxsets) do
		boxset:ShowMover();
	end
end
function BuffBoxes:HideMovers()
	for _, boxset in pairs(self.Boxsets) do
		boxset:HideMover();
	end
end

__setBoxsetDefaults = function(self)
	local fdefs = self.BoxsetDefs;
	
	-- remove groups that might be defined prior to v0.1.3. They now all use "group01" or "group01.pet", etc as a shared definition
	local remove = {}
	for unit, _ in pairs(fdefs) do
		if (string.find(unit, "group%d%d")) then
			local gnum = string.sub(unit, 6, 8)
			if (gnum ~= "01") then
				table.insert(remove, unit)
			end
		end
	end
	for _, unit in ipairs(remove) do
		fdefs[unit] = nil
	end
	
	local x = 200
	local y = 0
	
	for unit, def in pairs(fdefs) do
		def.enabled = ifnil(def.enabled, true)
		
		y = ifnil(def.y, y+200)
		def.y = ifnil(def.y, y)		
		def.x = ifnil(def.x, x)
		
		if (string.find(unit, "group")) then
			def.cxyBox = ifnil(def.cxyBox, 24)
			def.maxCols = ifnil(def.maxCols, 6)
			def.maxRows = ifnil(def.maxRows, 2)
			def.colspacing = ifnil(def.colspacing, 8)
			def.rowspacing = ifnil(def.rowspacing, 8)
		else
			def.cxyBox = ifnil(def.cxyBox, 36)
			def.maxCols = ifnil(def.maxCols, 6)
			def.maxRows = ifnil(def.maxRows, 2)
		end
		
		def.ownBuffs = ifnil(def.ownBuffs, true)
		def.ownDebuffs = ifnil(def.ownDebuffs, true)
		def.otherBuffs = ifnil(def.otherBuffs, true)
		def.otherDebuffs = ifnil(def.otherDebuffs, true)
		def.growVert = ifnil(def.growVert, "DOWN")
		def.growHorz = ifnil(def.growHorz, "RIGHT")
		def.anchorTo = ifnil(def.anchorTo, "TOPLEFT")
		def.anchorFrom = ifnil(def.anchorFrom, "TOPLEFT")
		def.anchor = ifnil(def.anchor,"SCREEN")
		
		def.whiteLists = ifnil(def.whiteLists, "")
		def.blackLists = ifnil(def.blackLists, "")
	end
end

local _configFrame = nil

function BuffBoxes:GetConfigFrame(parent)
	if (not _configFrame) then
		_configFrame = myUI:CreateFrame("Frame", parent)
		_configFrame:SetLayer(parent:GetLayer() + 1)
	
		local border = 16
	
		if (true) then -- just for frame collapsing in the editor
	
		local tabLabels = {
			L["Player"], 
			L["Target"], 
			L["Target Target"], 
			L["Pet"], 
			L["Pet Target"],
			L["Raid"], 
			L["Raid Pets"],
			L["Focus"],
			L["Focus Target"],
			L["Mouseover"],
			L["Mouseover Target"]
		}
		local tabunits = {
			"player", "player.target", "player.target.target", "player.pet", "player.pet.target",
			"group01", "group01.pet", "focus", "focus.target", "mouseover", "mouseover.target"
		}
		
		local f = myUI.Widgets.Tabstrip:new(tabLabels, "VERTICAL" , _configFrame)
		
		for i = 1, #tabunits, 1 do
			f:GetTab(i).value = tabunits[i]
		end
		
		f:SetPoint("TOPLEFT", _configFrame, "TOPLEFT", border, border)
		f:SetPoint("TOPRIGHT", _configFrame, "TOPRIGHT", -border, border)
		f:SetVisible(true)		
		f:SetSelectedIndex(1)
		f.SelectedIndexChanged = function(tabstrip, index)
			local unit = tabstrip:GetTab(index).value
			_configFrame:LoadUnit(unit)			
		end
		
		_configFrame.tabstrip = f
		
		f = myUI:CreateFrame("Frame", _configFrame)
		f:SetPoint("TOPLEFT", _configFrame.tabstrip, "TOPRIGHT")
		f:SetPoint("BOTTOMRIGHT", _configFrame, "BOTTOMRIGHT", -border, -border)
		f:SetBackgroundColor(_configFrame.tabstrip:GetTab(1):GetBackgroundColor())		
		_configFrame.tabpage = f
		
		f = myUI:CreateFrame("RiftCheckbox", _configFrame)
		f:SetLayer(parent:GetLayer() + 1)
		f:SetPoint("TOPLEFT", _configFrame.tabpage, "TOPLEFT", border, border)				
		_configFrame.chkEnabled = f
		
		f = myUI:CreateFrame("Text", _configFrame.chkEnabled)
		f:Autosize(L["Enabled"])
		f:SetPoint("LEFTCENTER", _configFrame.chkEnabled, "RIGHTCENTER", 4, 0)
		_configFrame.txtEnabled  = f
		
		local cxscreen = UIParent:GetWidth()
		local cyscreen = UIParent:GetHeight()
		
		f = myUI.Widgets.AnchorBox:new(_configFrame)
		f:SetPoint("TOPLEFT", _configFrame.chkEnabled, "BOTTOMLEFT", 160, border)
		f:SetScale(0.60)		
		_configFrame.anchorBox = f	

		f = myUI.Widgets.AnchorBox:new(_configFrame)
		f:SetPoint("TOPLEFT", _configFrame.anchorBox, "TOPRIGHT", border, 0)
		f:SetScale(0.60)
		
		f:SetBackgroundColor(1,0,0,0.2)
		_configFrame.anchorBox2 = f	
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Anchor"])		
		f:SetPoint("TOPLEFT", _configFrame.chkEnabled, "BOTTOMLEFT", 0,  border)
		_configFrame.lblAnchor = f
		
		f = myUI:CreateFrame("RiftCheckbox", _configFrame)
		f:SetPoint("TOPLEFT", _configFrame.lblAnchor,  "BOTTOMLEFT", border * 2, 0)
		_configFrame.chkAnchorScreen = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Screen"])
		f:SetPoint("LEFTCENTER", _configFrame.chkAnchorScreen, "RIGHTCENTER",  0, 0)
		_configFrame.lblAnchorScreen = f
		
		f = myUI:CreateFrame("RiftCheckbox", _configFrame)
		f:SetPoint("TOPLEFT", _configFrame.chkAnchorScreen,  "BOTTOMLEFT", 0, 0)		
		_configFrame.chkAnchorUF = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["myUI Unit Frame"])
		f:SetPoint("LEFTCENTER", _configFrame.chkAnchorUF, "RIGHTCENTER",  0, 0)
		_configFrame.lblAnchorUF = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["X-Offset:"])				
		f:SetPoint("TOPRIGHT", _configFrame.anchorBox, "BOTTOMLEFT", -border, border)
		_configFrame.lblX = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Y-Offset:"])
		f:SetPoint("TOPRIGHT", _configFrame.lblX, "BOTTOMRIGHT", 0, 4)
		_configFrame.lblY = f
		
		local activebg = myUI.color:clone(black, 100)
		
		f = myUI:CreateFrame("RiftTextfield", _configFrame)
		f:SetWidth(60)
		f:SetBackgroundColor(activebg:rgba())
		f:SetPoint("LEFTCENTER", _configFrame.lblX, "RIGHTCENTER", border, 0)		
		_configFrame.txtX = f
		
		f = myUI:CreateFrame("RiftTextfield", _configFrame)
		f:SetWidth(60)
		f:SetBackgroundColor(activebg:rgba())
		f:SetPoint("LEFTCENTER", _configFrame.lblY, "RIGHTCENTER", border, 0)	
		_configFrame.txtY = f
		
		f = myUI:CreateFrame("Text", _configFrame)
		f.max = math.floor(cxscreen)
		f.min = -f.max
		f:Autosize("(" .. tostring(f.min) .. " <-> " .. tostring(f.max) .. ")")
		f:SetPoint("LEFTCENTER", _configFrame.txtX, "RIGHTCENTER", border, 0)
		_configFrame.lblX2 = f
		
		f = myUI:CreateFrame("Text", _configFrame)
		f.max = math.floor(cyscreen)
		f.min = -f.max
		f:Autosize("(" .. tostring(f.min) .. " <-> " .. tostring(f.max) .. ")")
		f:SetPoint("LEFTCENTER", _configFrame.txtY, "RIGHTCENTER", border, 0)
		_configFrame.lblY2 = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Buff Size:"])
		f:SetPoint("TOPRIGHT", _configFrame.lblY, "BOTTOMRIGHT", 0, 4)
		_configFrame.lblCXY = f
		
		f = myUI:CreateFrame("RiftTextfield", _configFrame)
		f:SetWidth(60)
		f:SetBackgroundColor(activebg:rgba())
		f:SetPoint("LEFTCENTER", _configFrame.lblCXY, "RIGHTCENTER", border, 0)
		_configFrame.txtCXY = f
		
		f = myUI:CreateFrame("Text", _configFrame)
		f.max = 128
		f.min = 16
		f:Autosize("(" .. tostring(f.min) .. " <-> " .. tostring(f.max) .. ")")
		f:SetPoint("LEFTCENTER", _configFrame.txtCXY, "RIGHTCENTER", border, 0)
		_configFrame.lblCXY2 = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Max Columns:"])
		f:SetPoint("TOPRIGHT", _configFrame.lblCXY, "BOTTOMRIGHT", 0, 4)
		_configFrame.lblCols = f
		
		f = myUI:CreateFrame("RiftTextfield", _configFrame)
		f:SetWidth(60)
		f:SetBackgroundColor(activebg:rgba())
		f:SetPoint("LEFTCENTER", _configFrame.lblCols, "RIGHTCENTER", border, 0)
		_configFrame.txtCols = f
		
		f = myUI:CreateFrame("Text", _configFrame)
		f.max = 40
		f.min = 1
		f:Autosize("(" .. tostring(f.min) .. " <-> " .. tostring(f.max) .. ")")
		f:SetPoint("LEFTCENTER", _configFrame.txtCols, "RIGHTCENTER", border, 0)
		_configFrame.lblCols2 = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Max Rows:"])
		f:SetPoint("TOPRIGHT", _configFrame.lblCols, "BOTTOMRIGHT", 0, 4)
		_configFrame.lblRows = f
		
		f = myUI:CreateFrame("RiftTextfield", _configFrame)
		f:SetWidth(60)
		f:SetBackgroundColor(activebg:rgba())
		f:SetPoint("LEFTCENTER", _configFrame.lblRows, "RIGHTCENTER", border, 0)		
		_configFrame.txtRows = f
		
		f = myUI:CreateFrame("Text", _configFrame)
		f.max = 40
		f.min = 1
		f:Autosize("(" .. tostring(f.min) .. " <-> " .. tostring(f.max) .. ")")
		f:SetPoint("LEFTCENTER", _configFrame.txtRows, "RIGHTCENTER", border, 0)
		_configFrame.lblRows2 = f
				
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Grow:"])
		f:SetPoint("TOPRIGHT", _configFrame.lblRows, "BOTTOMRIGHT", 0, 4)
		_configFrame.lblGrow = f
		
		f = myUI:CreateFrame("RiftCheckbox", _configFrame)
		f:SetLayer(parent:GetLayer() + 1)
		f:SetPoint("TOPLEFT", _configFrame.txtRows, "BOTTOMLEFT", 0, 8)
		_configFrame.chkGrowLeft = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Left"])
		f:SetPoint("TOPLEFT", _configFrame.chkGrowLeft, "TOPRIGHT", 0, -2)
		_configFrame.lblGrowLeft = f
		
		f = myUI:CreateFrame("RiftCheckbox", _configFrame)
		f:SetLayer(parent:GetLayer() + 1)
		f:SetPoint("LEFTCENTER", _configFrame.chkGrowLeft, "RIGHTCENTER", 48, 0)
		_configFrame.chkGrowRight = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Right"])
		f:SetPoint("TOPLEFT", _configFrame.chkGrowRight, "TOPRIGHT", 0, -2)
		_configFrame.lblGrowRight = f
		
		f = myUI:CreateFrame("RiftCheckbox", _configFrame)
		f:SetPoint("TOPLEFT", _configFrame.chkGrowLeft, "BOTTOMLEFT", 0, 2)
		_configFrame.chkGrowUp = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Up"])
		f:SetPoint("TOPLEFT", _configFrame.chkGrowUp, "TOPRIGHT", 0, -2)
		_configFrame.lblGrowUp = f
		
		f = myUI:CreateFrame("RiftCheckbox", _configFrame)
		f:SetPoint("TOPLEFT", _configFrame.chkGrowRight, "BOTTOMLEFT", 0, 2)
		_configFrame.chkGrowDown = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Down"])
		f:SetPoint("TOPLEFT", _configFrame.chkGrowDown, "TOPRIGHT", 0, -2)
		_configFrame.lblGrowDown = f
				
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Show Buffs:"])
		f:SetPoint("BOTTOMLEFT", _configFrame.tabpage, "BOTTOMLEFT", border, -220)
		_configFrame.lblShowBuffs = f
		
		f = myUI:CreateFrame("RiftCheckbox", _configFrame)
		f:SetLayer(parent:GetLayer() + 1)
		f:SetPoint("TOPLEFT", _configFrame.lblShowBuffs, "TOPLEFT", 180, 2)
		_configFrame.chkOwnBuffs = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Self"])
		f:SetWidth(32)
		f:SetPoint("TOPLEFT", _configFrame.chkOwnBuffs, "TOPRIGHT", 0, -2)
		_configFrame.lblOwnBuffs = f
		
		f = myUI:CreateFrame("RiftCheckbox", _configFrame)
		f:SetPoint("TOPLEFT", _configFrame.chkOwnBuffs, "TOPRIGHT", 48, 0)
		_configFrame.chkOthersBuffs = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Others"])
		f:SetWidth(32)
		f:SetPoint("TOPLEFT", _configFrame.chkOthersBuffs, "TOPRIGHT", 0, -2)
		_configFrame.lblOthersBuffs = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Show Debuffs:"])
		f:SetPoint("TOPLEFT", _configFrame.lblShowBuffs, "BOTTOMLEFT", 0, 2)
		_configFrame.lblShowDebuffs = f
		
		f = myUI:CreateFrame("RiftCheckbox", _configFrame)
		f:SetPoint("TOPLEFT", _configFrame.lblShowDebuffs, "TOPLEFT", 180, 2)
		_configFrame.chkOwnDebuffs = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Self"])
		f:SetWidth(32)
		f:SetPoint("TOPLEFT", _configFrame.chkOwnDebuffs, "TOPRIGHT", 0, -2)
		_configFrame.lblOwnDebuffs = f
		
		f = myUI:CreateFrame("RiftCheckbox", _configFrame)
		f:SetPoint("TOPLEFT", _configFrame.chkOwnDebuffs, "TOPRIGHT", 48, 0)
		_configFrame.chkOthersDebuffs = f
		
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Others"])
		f:SetPoint("TOPLEFT", _configFrame.chkOthersDebuffs, "TOPRIGHT", 0, -2)
		_configFrame.lblOthersDebuffs = f
		
		
		
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["White Lists"])
		f:SetPoint("TOPLEFT", _configFrame.lblShowDebuffs, "BOTTOMLEFT", 0, border)
		_configFrame.lblWhiteLists = f
		
		f.Event.MouseIn = function(label) myUI:ShowTooltip(label, L["List Instructions"]) end
		f.Event.MouseOut = function(label) myUI:HideTooltip(label) end
		
		f = myUI:CreateFrame("RiftTextfield", _configFrame)
		f:SetBackgroundColor(activebg:rgba())
		f:SetWidth(300)
		f:SetPoint("LEFTCENTER", _configFrame.lblWhiteLists, "LEFTCENTER", 120, 0)
		_configFrame.txtWhiteLists = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Black Lists"])
		f:SetPoint("TOPLEFT", _configFrame.lblWhiteLists, "BOTTOMLEFT", 0, 4)
		_configFrame.lblBlackLists = f
		
		f.Event.MouseIn = function(label) myUI:ShowTooltip(label, L["List Instructions"]) end
		f.Event.MouseOut = function(label) myUI:HideTooltip(label) end
		
		
		f = myUI:CreateFrame("RiftTextfield", _configFrame)
		f:SetBackgroundColor(activebg:rgba())
		f:SetWidth(300)
		f:SetPoint("LEFTCENTER", _configFrame.lblBlackLists, "LEFTCENTER", 120, 0)
		_configFrame.txtBlackLists = f
		
		
		
		
		
		
		
		
		
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Spacing"])
		f:SetPoint("TOPRIGHT", _configFrame.lblGrow, "BOTTOMRIGHT", 0, 24)
		_configFrame.lblSpacing = f
		
		f = myUI:CreateFrame("RiftTextfield", _configFrame)
		f:SetBackgroundColor(activebg:rgba())
		f:SetWidth(30)
		f:SetPoint("LEFTCENTER", _configFrame.lblSpacing, "RIGHTCENTER", border, 0)
		_configFrame.txtColspacing = f
		
		f = myUI:CreateFrame("RiftTextfield", _configFrame)
		f:SetBackgroundColor(activebg:rgba())
		f:SetWidth(30)
		f:SetPoint("LEFTCENTER", _configFrame.txtColspacing, "RIGHTCENTER", 2, 0)
		_configFrame.txtRowspacing = f
		
		f = myUI:CreateFrame("Text", _configFrame)
		f.min = 0; f.max = 48
		f:Autosize("(" .. tostring(f.min) .. " <-> " .. tostring(f.max) .. ")")
		f:SetPoint("LEFTCENTER", _configFrame.txtRowspacing, "RIGHTCENTER", border, 0)
		_configFrame.lblSpacing2 = f
		
		f = myUI:CreateFrame("RiftButton", _configFrame)
		f:SetText(L["Apply"])
		f:SetPoint("BOTTOMLEFT", _configFrame.tabpage, "BOTTOMLEFT", border, -2)
		
		_configFrame.btnApply = f
		
		f = myUI:CreateFrame("Text", _configFrame)
		f:Autosize(L["myUI Unit Frames required"])
		f:SetPoint("BOTTOMLEFT", _configFrame.btnApply, "TOPLEFT", 0, -24)
		_configFrame.lblRequired = f
		
		
		end
		
		_configFrame.LoadUnit = function(self, unit)
			self._loading = true
			self._currentUnit = unit
			
			local bx = me.Boxsets[unit]
			local enabled = ((bx ~= nil) and (bx.Definition.enabled))
			
			
			local txts = {self.txtX, self.txtY, self.txtCXY, self.txtCols, self.txtRows, self.txtWhiteLists, self.txtBlackLists}
			local chks = {self.chkAnchorScreen, self.chkAnchorUF, self.chkGrowLeft, self.chkGrowRight,
				self.chkGrowUp, self.chkGrowDown, self.chkOwnBuffs, self.chkOwnDebuffs,
				self.chkOthersBuffs, self.chkOthersDebuffs
			}
			
			for _, txt in ipairs(txts) do txt:SetVisible(enabled) end
			for _, chk in ipairs(chks) do chk:SetEnabled(enabled) end			
			if (not enabled) then
				for _, txt in ipairs(txts) do txt:SetText("") end
				for _, chk in ipairs(chks) do chk:SetChecked(false) end
			end

			self.anchorBox:SetEnabled(enabled)
			self.anchorBox2:SetEnabled(enabled)
			self.chkEnabled:SetChecked(enabled)
			self.btnApply:SetEnabled(enabled)
			
			if (enabled) then
				local def = bx.Definition
				
				self.txtX:SetText(tostring(def.x))
				self.txtY:SetText(tostring(def.y))
				self.txtCXY:SetText(tostring(def.cxyBox))
				self.txtRows:SetText(tostring(def.maxRows))
				self.txtCols:SetText(tostring(def.maxCols))
				
				self.chkAnchorScreen:SetChecked(def.anchor == "SCREEN")
				self.chkAnchorUF:SetChecked(def.anchor == "UF")
				
				self.chkGrowUp:SetChecked(def.growVert == "UP")
				self.chkGrowDown:SetChecked(def.growVert == "DOWN")
				self.chkGrowLeft:SetChecked(def.growHorz == "LEFT")
				self.chkGrowRight:SetChecked(def.growHorz == "RIGHT")
				
				self.chkOwnBuffs:SetChecked(def.ownBuffs)
				self.chkOwnDebuffs:SetChecked(def.ownDebuffs)
				self.chkOthersBuffs:SetChecked(def.otherBuffs)
				self.chkOthersDebuffs:SetChecked(def.otherDebuffs)
				
				self.anchorBox:SetAnchor(def.anchorTo)
				self.anchorBox2:SetAnchor(def.anchorFrom)
				
				if (unit == "group01") or (unit == "group01.pet") then
					self.lblSpacing:SetVisible(true)
					self.lblSpacing2:SetVisible(true)
					self.txtColspacing:SetText(tostring(def.colspacing))
					self.txtColspacing:SetVisible(true)
					self.txtRowspacing:SetText(tostring(def.rowspacing))
					self.txtRowspacing:SetVisible(true)
				else
					self.txtColspacing:SetVisible(false)
					self.txtRowspacing:SetVisible(false)
					self.lblSpacing:SetVisible(false)
					self.lblSpacing2:SetVisible(false)
				end
				
				self.txtWhiteLists:SetText(ifnil(def.whiteLists, ""))
				self.txtBlackLists:SetText(ifnil(def.blackLists, ""))
			else
				self.anchorBox:SetAnchor("TOPLEFT")		
				self.anchorBox2:SetAnchor("TOPLEFT")
				
				self.txtColspacing:SetVisible(false)
				self.txtRowspacing:SetVisible(false)
				self.lblSpacing:SetVisible(false)
				self.lblSpacing2:SetVisible(false)
			end
			
			
			self._loading = false
		end
		
		_configFrame.chkAnchorScreen.Event.CheckboxChange = function(chk)
			if (_configFrame._loading) then return end			
			_configFrame.chkAnchorUF:SetChecked(not chk:GetChecked())
			
		end
		_configFrame.chkAnchorUF.Event.CheckboxChange = function(chk)
			if (_configFrame._loading) then return end			
			_configFrame.chkAnchorScreen:SetChecked(not chk:GetChecked())			
		end
		_configFrame.chkGrowLeft.Event.CheckboxChange = function(chk)
			if (_configFrame._loading) then return end			
			_configFrame.chkGrowRight:SetChecked(not chk:GetChecked())			
		end
		_configFrame.chkGrowRight.Event.CheckboxChange = function(chk)
			if (_configFrame._loading) then return end			
			_configFrame.chkGrowLeft:SetChecked(not chk:GetChecked())			
		end
		_configFrame.chkGrowUp.Event.CheckboxChange = function(chk)
			if (_configFrame._loading) then return end			
			_configFrame.chkGrowDown:SetChecked(not chk:GetChecked())			
		end
		_configFrame.chkGrowDown.Event.CheckboxChange = function(chk)
			if (_configFrame._loading) then return end			
			_configFrame.chkGrowUp:SetChecked(not chk:GetChecked())			
		end
		
		
		_configFrame.btnApply.Event.LeftClick = function(btn)
			local units = {}
			local raid = false
			
			if (_configFrame._currentUnit == "group01") then
				for i = 1, 20, 1 do					
					table.insert(units, "group" .. lpad(string.format("%d", i), 2, "0"))
				end
				raid = true
			elseif (_configFrame._currentUnit == "group01.pet") then
				for i = 1, 20, 1 do					
					table.insert(units, "group" .. lpad(string.format("%d", i), 2, "0") .. ".pet")
				end
				raid = true
			else
				table.insert(units, _configFrame._currentUnit)		
			end
		
			local x, y, cxy, cols, rows, colspacing, rowspacing
				
			if (pcall(function()
				x = 0+tonumber(_configFrame.txtX:GetText())
				y = 0+tonumber(_configFrame.txtY:GetText())
				cxy = 0+tonumber(_configFrame.txtCXY:GetText())
				cols = 0+tonumber(_configFrame.txtCols:GetText())
				rows = 0+tonumber(_configFrame.txtRows:GetText())
				
				if (raid) then
					colspacing = 0+tonumber(_configFrame.txtColspacing:GetText())
					rowspacing = 0+tonumber(_configFrame.txtRowspacing:GetText())				
				end
				
				end)) then
				
				if (x < _configFrame.lblX2.min) or (x > _configFrame.lblX2.max) then
					myUI.msgbox(_configFrame, L["Bad X"], _, "OK", "WARNING")
					return
				end
				if (y < _configFrame.lblY2.min) or (y > _configFrame.lblY2.max) then
					myUI.msgbox(_configFrame, L["Bad Y"], _, "OK", "WARNING")
					return
				end
				if (cxy < _configFrame.lblCXY2.min) or (cxy > _configFrame.lblCXY2.max) then
					myUI.msgbox(_configFrame, L["Bad CXY"], _, "OK", "WARNING")
					return 
				end
				if (cols < _configFrame.lblCols2.min) or (cols > _configFrame.lblCols2.max) then
					myUI.msgbox(_configFrame, L["Bad Cols"], _, "OK", "WARNING")
					return 
				end
				if (rows < _configFrame.lblRows2.min) or (rows > _configFrame.lblRows2.max) then
					myUI.msgbox(_configFrame, L["Bad Rows"], _, "OK", "WARNING")
					return 
				end
				
				if (raid) then
					if (colspacing < _configFrame.lblSpacing2.min) or (colspacing > _configFrame.lblSpacing2.max) then
						myUI.msgbox(_configFrame, L["Bad Col Spacing"], _, "OK", "WARNING")
						return
					end
					if (rowspacing < _configFrame.lblSpacing2.min) or (rowspacing > _configFrame.lblSpacing2.max) then
						myUI.msgbox(_configFrame, L["Bad Row Spacing"], _, "OK", "WARNING")
						return
					end
				end			
				
				for _, unit in ipairs(units) do
					local bx = BuffBoxes.Boxsets[unit]
					local def = bx.Definition
					def.x = x
					def.y = y
					def.cxyBox = cxy
					def.maxCols = cols
					def.maxRows = rows
					
					def.growVert = iif(_configFrame.chkGrowUp:GetChecked(), "UP", "DOWN")
					def.growHorz = iif(_configFrame.chkGrowLeft:GetChecked(), "LEFT", "RIGHT")
					
					def.ownBuffs = _configFrame.chkOwnBuffs:GetChecked()
					def.ownDebuffs = _configFrame.chkOwnDebuffs:GetChecked()
					def.otherBuffs = _configFrame.chkOthersBuffs:GetChecked()
					def.otherDebuffs = _configFrame.chkOthersDebuffs:GetChecked()
					
					def.anchor = iif(_configFrame.chkAnchorScreen:GetChecked(), "SCREEN", "UF")
					def.anchorTo = _configFrame.anchorBox:GetAnchor()
					def.anchorFrom = _configFrame.anchorBox2:GetAnchor()
					
					if (raid) then
						def.rowspacing = rowspacing
						def.colspacing = colspacing
					end
					
					def.whiteLists = myUI.string.trim(_configFrame.txtWhiteLists:GetText())
					def.blackLists = myUI.string.trim(_configFrame.txtBlackLists:GetText())
					
					local badlist = false
					if (string.len(def.whiteLists) > 0) then
						local lists = myUI.string.fromCSV(def.whiteLists)
						for _, list in ipairs(lists) do
							list = myUI.string.trim(list)
							if (not myUI.Lists:Exists(list)) then
								badlist = true
								myUI.msgbox(_configFrame, L["Missing White Lists"], nil, "OK", "WARNING")
								break
							end
						end
					end
					
					if (not badlist) and (string.len(def.blackLists) > 0) then
						local lists = myUI.string.fromCSV(def.blackLists)
						for _, list in ipairs(lists) do
							list = myUI.string.trim(list)
							if (not myUI.Lists:Exists(list)) then
								myUI.msgbox(_configFrame, L["Missing Black Lists"], nil, "OK", "WARNING")
								break
							end
						end
					end
					
					bx:RebuildLists()
					bx:Layout()
					bx:Resync()
				end
			else
				myUI.msgbox(_configFrame, L["Invalid value"], _, "OK", "WARNING")
				return
			end
			
		end
		
		_configFrame.chkEnabled.Event.CheckboxChange = function(chk)
			if (_configFrame._loading) then return end
			
			units = {}
			
			if (_configFrame._currentUnit == "group01") then
				for i = 1, 20, 1 do					
					table.insert(units, "group" .. lpad(string.format("%d", i), 2, "0"))
				end
			elseif (_configFrame._currentUnit == "group01.pet") then
				for i = 1, 20, 1 do					
					table.insert(units, "group" .. lpad(string.format("%d", i), 2, "0") .. ".pet")
				end
			else
				table.insert(units, _configFrame._currentUnit)			
			end
			
			for _, unit in ipairs(units) do
				BuffBoxes:SetUnitEnabled(unit, chk:GetChecked())
			end
			
			_configFrame:LoadUnit(_configFrame._currentUnit)			
		end
		
		_configFrame:LoadUnit("player")
		
	end
	
	return _configFrame
end

local __recycledBoxes = {}

function BuffBoxes:BeginSimulation()
	_simulating = true
end

function BuffBoxes:EndSimulation()
	_simulating = false
	for _, barset in pairs(me.Boxsets) do
		if (barset.Definition.enabled) then
			barset:Resync()
		end
	end
end

function BuffBoxes:GetSimulating()
	return _simulating
end

function BuffBoxes:SetUnitEnabled(unit, enabled)
	units = {}
	
	if (unit == "group01") then
		for i = 1, 20, 1 do					
			table.insert(units, "group" .. lpad(string.format("%d", i), 2, "0"))
		end
	elseif (unit == "group01.pet") then
		for i = 1, 20, 1 do					
			table.insert(units, "group" .. lpad(string.format("%d", i), 2, "0") .. ".pet")
		end
	else
		table.insert(units, unit) --_configFrame._currentUnit)			
	end
	
	for _, u in ipairs(units) do
		local box = self.Boxsets[u]
		if (box) then
			box.Definition.enabled = enabled			
		else
			if (enabled) then
				if (not self.BoxsetDefs[unit]) then -- note, we're using [unit] rather than [u] so that all our raid buff boxes share the same definition
					self.BoxsetDefs[unit] = {}
					__setBoxsetDefaults(self)
				end
				
				box = BuffBoxset:new(u, self.BoxsetDefs[unit])
				self.Boxsets[u] = box
				
				box.Definition.enabled = true
				table.insert(q, box.Unit)
			end		
		end		
	end
	
	for _, u in ipairs(units) do
		local box = self.Boxsets[u]
		if (box) then
			box:Layout()
			box:Resync()
		
			if (box.Definition.enabled) then
				box:ShowMover()
			else
				box:HideMover()
			end
		end
	end
	
	
	
end

function BuffBoxes:__onUnitDetails(units)		
	if (units) then
		for _, boxset in pairs(me.Boxsets) do
			if (boxset.Definition.enabled) then
				if (units[Inspect.Unit.Lookup(boxset.Unit)]) then
					boxset:Resync()
				end
			end
		end
	end
end

function BuffBoxes:__onBuff(unitID, buffs)	
	for _, boxset in pairs(me.Boxsets) do
		if (boxset.Definition.enabled) then
			if (unitID == Inspect.Unit.Lookup(boxset.Unit)) then
				boxset:Resync()
			end 
		end
	end
end



function BuffBoxes:__onAdaptiveUpdatingChanged(enabled)
	adaptiveUpdating = enabled
end
function BuffBoxes:__onTimer(now, elapsed)
	if (adaptiveUpdating) then
		local maxupdate = math.max(math.ceil(#q / 2), 14) -- anything under #q/2 creates problems due to not penetrating deeply enough into the queue on each update
		local nq = {}
		
		for i, unitid in ipairs(q) do
			if (i > maxupdate) then
				table.insert(nq, 1, unitid)
			else
				self.Boxsets[unitid]:TimerUpdate(now, elapsed)
				table.insert(nq, unitid)
			end
		end
		
		q = nq
	else
		for _, boxset in pairs(self.Boxsets) do		
			boxset:TimerUpdate(now, elapsed)		
		end
	end
end





local function __recycleBox(box)
	if (box ~= nil) then
		box:Reset()
		
		--print("bar recycled")
		table.insert(__recycledBoxes, box)
	end
end

local function __getBox(boxset)
	local bar
	
	if (#__recycledBoxes > 0) then
		bar = table.remove(__recycledBoxes)				
	else
		bar = BuffBox:new()
		
	end
	
	bar:Init(boxset)
	bar:Layout()
	return bar
end


local function __layoutBBS(self)
	local def = self.Definition
	local p
	
	local x = def.x
	local y = def.y
	local anchorTo = def.anchorTo
	local anchorFrom = def.anchorFrom
	
	--local pstr 
	
	if (def.anchor == "SCREEN") then
		if (string.find(self.Unit, "group%d%d")) then
			local gnum = tonumber(string.sub(self.Unit, 6, 8))
			local ispet = string.find(self.Unit, "pet")
			
			if (gnum ~= 1) then
				if (gnum == 6) or (gnum == 11) or (gnum == 16) then
					gnum = gnum - 5
					anchorFrom = "TOPLEFT"
					anchorTo = "TOPRIGHT"
					y = 0; x = def.colspacing
				else
					gnum = gnum - 1
					anchorFrom = "TOPLEFT"
					anchorTo = "BOTTOMLEFT"
					x = 0; y = def.rowspacing
				end
				
				local g = "group" .. lpad(string.format("%d", gnum), 2, "0")
				if (ispet) then g = g .. "." .. string.sub(self.Unit, ispet) end
				--pstr = g
				
				p = BuffBoxes.Boxsets[g]
				
				if (p == nil) then
					--print ("no p !!!! (" .. g .. ")")
					x = def.x
					y = def.y
					anchorTo = def.anchorTo
					anchorFrom = def.anchorFrom
					--pstr = nil
				end
			end
		end
	end
		
	if (def.anchor == "UF") and (myUI.UnitFrames) and (myUI.UnitFrames.Frames[self.Unit]) then
		p = myUI.UnitFrames.Frames[self.Unit]
	end

	local cx = def.cxyBox * def.maxCols
	local cy = def.cxyBox * def.maxRows
	
	self:ClearAll()
	self:SetWidth(cx)
	self:SetHeight(cy)
	
	p = ifnil(p, UIParent)
	
	
	self:SetPoint(anchorFrom, p, anchorTo, x, y)
	
	for _, box in ipairs(self.Boxes) do
		box:Layout()
	end
	
	self:LayoutBars()
end

local function __toggleMoverBBS(self)
	if (self._mover) then
		self:HideMover()
	elseif (self.Definition.enabled) then
		self:ShowMover()
	end
	
end
local function __showMoverBBS(self)
	if (self.Definition.enabled) then
		self._mover = true
		self:SetBackgroundColor(0.6,0,0,0.4)
		self.txtMove:Autosize(self.Unit); self.txtMove:SetVisible(true)
	end
end
local function __hideMoverBBS(self)
	self._mover = false
	self:SetBackgroundColor(0,0,0,0)
	self.txtMove:SetVisible(false)
end

local function __disposeBBS(self)
	if (notnil(self.Mover)) then self.Mover:Dispose(); self.Mover = nil; end

	for _, box in ipairs(self.Boxes) do
		__recycleBox(box)			
	end

	self.Boxes = nil
	self:_dispose()
end

local function __resyncBBS(self)
	self.resync = true
end


local function __timerUpdateBBS (self, now, elapsed)
	if (self.updating) then return; else self.updating = true end
	
	if (not self.Definition.enabled) or (not Inspect.Unit.Detail(self.Unit)) then 
		while #self.Boxes > 0 do
			__recycleBox(table.remove(self.Boxes))
		end
		self.resync = false
	end
	
	if (self.resync or _simulating) then
		local buffs = Inspect.Buff.List(self.Unit)
		
		local def = self.Definition
		if (buffs or _simulating) then
			local details
			if (_simulating) then
				details = __getSimDetailList(self, now, elapsed)
			else
				details = Inspect.Buff.Detail(self.Unit, buffs)
			end
			if (details) then
				local pid = Inspect.Unit.Lookup("player")
				local myuid = Inspect.Unit.Lookup(self.Unit)
				local goodbuffs = {}
				
				for id, buff in pairs(details) do
					if (not buff.expired) or (buff.expired >= 0) then
						buff.id = id						
						buff.playerCast = (pid == buff.caster)
					
						-- check our definition filters
						if (buff.playerCast and def.ownBuffs and not buff.debuff)
							or (buff.playerCast and def.ownDebuffs and buff.debuff)
							or (not buff.playerCast and def.otherBuffs and not buff.debuff)
							or (not buff.playerCast and def.otherDebuffs and buff.debuff)
							or (_simulating)
							then
						
							-- If there are both white and black lists, we've already applied the black list to the white, so 
							-- there's no need to ever evaluate both.
							
							if (self.wlist) then
								if (self.wlist[buff.name]) then
									table.insert(goodbuffs, buff)
								end
							elseif (self.blist) then
								if (not self.blist[buff.name]) then
									table.insert(goodbuffs, buff)
								end
							else						
								table.insert(goodbuffs, buff)
							end						
						end
					end							
				end
				
				if (#goodbuffs > 0) then
					table.sort(goodbuffs, function(a, b)
							-- to get into this table, the buff already has to have a duration so we don't need to check that
							if (a.playerCast and b.playerCast) then
								if (a.debuff ~= b.debuff) then
									return b.debuff
								else
									if (a.debuff and not b.debuff) then
										return false
									elseif (b.debuff and not a.debuff) then
										return true
									else
										if (notnil(a.remaining) and notnil(b.remaining)) then
											return a.remaining < b.remaining
										else
											return a.remaining
										end
									end
								end									
							else
								return a.playerCast
							end
							
						end
					)
					
					local maxA = self.Definition.maxRows * self.Definition.maxCols
					
					while #goodbuffs > maxA do
						table.remove(goodbuffs)
					end
				
				end
				while #goodbuffs > #self.Boxes do
					table.insert(self.Boxes, __getBox(self))
				end
				
				while #self.Boxes > #goodbuffs do
					__recycleBox(table.remove(self.Boxes))
				end
				
				--print("unit = ", self.Unit, " boxes = ", #self.Boxes, " goodbuffs = ", #goodbuffs)
				
				for i, v in ipairs(goodbuffs) do
					self.Boxes[i]:SetBuff(v)
				end			
				
				
			end			
		else
			while #self.Boxes > 0 do
				__recycleBox(table.remove(self.Boxes))
			end
				
			--print ("no buffs for ", self.Unit)
		end
	end

	for _, bar in ipairs(self.Boxes) do	
		bar:TimerUpdate(now, elapsed)
	end
	
	self.resync = false
	self:LayoutBars() -- we have to layout regardless of syncing b/c some bars may hide themselfs if their duration goes negative
	self.updating = false
	
	
end

local function __layoutBarsBBS(self)
	local def = self.Definition
	
	if (not def.enabled) then return end
	
	local row = 1
	local col = 1
	local maxRows = def.maxRows
	local maxCols = def.maxCols
	local n = 0
	local max = maxRows * maxCols
	
	local anchorNext = self
	local anchorRow = self
	
	local anchorFrom
	local anchorTo
	local anchorFromRow 
	local anchorToRow
	local anchorFromFirst
	local anchorToFirst
	
	if (def.growHorz == "RIGHT") then
		if (def.growVert == "DOWN") then
			anchorFromFirst = "TOPLEFT"; anchorToFirst = "TOPLEFT";
			anchorFrom = "TOPLEFT"; anchorTo = "TOPRIGHT";
			anchorFromRow = "TOPLEFT"; anchorToRow = "BOTTOMLEFT";			
		else
			anchorFromFirst = "BOTTOMLEFT"; anchorToFirst = "BOTTOMLEFT";
			anchorFrom = "BOTTOMLEFT"; anchorTo = "BOTTOMRIGHT";
			anchorFromRow = "BOTTOMLEFT"; anchorToRow = "TOPLEFT";
		end
	else
		if (def.growVert == "DOWN") then
			anchorFromFirst = "TOPRIGHT"; anchorToFirst = "TOPRIGHT";
			anchorFrom = "TOPRIGHT"; anchorTo = "TOPLEFT";
			anchorFromRow = "TOPRIGHT"; anchorToRow = "BOTTOMRIGHT";
		else
			anchorFromFirst = "BOTTOMRIGHT"; anchorToFirst = "BOTTOMRIGHT";
			anchorFrom = "BOTTOMRIGHT"; anchorTo= "BOTTOMLEFT";
			anchorFromRow = "BOTTOMRIGHT"; anchorToRow = "TOPRIGHT";
		end
	end
	
	for _, box in ipairs(self.Boxes) do
		if (n > max) then break end
		
		box:ClearAll()
		box:Layout()
		
		if (n == 0) then
			--box:SetPoint("TOPLEFT", self, "TOPLEFT", 0, 0)
			box:SetPoint(anchorFromFirst, self, anchorToFirst, 0, 0)
			anchorRow = box
		elseif (col >= maxCols) then
			col = 1
			--box:SetPoint("TOPLEFT", anchorRow, "BOTTOMLEFT", 0, 0)										
			box:SetPoint(anchorFromRow, anchorRow, anchorToRow, 0, 0)										
			anchorRow = box	
			row = row + 1
		else
			col = col + 1
			--box:SetPoint("TOPLEFT", anchorNext, "TOPRIGHT", 0, 0)					
			box:SetPoint(anchorFrom, anchorNext, anchorTo, 0, 0)
		end
		
		anchorNext = box
		n = n + 1
		
		--print (tostring(n))
	end

	
	
end

local function __buildListsBBS(self)
	local def = self.Definition
	if (def.whiteLists) and (trim(def.whiteLists) ~= "") then
		self.wlist = myUI.Lists:BuildWorkingList(def.whiteLists)
	else
		self.wlist = nil
	end
	if (def.blackLists) and (trim(def.blackLists) ~= "") then
		self.blist = myUI.Lists:BuildWorkingList(def.blackLists)
	else
		self.blist = nil
	end
	
	-- So, if both black and white lists, then we only show white buffs unless hidden by the black list.
	-- Therefore, we can simplify things by removing anything from both lists that are in common, which
	-- effectively applies the black list in advance.
	
	if (self.wlist) and (self.blist) then
		for buff, _ in pairs(self.blist) do
			if (self.wlist[buff]) then
				self.wlist[buff] = nil
			end
		end
		
		self.blist = nil
	end
end




local function __disposeBBS(self)
	for _, box in ipairs(self.Boxes) do
		box:Dispose()
	end
	self.Boxes = nil
	self.Unit = nil
	self.Definition = nil
	self.txtMove:Dispose(); self.txtMove = nil
	
	self:Dispose()
end
BuffBoxset = {
	new = function(self, unit, def)
		local this = myUI:CreateFrame("Frame")
		
		this.txtMove = myUI:CreateFrame("Text", this)
		this.txtMove:SetFontSize(14)
		this.txtMove:SetHeight(20)		
		this.txtMove:SetPoint("CENTER", this, "CENTER")
		
		this.txtMove:SetVisible(false)
	
		this:SetWidth(1)
		this:SetHeight(1)
	
		this.Definition = def
		this.Unit = unit
		this.resync = true
	
		this.Boxes = {}
		
		this.Layout = __layoutBBS	
		this.ToggleMover = __toggleMoverBBS	
		this.ShowMover = __showMoverBBS
		this.HideMover = __hideMoverBBS
		this._dispose = this.Dispose
		this.Dispose = __disposeBBS		
		this.Resync = __resyncBBS	
		this.TimerUpdate = __timerUpdateBBS
		this.LayoutBars = __layoutBarsBBS
		this.RebuildLists = __buildListsBBS
		
		this._dispose = this.Dispose
		this.Dispose = __disposeBBS
		this:Layout()
		this:RebuildLists()
		
		
		return this
	end

}


local function __dismissBuffBB(self)
	if (self.buff and not self.buff.noncancelable) then
		--print ("cancel buff " .. self.buff.id)
		Command.Buff.Cancel(self.buff.id)
		self:SetVisible(false)
		
	end
end

local function __setBuffBB(self, buff)
	self.syncing = true
	
	local clr			
	if (buff.debuff) then
		if (buff.disease) then
			clr = darkGoldenrod
		elseif (buff.poison) then
			clr = green
		elseif (buff.curse) then
			clr = purple
		else
			clr = red
		end
	else
		if (buff.noncancelable) then
			clr = blue
		else
			clr = dodgerBlue
		end
	end
	
	
	self.Icon:SetTexture("Rift", buff.icon)
	self:SetBackgroundColor(clr:rgba())
	self:SetAlpha(iif(buff.playerCast, 1, 0.6))
	
	--self.endsAt = buff.begin + buff.duration	
	self.buff = buff
	self.duration = buff.duration
	self.begin = buff.begin
	self.stack = buff.stack
	self.syncing = false	
end

local function __timerUpdateBB(self, now, elapsed)
	if (self.syncing) then return end
			
	local show = true
	if (self.duration and self.begin) then				
		local remaining = self.begin + self.duration - now
		if (remaining <= 0) then
			self.Duration:SetVisible(false)
			show = false
		else
			self.Duration:Autosize(self.timeString(remaining, true))			
			self.Duration:SetVisible(true)			
		end
	else
		self.Duration:SetVisible(false)
	end
	
	if (notnil(self.stack)) then
		self.Stacks:Autosize(tostring(self.stack))
		self.Stacks:SetVisible(true)
	else
		self.Stacks:SetVisible(false)
	end
	
	--self.Duration2:SetWidth(self.Duration:GetFullWidth())
	
	self.Duration2:Autosize(self.Duration:GetText())
	self.Duration2:SetVisible(self.Duration:GetVisible())
	
	self.Stacks2:Autosize(self.Stacks:GetText())
	self.Stacks2:SetVisible(self.Stacks:GetVisible())
	
	self:SetVisible(show)
end

local function __initBB(self,boxset)
	self.Buff = nil
	self.Boxset = boxset
	--self:SetParent(boxset)
	self:SetVisible(false)
	
	if (boxset.Unit == "player") then
		self.Event.RightClick = __dismissBuffBB
	end
end

local function __resetBB(self)	
	self.Boxset = nil
	self.Event.RightClick = nil
	self.Buff = nil
	self:SetVisible(false)
	self:ClearAll()
end

local function __layoutBB(self)
	local def = self.Boxset.Definition
	
	local cxyBox = def.cxyBox
	local cxyFont = math.floor(cxyBox * 0.35)
	cxyFont = minmax(cxyFont, cxyFont, 16)
	local sz = self.Duration:GetText()
	if (sz and sz~= "") then
		self.Duration:SetText("X")
	end
	
	if (cxyFont < 8) then
		self.Duration:SetVisible(false)
	else
		self.Duration:SetFontSize(cxyFont)
		self.Duration:SetHeight(cxyFont + 2)
		--self.Duration:SetWidth(self.Duration:GetFullWidth())
		--self.Duration:SetVisible(true)
	end
	
	self.Duration:SetText(sz)
	self.Duration2:SetFontSize(self.Duration:GetFontSize())
	self.Duration2:SetText(self.Duration:GetText())
	self.Duration2:SetHeight(self.Duration:GetHeight())
	--self.Duration2:SetWidth(self.Duration:GetWidth())
	self.Duration2:SetVisible(self.Duration:GetVisible())
	
	self.Stacks:SetFontSize(self.Duration:GetFontSize())
	self.Stacks:SetHeight(self.Duration:GetHeight())
	
	self.Stacks2:SetFontSize(self.Stacks:GetFontSize())
	self.Stacks2:SetText(self.Stacks:GetText())
	self.Stacks2:SetHeight(self.Stacks:GetHeight())
	self.Stacks2:SetVisible(self.Stacks:GetVisible())
	
	self:SetHeight(cxyBox)
	self:SetWidth(cxyBox)			
end


local function __mouseEnterBB(self)	
	if (self.buff) then
		local msg = ""
		
		if (self.buff.caster) then
			local detail = Inspect.Unit.Detail(self.buff.caster)
			if (detail) then
				msg = msg .. "Cast by: " .. detail.name
				if (detail.calling) then
					msg = msg .. "  (" .. L[detail.calling] .. ")"
				end
			end		
		end
		
		if (self.buff.curse) then
			msg = msg .. "\n\n" .. L["curse"]
		elseif (self.buff.disease) then
			msg = msg .. "\n\n" .. L["disease"]
		elseif (self.buff.poison) then
			msg = msg .. "\n\n" .. L["poison"] 
		end
		
		myUI:ShowTooltip(self, self.buff.name, msg)
	end
end
local function __mouseLeaveBB(self)
	myUI:HideTooltip(self)
end

BuffBox = {
	new = function(self)
		local this = myUI:CreateFrame("Frame")
		this.syncing = true
		this:SetLayer(15)
		
		this.Icon = myUI:CreateFrame("Texture", this)
		this.Icon:SetPoint("TOPLEFT", this, "TOPLEFT", 2, 2)
		this.Icon:SetPoint("BOTTOMRIGHT", this, "BOTTOMRIGHT", -2, -2)
		this.Icon:SetLayer(1)
		
		this.Duration = myUI:CreateFrame("Text", this)
		this.Duration:SetLayer(7)
		this.Duration:SetPoint("BOTTOMLEFT", this, "BOTTOMLEFT", 2, -4)
		--this.Duration:SetPoint("BOTTOMRIGHT", this, "BOTTOMRIGHT", 0, 0)
		this.Duration:SetFontColor(white:rgba())
		this.Duration:SetWordwrap(false)
		
		this.Duration2 = myUI:CreateFrame("Text", this)
		this.Duration2:SetLayer(6)
		this.Duration2:SetPoint("TOPLEFT", this.Duration, "TOPLEFT", 2, 0)
		--this.Duration2:SetPoint("BOTTOMRIGHT", this.Duration, "BOTTOMRIGHT", 2, 0)
		this.Duration2:SetFontColor(black:rgba())
		this.Duration2:SetWordwrap(false)
		
		this.Duration:SetBackgroundColor(0, 0, 0, 0.6)
		
		this.Stacks = myUI:CreateFrame("Text", this)
		this.Stacks:SetLayer(5)
		this.Stacks:SetPoint("TOPLEFT", this, "TOPLEFT", 2, 2)
		--this.Stacks:SetPoint("TOPRIGHT", this, "TOPRIGHT", 0, 0)
		this.Stacks:SetFontColor(white:rgba())
		this.Stacks:SetWordwrap(false)
		this.Stacks:SetBackgroundColor(0, 0, 0, 0.6)
		
		this.Stacks2 = myUI:CreateFrame("Text", this)
		this.Stacks2:SetLayer(4)
		this.Stacks2:SetPoint("TOPLEFT", this.Stacks, "TOPLEFT", 2, 0)
		--this.Stacks2:SetPoint("BOTTOMRIGHT", this.Stacks, "BOTTOMRIGHT", 2, 0)
		this.Stacks2:SetFontColor(black:rgba())
		this.Stacks2:SetWordwrap(false)
		
		this.Duration:SetVisible(false)
		this.Duration2:SetVisible(false)
		this.Stacks:SetVisible(false)
		this.Stacks2:SetVisible(false)
		
		this.timeString = myUI.string.getTimeString
		
		this.SetBuff = __setBuffBB		
		this.TimerUpdate = __timerUpdateBB
		this.Init = __initBB 	
		this.Reset = __resetBB
		this.Layout = __layoutBB
		
		this.Event.MouseIn = __mouseEnterBB
		this.Event.MouseOut = __mouseLeaveBB
		
		this:SetVisible(false)
		return this
	end
}



myUI.BuffBoxes = me

