local info, data = ...
local me = {Name = "myUI Unit Frames", Addon = info.identifier, SettingsKey = "UnitFrames", ModuleVersion="0.2.0", ReleaseType = "(public beta)"};
table.insert(Event.Addon.Load.End, {function(addon) if (addon == me.Addon) then myUI:RegisterModule(me)  end end, me.Addon, "_startup"})

myUI.UnitFrames = me
local UnitFrames = me

local UnitFrame, UnitFrameMover, BuffBox

local __setFrameDefaults, __setInCombat, __fadeCombat

local L
local isnil, ifnil, notnil, iif, lpad, minmax
local color, dodgerBlue, red, darkRed, white, black, yellow, lime, gray, firebrick, normal, transparent, deepskyblue
local q
local adaptiveUpdating 

local testdetail = {	
	calling = "mage", -- so it shows charge
	charge = 75,
	health = 800,
	healthMax = 1000,
	level = 50,
	mana = 600,
	manaMax = 1000,
	name = "#",
	guaranteedLoot = true,
	mark = 4,
	planar = 1,
	planarMax = 4,
	player = true,
	ready = true,
	role = "dps",
	vitality = 50,
	pvp = true	
}

local testcdetail = {
	abilityName = "Can o' whoopass !",
	duration = 2,
	remaining = 0.2	
}

local simdetail = {	
	calling = "mage", -- so it shows charge
	charge = 0,
	health = 1000,
	healthMax = 1000,
	level = 50,
	mana = 1000,
	manaMax = 1000,
	name = "#",
	guaranteedLoot = true,
	player = true,
	role = "dps"
}
local simcdetail = {
	abilityName = "Can o' whoopass !",
	duration = 2,
	remaining = 2
}
local _configFrame = nil
local _simulating = false

local function __setFrameDefaults(self)
	local fdefs = self.FrameDefs;

	self.Definition.Version = self.ModuleVersion
	
	-- remove groups that might be defined prior to v0.1.3. They now all use "group01" or "group01.pet", etc as a shared definition
	local remove = {}
	for unit, _ in pairs(fdefs) do
		if (string.find(unit, "group%d%d")) then
			local gnum = string.sub(unit, 6, 8)
			if (gnum ~= "01") then
				table.insert(remove, unit)
			end
		end
	end
	for _, unit in ipairs(remove) do
		fdefs[unit] = nil
	end
	-- end v0.1.3 update
	
	local defx = 100
	local defy = 200
	
	local bigcx = 320
	local bigcy = 80
	
	local elsecx = 140
	local elsecy = 70
	
	for unit, def in pairs(fdefs) do
		def.enabled = ifnil(def.enabled, true)
		
		
		local isgroup = (unit == "group01")
		local isgrouppet = (unit == "group01.pet")
		local notgroup = not (isgroup or isgrouppet)
		local isplayer = (unit == "player")
		
		if (isgroup or isgrouppet) then
			def.colspacing = ifnil(def.colspacing, 4)
			def.rowspacing = ifnil(def.rowspacing, 4)
		end
		
		def.showPvP = ifnil(def.showPvP, notgroup);
		def.showVitality = ifnil(def.showVitality, isplayer);
		def.showLevel = ifnil(def.showLevel, notgroup);
		def.showPortrait = ifnil(def.showPortrait, notgroup);			
		
		def.showHealthText = ifnil(def.showHealthText, true);
		def.showHealthAsPercent = ifnil(def.showHealthAsPercent, false)
		def.showPowerText = ifnil(def.showManaText, notgroup);
		def.showPowerAsPercent = ifnil(def.showPowerAsPercent, false)
		def.showTitleBar = ifnil(def.showTitleBar, notgroup);
		def.showCastBar = ifnil(def.showCastBar, notgroup);
		def.showPlanar = ifnil(def.showPlanar, isplayer);
		def.showRole = ifnil(def.showRole, true);
		def.showReady = ifnil(def.showReady, isgroup)
		def.showMark = ifnil(def.showMark, true);
		def.showLoot = ifnil(def.showLoot, (not isplayer));
		def.showCharge = ifnil(def.showChange, isplayer)
		
		def.anchorTo = ifnil(def.anchorTo, "TOPLEFT");
		def.anchorFrom = ifnil(def.anchorFrom, "TOPLEFT");
		
		def.followMouse = ifnil(def.followMouse, false)
		
		if (isnil(def.x)) then
			def.x = defx
		end
		if (isnil(def.y)) then 
			def.y = defy
			defy = defy + bigcy + 8
		end
		
		if (notgroup) or (def.isName) then
			def.cx = ifnil(def.cx, bigcx);
			def.cy = ifnil(def.cy, bigcy);
		else
			def.cx = ifnil(def.cx, elsecx);
			def.cy = ifnil(def.cy, elsecy);
		end
		
		if (isplayer) then
			if (def.longBuffs) then
				def.longBuffs.enabled = ifnil(def.longBuffs.enabled, true)
				def.longBuffs.cxyExpired = ifnil(def.longBuffs.cxyExpired, 36);
				def.longBuffs.cxyNormal = ifnil(def.longBuffs.cxyNormal, 24);
				def.longBuffs.anchor = ifnil(def.longBuffs.anchor, "TOP");
			end
		else
			def.longBuffs = nil
		end
		
		
	end

end
local function __setInCombat(self, inCombat)
	if (self.CombatIndicator) then 
		if (incombat) then 
			self.CombatIndicator:SetAlpha(1) 
			self.CombatIndicator.delta = nil
		end
		self.CombatIndicator:SetVisible(inCombat)		
	end
end
local function __fadeCombat(self, elapsed)
	local c = self.CombatIndicator
	if (c) and (c:GetVisible()) then
		local aold = c:GetAlpha()
		c.delta = ifnil(c.delta, -0.1)
		local anew = aold + c.delta
		
		if (anew >= 1) then
			c.delta = -0.1
		elseif (anew <= 0) then
			c.delta = 0.1
		end
		
		c:SetAlpha(minmax(anew, 0, 1))
	
	end
end



local function  __updateSimDetails()
	local newcharge = simdetail.charge + 10
	if (newcharge > 100) then newcharge = 0 end
	simdetail.charge = newcharge
	
	local newhealth = simdetail.health + math.random(-simdetail.healthMax / 2, simdetail.healthMax / 2)	
	simdetail.health = minmax(newhealth, 0, simdetail.healthMax)
	
	local newmana = simdetail.mana - 50
	if (newmana < 0) then newmana = simdetail.manaMax end
	simdetail.mana = newmana	
	
	return myUI.global.shallowCopy(simdetail)
end
local function __updateSimCastDetails(now, elapsed)
	local newrem = simcdetail.remaining - elapsed
	if (newrem < 0) then
		simcdetail.remaining = simcdetail.duration
	else
		simcdetail.remaining = newrem
	end
	
	simcdetail.begin = now + simcdetail.remaining - simcdetail.duration
	
	--return myUI.global.shallowCopy(simcdetail)
end


function UnitFrames:GetDefaults()
	local defaults = {
		FirstRun = true,
		Enabled = true,
		Version = self.ModuleVersion,
		Frames = {
			["player"] = {x = 200, y = 200},
			["player.target"] = {x = 200, y = 300},
			["player.target.target"] = {x = 200, y = 400},
			["player.pet"] = {x = 200, y = 500},
			["focus"] = {x = 200, y = 600},
			["group01"] = {x = 560, y = 200}
		}
	}
	return defaults
end

function UnitFrames:Initialize(profile)
	myUI.UnitFrames = self
	
	profile.Version = self.ModuleVersion
	
	isnil = myUI.global.isnil
	ifnil = myUI.global.ifnil
	notnil = myUI.global.notnil
	iif = myUI.global.iif
	lpad = myUI.string.lpad
	minmax = myUI.global.minmax
	color = myUI.color
	
	local colors = myUI.colors
	dodgerBlue = colors.DodgerBlue
	red = colors.Red
	darkRed = colors.DarkRed
	white = colors.White
	black = colors.Black
	yellow = colors.Yellow
	lime = colors.Lime
	gray = colors.Gray
	firebrick = colors.Firebrick
	transparent = colors.Transparent
	deepskyblue = colors.DeepSkyBlue
	
	L = self:LoadLanguage()
	
	-- Setup some UnitFrame-wide constants
	UnitFrames.Border = 2;	
	UnitFrames.bgNormal = color:new(0, 0, 0, 200);
	UnitFrames.bgLowMana = dodgerBlue;
	UnitFrames.bgLowHealth = darkRed;
	UnitFrames.lowHealth = 0.4;
	UnitFrames.lowMana = 0.3;
	UnitFrames.longBuffDuration = 60 * 29; -- self-cast buffs over this number of seconds will be monitored
	UnitFrames.expiredBuff = red;
	UnitFrames.shortBuff = yellow;
	
	normal = UnitFrames.bgNormal
	
	self.Definition = profile.UnitFrames
	self.FrameDefs = profile.UnitFrames.Frames
	
	__setFrameDefaults(self);
	self.Frames = {} -- key = unitID or specifier, value = unit frame
	
	for unit, def in pairs(UnitFrames.FrameDefs) do
		if (not def.isName) and (def.enabled) then		
			if (unit == "group01") then
				for i = 1, 20, 1 do
					local u = "group" .. lpad(string.format("%d", i), 2, "0")
					UnitFrames.Frames[u] = UnitFrame:new(u, def)
				end
			elseif (unit == "group01.pet") then
				for i = 1, 20, 1 do
					local u = "group" .. lpad(string.format("%d", i), 2, "0") .. ".pet"
					UnitFrames.Frames[u] = UnitFrame:new(u, def)
				end
			else
				UnitFrames.Frames[unit] = UnitFrame:new(unit, def);
			end
		end
	end
	
	--[[
	table.insert(Event.Unit.Available, {UnitFrames.OnUnitAvailable, "myUIUF", "ufOnUnitAvailable"});
	table.insert(Event.Unit.Unavailable, {UnitFrames.OnUnitAvailable, "myUIUF", "ufOnUnitAvailable"});
	table.insert(Event.Unit.Add, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	table.insert(Event.Unit.Remove, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	]]--
	--[[
	table.insert(Event.Unit.Detail.Offline, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	table.insert(Event.Unit.Detail.Afk,{UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	
	table.insert(Event.Unit.Detail.Name, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	table.insert(Event.Unit.Detail.Level, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	
	table.insert(Event.Unit.Detail.Health, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	table.insert(Event.Unit.Detail.HealthMax, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	table.insert(Event.Unit.Detail.HealthCap, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	
	table.insert(Event.Unit.Detail.Mana, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	table.insert(Event.Unit.Detail.ManaMax, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	table.insert(Event.Unit.Detail.Power, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	table.insert(Event.Unit.Detail.Energy, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	table.insert(Event.Unit.Detail.EnergyMax, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	
	table.insert(Event.Unit.Detail.Pvp, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	table.insert(Event.Unit.Detail.Vitality, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	table.insert(Event.Unit.Detail.Planar, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	table.insert(Event.Unit.Detail.PlanarMax, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	table.insert(Event.Unit.Detail.Mark, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	table.insert(Event.Unit.Detail.Charge, {UnitFrames.OnUnitDetail, "myUIUF", "ufOnUnitDetail"});
	]]--
	table.insert(Event.Unit.Castbar, {UnitFrames.OnCastbar, "myUIUF", "ufOnCastbar"});
	--[[
	table.insert(Event.Buff.Add, {UnitFrames.OnBuff, "myUIUF", "ufOnBuff"});
	table.insert(Event.Buff.Remove, {UnitFrames.OnBuff, "myUIUF", "ufOnBuff"});
	table.insert(Event.Buff.Change, {UnitFrames.OnBuff, "myUIUF", "ufOnBuff"});
	]]--
	myUI.Events.UnitsChanged:Subscribe("myUI.UnitFrames.OnUnitAvailable", UnitFrames, UnitFrames.OnUnitAvailable);
	myUI.Events.UnitUpdate:Subscribe("myUI.UnitFrames.OnUnitUpdate", UnitFrames, UnitFrames.OnUnitUpdate);
	myUI.Events.Update:Subscribe("myUI.UnitFrames.OnUpdate", UnitFrames, UnitFrames.OnUpdate);
	myUI.Events.CombatChanged:Subscribe("myUI.UnitFrames.OnCombat", UnitFrames, UnitFrames.OnCombat);
	myUI.Events.AdaptiveUpdatingChanged:Subscribe("myUI.UnitFrames.OnAdaptiveUpdatingChanged", UnitFrames, UnitFrames.__onAdaptiveUpdatingChanged);
	
	q = {}
	adaptiveUpdating = myUI:GetAdaptiveUpdating()
	
	for _, uf in pairs(self.Frames) do
		table.insert(q, uf.UnitID)
		uf:Layout() -- so group frames can sync properly
	end
	
	
end





function UnitFrames:SetUnitEnabled(unit, enabled)
	local defid = unit -- note: all raid frames share a common definition under "group01" ...
	local isgroup = string.find(unit, "group%d%d")
	local postgroup = ""
	if (isgroup) then
		defid = "group01"
		if (string.len(unit) > 7) then postgroup = string.sub(unit, 8) end
		defid = defid .. postgroup
	end
	
	local units = {}
	if (isgroup) then
		for i = 1, 20, 1 do
			local gid = "group" .. lpad(string.format("%d", i), 2, "0") .. postgroup
			table.insert(units, gid)
		end
	else
		table.insert(units, unit)
	end
	
	for _, u in ipairs(units) do
		local uf = self.Frames[u]
		if (uf) then
			uf.Definition.enabled = enabled
		else
			if (enabled) then
				if (not self.FrameDefs[defid]) then
					self.FrameDefs[defid] = {}
					__setFrameDefaults(self)
				end
				
				uf = UnitFrame:new(u, self.FrameDefs[defid])
				self.Frames[u] = uf
				uf.Definition.enabled = true
				table.insert(q, uf.UnitID)
			end
		end
	end
	
	-- do this in a seperate loop b/c if we created unit frames we need them all established before
	-- doing our final layout as they anchor from each other
	for _, u in ipairs(units) do
		local uf = self.Frames[u]
		if (uf) then
			uf:Reset()
			
			if (uf.Definition.enabled) then
				uf:ShowMover()
			else
				uf:HideMover()
			end
		end
	end	
end

function UnitFrames:ResetUnit(unit)	
	local isgroup = string.find(unit, "group%d%d")
	if (isgroup) then
		local postgroup = ""
		if (string.len(unit) > 7) then postgroup = string.sub(unit, 8) end
		
		for i = 1, 20, 1 do
			local gid = "group" .. lpad(string.format("%d", i), 2, "0") .. postgroup
			if (self.Frames[gid]) then self.Frames[gid]:Reset() end
		
		end
	else
		if (self.Frames[unit]) then self.Frames[unit]:Reset() end
	end
end

function UnitFrames:GetConfigFrame(parent)
	if (_configFrame == nill) then
		_configFrame = myUI:CreateFrame("Frame", parent)
		_configFrame:SetLayer(parent:GetLayer() + 1)
	
		local border = 16
	
		if (true) then -- just for frame collapsing in the editor
	
		local labels = {
			L["Player"], 
			L["Target"], 
			L["Target Target"], 
			L["Pet"], 
			L["Pet Target"],
			L["Raid"], 
			L["Raid Pets"],
			L["Focus"],
			L["Focus Target"],
			L["Mouseover"],
			L["Mouseover Target"]
		}
		local tabunits = {
			"player", "player.target", "player.target.target", "player.pet", "player.pet.target",
			"group01", "group01.pet", "focus", "focus.target", "mouseover", "mouseover.target"
		}
		
		local f = myUI.Widgets.Tabstrip:new(labels, "VERTICAL" , _configFrame)
		
		for i = 1, #tabunits, 1 do
			f:GetTab(i).value = tabunits[i]

		end
		
		f:SetPoint("TOPLEFT", _configFrame, "TOPLEFT", border, border)
		f:SetPoint("TOPRIGHT", _configFrame, "TOPRIGHT", -border, border)
		f:SetVisible(true)		
		f:SetSelectedIndex(1)
		f.SelectedIndexChanged = function(tabstrip, index)
			local unit = tabstrip:GetTab(index).value
			_configFrame:LoadUnit(unit)
			
		end
		
		_configFrame.tabstrip = f
		
		f = myUI:CreateFrame("Frame", _configFrame)
		f:SetPoint("TOPLEFT", _configFrame.tabstrip, "TOPRIGHT")
		f:SetPoint("BOTTOMRIGHT", _configFrame, "BOTTOMRIGHT", -border, -border)
		f:SetBackgroundColor(_configFrame.tabstrip:GetTab(1):GetBackgroundColor())
		
		_configFrame.tabpage = f
		
		f = myUI:CreateFrame("RiftCheckbox", _configFrame)
		f:SetLayer(parent:GetLayer() + 1)
		f:SetPoint("TOPLEFT", _configFrame.tabpage, "TOPLEFT", border, border)				
		_configFrame.chkEnabled = f
		
		f = myUI:CreateFrame("Text", _configFrame.chkEnabled); f:Autosize(L["Enabled"])
		f:SetPoint("LEFTCENTER", _configFrame.chkEnabled, "RIGHTCENTER", 4, 0)
		_configFrame.txtEnabled  = f
		
		local cxscreen = UIParent:GetWidth()
		local cyscreen = UIParent:GetHeight()
		
		f = myUI.Widgets.AnchorBox:new(_configFrame)
		f:SetPoint("TOPLEFT", _configFrame.chkEnabled, "BOTTOMLEFT", 160, border)
		f:SetScale(0.60)
		_configFrame.anchorBox = f	

		f = myUI.Widgets.AnchorBox:new(_configFrame)
		f:SetPoint("TOPLEFT", _configFrame.anchorBox, "TOPRIGHT", border, 0)
		f:SetScale(0.60)
		
		f:SetBackgroundColor(1,0,0,0.2)
		_configFrame.anchorBox2 = f	
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Anchor"])
		f:SetPoint("TOPLEFT", _configFrame.chkEnabled, "BOTTOMLEFT", 0,  border)
		_configFrame.lblAnchor = f
		
		f = myUI:CreateFrame("RiftCheckbox", _configFrame)
		f:SetPoint("TOPLEFT", _configFrame.lblAnchor, "BOTTOMLEFT", border, border)
		_configFrame.chkMouseover = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Follow Mouse"])
		f:SetPoint("LEFTCENTER", _configFrame.chkMouseover, "RIGHTCENTER")
		_configFrame.lblMouseover = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["X-Offset:"])
		f:SetPoint("TOPRIGHT", _configFrame.anchorBox, "BOTTOMLEFT", -border, border)
		_configFrame.lblX = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Y-Offset:"])		
		f:SetPoint("TOPRIGHT", _configFrame.lblX, "BOTTOMRIGHT", 0, 4)
		_configFrame.lblY = f
		
		local activebg = myUI.color:clone(black, 100)
		
		f = myUI:CreateFrame("RiftTextfield", _configFrame)
		f:SetWidth(60)
		f:SetBackgroundColor(activebg:rgba())
		f:SetPoint("LEFTCENTER", _configFrame.lblX, "RIGHTCENTER", border, 0)		
		_configFrame.txtX = f
		
		f = myUI:CreateFrame("RiftTextfield", _configFrame)
		f:SetWidth(60)
		f:SetBackgroundColor(activebg:rgba())
		f:SetPoint("LEFTCENTER", _configFrame.lblY, "RIGHTCENTER", border, 0)		
		_configFrame.txtY = f
		
		f = myUI:CreateFrame("Text", _configFrame)
		f.max = math.floor(cxscreen); f.min = -f.max
		f:Autosize("(" .. tostring(f.min) .. " <-> " .. tostring(f.max) .. ")")
		
		f:SetPoint("LEFTCENTER", _configFrame.txtX, "RIGHTCENTER", border, 0)
		_configFrame.lblX2 = f
		
		f = myUI:CreateFrame("Text", _configFrame)
		f.max = math.floor(cyscreen); f.min = -f.max
		f:Autosize("(" .. tostring(f.min) .. " <-> " .. tostring(f.max) .. ")")
		
		f:SetPoint("LEFTCENTER", _configFrame.txtY, "RIGHTCENTER", border, 0)
		_configFrame.lblY2 = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["CX:"])
		f:SetPoint("TOPRIGHT", _configFrame.lblY, "BOTTOMRIGHT", 0, 4)
		_configFrame.lblCX = f
		
		f = myUI:CreateFrame("RiftTextfield", _configFrame)
		f:SetWidth(60)
		f:SetBackgroundColor(activebg:rgba())
		f:SetPoint("LEFTCENTER", _configFrame.lblCX, "RIGHTCENTER", border, 0)
		_configFrame.txtCX = f
		
		f = myUI:CreateFrame("Text", _configFrame)
		f.max = 400; f.min = 48
		f:Autosize("(" .. tostring(f.min) .. " <-> " .. tostring(f.max) .. ")")
		f:SetPoint("LEFTCENTER", _configFrame.txtCX, "RIGHTCENTER", border, 0)
		_configFrame.lblCX2 = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["CY:"])
		f:SetPoint("TOPRIGHT", _configFrame.lblCX, "BOTTOMRIGHT", 0, 4)
		_configFrame.lblCY = f
		
		f = myUI:CreateFrame("RiftTextfield", _configFrame)
		f:SetWidth(60)
		f:SetBackgroundColor(activebg:rgba())
		f:SetPoint("LEFTCENTER", _configFrame.lblCY, "RIGHTCENTER", border, 0)
		_configFrame.txtCY = f
		
		f = myUI:CreateFrame("Text", _configFrame)		
		f.max = 400; f.min = 24
		f:Autosize("(" .. tostring(f.min) .. " <-> " .. tostring(f.max) .. ")")
		f:SetPoint("LEFTCENTER", _configFrame.txtCY, "RIGHTCENTER", border, 0)
		_configFrame.lblCY2 = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Spacing"])		
		f:SetPoint("TOPRIGHT", _configFrame.lblCY, "BOTTOMRIGHT", 0, 4)
		_configFrame.lblSpacing = f
		
		f = myUI:CreateFrame("RiftTextfield", _configFrame)
		f:SetBackgroundColor(activebg:rgba())
		f:SetWidth(30)
		f:SetPoint("LEFTCENTER", _configFrame.lblSpacing, "RIGHTCENTER", border, 0)
		_configFrame.txtColspacing = f
		
		f = myUI:CreateFrame("RiftTextfield", _configFrame)
		f:SetBackgroundColor(activebg:rgba())
		f:SetWidth(30)
		f:SetPoint("LEFTCENTER", _configFrame.txtColspacing, "RIGHTCENTER", 2, 0)
		_configFrame.txtRowspacing = f
		
		f = myUI:CreateFrame("Text", _configFrame)
		f.min = 0; f.max = 48
		f:Autosize("(" .. tostring(f.min) .. " <-> " .. tostring(f.max) .. ")")
		f:SetPoint("LEFTCENTER", _configFrame.txtRowspacing, "RIGHTCENTER", border, 0)
		_configFrame.lblSpacing2 = f
		
		f = myUI:CreateFrame("Text", _configFrame); f:Autosize(L["Show:"])
		f:SetPoint("TOPLEFT", _configFrame.tabpage, "TOPLEFT", border, 280)
		
		_configFrame.lblShow = f
		
		
		labels = {
			{"Health Values", "Health %", "Titlebar"},
			{"Power Values", "Power %", "Portrait"},
			{"Castbar", "Level", "Role"},
			{"PvP", "Raid Mark", "Ready"},
			{"Loot", "Planar", "Vitality"},
			{"Charge", "Buff Warning", "Left"},
			{"Right", "Top", "Bottom"}
		}
		
		local rowspacing = 120		
		local lastanchor = _configFrame.lblShow
		local lastrow = _configFrame.lblShow
		local startcx = border * 2
		local startcy = 2
		local lbl, chk
		local chkboxes = {}
		
		for _, row in ipairs(labels) do
			for i, col in ipairs(row) do
				chk = myUI:CreateFrame("RiftCheckbox", _configFrame)
				table.insert(chkboxes, chk)
				if (i == 1) then
					chk:SetPoint("TOPLEFT", lastrow, "BOTTOMLEFT", startcx, startcy)
					lastrow = chk
				else
					chk:SetPoint("LEFTCENTER", lastanchor, "RIGHTCENTER", rowspacing, 0)
				end
				lastanchor = chk
				
				lbl = myUI:CreateFrame("Text", chk)
				lbl:Autosize(L[col])
				lbl:SetPoint("LEFTCENTER", chk, "RIGHTCENTER")
				
				chk.label = lbl
			end
			
			startcx = 0; startcy = 4
		end
		
		_configFrame.chkHValues = chkboxes[1]
		_configFrame.chkHPercent = chkboxes[2]
		_configFrame.chkTitlebar = chkboxes[3]
		_configFrame.chkPValues = chkboxes[4]
		_configFrame.chkPPercent = chkboxes[5]
		_configFrame.chkPortrait = chkboxes[6]		
		_configFrame.chkCastbar = chkboxes[7]
		_configFrame.chkLevel = chkboxes[8]
		_configFrame.chkRole = chkboxes[9]
		_configFrame.chkPvP = chkboxes[10]
		_configFrame.chkMark = chkboxes[11]
		_configFrame.chkReady = chkboxes[12]		
		_configFrame.chkLoot = chkboxes[13]
		_configFrame.chkPlanar = chkboxes[14]		
		_configFrame.chkVitality = chkboxes[15]				
		_configFrame.chkCharge = chkboxes[16]		
		_configFrame.chkBuffWarn = chkboxes[17]
		_configFrame.chkWarnLeft = chkboxes[18]
		_configFrame.chkWarnRight = chkboxes[19]
		_configFrame.chkWarnTop = chkboxes[20]
		_configFrame.chkWarnBottom = chkboxes[21]
		
		-- loot and charge are mutually exclusive so
		_configFrame.chkCharge:ClearAll()
		_configFrame.chkCharge:SetPoint("CENTER", _configFrame.chkLoot, "CENTER")
		 
		-- put buff warning under charge
		_configFrame.chkBuffWarn:ClearAll()
		_configFrame.chkBuffWarn:SetPoint("TOPLEFT", _configFrame.chkLoot, "BOTTOMLEFT", 0, 18)
		 
		-- special for buff warning alignment
		_configFrame.chkWarnLeft:ClearAll()
		_configFrame.chkWarnLeft:SetPoint("TOPLEFT", _configFrame.chkBuffWarn, "BOTTOMLEFT", border, startcy)
		
		_configFrame.chkWarnRight:ClearAll()
		_configFrame.chkWarnRight:SetPoint("LEFTCENTER", _configFrame.chkWarnLeft, "RIGHTCENTER", rowspacing, 0)
		
		_configFrame.chkWarnTop:ClearAll()
		_configFrame.chkWarnTop:SetPoint("TOPLEFT", _configFrame.chkWarnLeft, "BOTTOMLEFT", 0, startcy)
		
		_configFrame.chkWarnBottom:ClearAll()
		_configFrame.chkWarnBottom:SetPoint("LEFTCENTER", _configFrame.chkWarnTop, "RIGHTCENTER", rowspacing, 0)
		
		_configFrame.chkBuffWarn.aligns = {
			_configFrame.chkWarnLeft, 
			_configFrame.chkWarnRight, 
			_configFrame.chkWarnTop, 
			_configFrame.chkWarnBottom
		}
		
		local function _warnAlignChecked(chk)		
			if (chk:GetChecked()) then
				local chkname = chk:GetName()
				for _, align in ipairs(_configFrame.chkBuffWarn.aligns) do
					if (chkname ~= align:GetName()) then
						align:SetChecked(false)
					end
				end
			end
		end
		 
		for _, align in ipairs(_configFrame.chkBuffWarn.aligns) do
			align.Event.CheckboxChange = _warnAlignChecked
		end
		 
		f = myUI:CreateFrame("RiftButton", _configFrame)
		f:SetText(L["Apply"])
		f:SetPoint("BOTTOMLEFT", _configFrame.tabpage, "BOTTOMLEFT", border, -2)
		_configFrame.btnApply = f
		
		end -- if (true)
	
	
		_configFrame.LoadUnit = function(self, unit)
			self._loading = true
			self._currentUnit = unit
			
			local uf = me.Frames[unit]
			local enabled = ((uf ~= nil) and (uf.Definition.enabled))
			
			local txts = {self.txtX, self.txtY, self.txtCX, self.txtCY, self.txtColspacing, self.txtRowspacing}
			local chks = {self.chkHValues, self.chkHPercent, self.chkPValues, self.chkPPercent, 
				self.chkPortrait, self.chkTitlebar, self.chkCastBar, self.chkLevel, self.chkRole, 
				self.chkPvP, self.chkPlanar, self.chkVitality, self.chkMark,
				self.chkReady, self.chkBuffWarn, self.chkCharge, self.chkLoot
			}
			
			local isplayer = (unit == "player")
			local isplayerpet = (unit == "player.pet")
			local isgroup = (unit == "group01")
			local isgrouppet = (unit == "group01.pet")
			local ismouseover = (unit == "mouseover")
			local ismouseovertarget = (unit == "mouseover.target")
			
			for _, txt in ipairs(txts) do txt:SetVisible(enabled) end
			for _, chk in ipairs(chks) do chk:SetEnabled(enabled) end			
			if (not enabled) then
				for _, txt in ipairs(txts) do txt:SetText("") end
				for _, chk in ipairs(chks) do chk:SetChecked(false) end
			end

			self.anchorBox:SetEnabled(enabled)
			self.anchorBox2:SetEnabled(enabled)
			self.chkEnabled:SetChecked(enabled)
			self.btnApply:SetEnabled(enabled)
		
			
			
			self.chkBuffWarn:SetVisible(isplayer)
			self.chkWarnLeft:SetVisible(isplayer)
			self.chkWarnRight:SetVisible(isplayer)
			self.chkWarnTop:SetVisible(isplayer)
			self.chkWarnBottom:SetVisible(isplayer)
			self.chkCharge:SetVisible(isplayer)
			self.chkPlanar:SetVisible(isplayer)
			self.chkVitality:SetVisible(isplayer)
			self.txtColspacing:SetVisible((isgroup or isgrouppet) and enabled)
			self.txtRowspacing:SetVisible((isgroup or isgrouppet) and enabled)
			self.lblSpacing:SetVisible(isgroup or isgrouppet)
			self.lblSpacing2:SetVisible(isgroup or isgrouppet)
			self.chkLoot:SetVisible(not (isplayer or isplayerpet or isgroup or isgrouppet))
			self.txtX:SetVisible(enabled) --(not (ismouseover or ismouseovertarget)) and enabled)
			self.lblX:SetVisible(true) --not (ismouseover or ismouseovertarget)) 
			self.lblX2:SetVisible(true) --not (ismouseover or ismouseovertarget))
			self.txtY:SetVisible(enabled) --(not (ismouseover or ismouseovertarget)) and enabled)
			self.lblY:SetVisible(true) -- not (ismouseover or ismouseovertarget))
			self.lblY2:SetVisible(true) --not (ismouseover or ismouseovertarget))
			
			self.chkMouseover:SetVisible(ismouseover or ismouseovertarget)
			self.lblMouseover:SetVisible(ismouseover or ismouseovertarget)
			
			if (enabled) then
				local def = uf.Definition
			
				if (isplayer) then
					self.chkBuffWarn:SetChecked(notnil(def.longBuffs) and (def.longBuffs.enabled))
					self.chkWarnLeft:SetChecked(notnil(def.longBuffs) and (def.longBuffs.anchor == "LEFT"))
					self.chkWarnRight:SetChecked(notnil(def.longBuffs) and (def.longBuffs.anchor == "RIGHT"))
					self.chkWarnTop:SetChecked(notnil(def.longBuffs) and (def.longBuffs.anchor == "TOP"))
					self.chkWarnBottom:SetChecked(notnil(def.longBuffs) and (def.longBuffs.anchor == "BOTTOM"))
					
					self.chkCharge:SetChecked(def.showCharge)
					self.chkPlanar:SetChecked(def.showPlanar)
					self.chkVitality:SetChecked(def.showVitality)
				end
			
				if (isgroup or isgrouppet) then					
					self.txtRowspacing:SetText(tostring(def.rowspacing))
					self.txtColspacing:SetText(tostring(def.colspacing))
				end
				if (not(isplayer or isplayerpet or isgroup or isgrouppet)) then
					self.chkLoot:SetChecked(def.showLoot)
				end
				--if (not (ismouseover or ismouseovertarget)) then
					self.txtX:SetText(tostring(math.floor(def.x)))
					self.txtY:SetText(tostring(math.floor(def.y)))
				--end
				
				if (ismouseover or ismouseovertarget) then
					self.chkMouseover:SetChecked(def.followMouse)
				end
				
				self.txtCX:SetText(tostring(math.floor(def.cx)))
				self.txtCY:SetText(tostring(math.floor(def.cy)))
				
				self.chkPvP:SetChecked(def.showPvP)
				self.chkLevel:SetChecked(def.showLevel)
				self.chkPortrait:SetChecked(def.showPortrait)
				self.chkHValues:SetChecked(def.showHealthText)
				self.chkHPercent:SetChecked(def.showHealthAsPercent)
				self.chkPValues:SetChecked(def.showPowerText)
				self.chkPPercent:SetChecked(def.showPowerAsPercent)
				self.chkTitlebar:SetChecked(def.showTitleBar)
				self.chkCastbar:SetChecked(def.showCastBar)
				self.chkRole:SetChecked(def.showRole)
				self.chkReady:SetChecked(def.showReady)
				self.chkMark:SetChecked(def.showMark)
				
				self.anchorBox:SetAnchor(def.anchorTo)
				self.anchorBox2:SetAnchor(def.anchorFrom)
			
			else
				self.anchorBox:SetAnchor("TOPLEFT")
				self.anchorBox2:SetAnchor("TOPLEFT")
				
					
			end
			
			self._loading = false
		end
		
		_configFrame.btnApply.Event.LeftClick = function(btn)		
			local def = UnitFrames.FrameDefs[_configFrame._currentUnit]
			local raid = string.find(_configFrame._currentUnit, "group")
			local mouseover = string.find(_configFrame._currentUnit, "mouseover")
			local player = (_configFrame._currentUnit == "player")
			local playerpet = (_configFrame._currentUnit ==  "player.pet")
			local raidpet = (raid) and (string.find(_configFrame._currentUnit, "pet"))
			
			local x, y, cx, cy, rowspacing, colspacing
			
			if (pcall(function()
				--if (not mouseover) then
					x = 0+tonumber(_configFrame.txtX:GetText())
					y = 0+tonumber(_configFrame.txtY:GetText())
				--end
				
				cx = 0+tonumber(_configFrame.txtCX:GetText())
				cy = 0+tonumber(_configFrame.txtCY:GetText())
				
				
				if (raid) then
					colspacing = 0+tonumber(_configFrame.txtColspacing:GetText())
					rowspacing = 0+tonumber(_configFrame.txtRowspacing:GetText())				
				end
				
				end)) then
				
				--if (not mouseover) then
					if (x < _configFrame.lblX2.min) or (x > _configFrame.lblX2.max) then
						myUI.msgbox(_configFrame, L["Bad X"], _, "OK", "WARNING")
						return
					end
					if (y < _configFrame.lblY2.min) or (y > _configFrame.lblY2.max) then
						myUI.msgbox(_configFrame, L["Bad Y"], _, "OK", "WARNING")
						return
					end
				--end
				if (cx < _configFrame.lblCX2.min) or (cx > _configFrame.lblCX2.max) then
					myUI.msgbox(_configFrame, L["Bad CX"], _, "OK", "WARNING")
					return 
				end
				if (cy < _configFrame.lblCY2.min) or (cy > _configFrame.lblCY2.max) then
					myUI.msgbox(_configFrame, L["Bad CY"], _, "OK", "WARNING")
					return 
				end
				
				if (raid) then
					if (colspacing < _configFrame.lblSpacing2.min) or (colspacing > _configFrame.lblSpacing2.max) then
						myUI.msgbox(_configFrame, L["Bad Col Spacing"], _, "OK", "WARNING")
						return
					end
					if (rowspacing < _configFrame.lblSpacing2.min) or (rowspacing > _configFrame.lblSpacing2.max) then
						myUI.msgbox(_configFrame, L["Bad Row Spacing"], _, "OK", "WARNING")
						return
					end
				end			
				
				
				def.x = x
				def.y = y
				def.cx = cx
				def.cy = cy
				
				if (raid) then
					def.colspacing = colspacing
					def.rowspacing = rowspacing
				end
				if (not mouseover) then
					def.anchorTo = _configFrame.anchorBox:GetAnchor()
					def.anchorFrom = _configFrame.anchorBox2:GetAnchor()
				end
				if (mouseover) then
					def.followMouse = _configFrame.chkMouseover:GetChecked()
				end
				
				def.showHealthText = _configFrame.chkHValues:GetChecked()
				def.showHealthAsPercent = _configFrame.chkHPercent:GetChecked()
				def.showPowerText = _configFrame.chkPValues:GetChecked()
				def.showPowerAsPercent = _configFrame.chkPPercent:GetChecked()
				def.showLevel = _configFrame.chkLevel:GetChecked()
				def.showTitleBar = _configFrame.chkTitlebar:GetChecked()
				def.showPortrait = _configFrame.chkPortrait:GetChecked()
				def.showPvP = _configFrame.chkPvP:GetChecked()
				def.showRole = _configFrame.chkRole:GetChecked()
				def.showReady = _configFrame.chkReady:GetChecked()
				def.showMark = _configFrame.chkMark:GetChecked()
				
				
				
				if (player) then
					def.showVitality = _configFrame.chkVitality:GetChecked()
					def.showPlanar = _configFrame.chkPlanar:GetChecked()
					def.showCharge = _configFrame.chkCharge:GetChecked()
					
					if (_configFrame.chkBuffWarn:GetChecked()) then
						if (def.longBuffs) then
							def.longBuffs.enabled = true
							
							if (_configFrame.chkWarnLeft:GetChecked()) then def.longBuffs.anchor = "LEFT"
							elseif (_configFrame.chkWarnRight:GetChecked()) then def.longBuffs.anchor = "RIGHT"
							elseif (_configFrame.chkWarnTop:GetChecked()) then def.longBuffs.anchor = "TOP"
							elseif (_configFrame.chkWarnBottom:GetChecked()) then def.longBuffs.anchor = "BOTTOM"
							end
						else
							def.longBuffs = {enabled = true, cyxExpired = 36, cxyNormal = 24
							}
							if (_configFrame.chkWarnLeft:GetChecked()) then def.longBuffs.anchor = "LEFT"
							elseif (_configFrame.chkWarnRight:GetChecked()) then def.longBuffs.anchor = "RIGHT"
							elseif (_configFrame.chkWarnTop:GetChecked()) then def.longBuffs.anchor = "TOP"
							elseif (_configFrame.chkWarnBottom:GetChecked()) then def.longBuffs.anchor = "BOTTOM"
							else
								def.longBuffss.anchor = "TOP"
							end
						end
					else
						if (def.longBuffs) then
							def.longBuffs.enabled = false
						end
					end
				end
				if not (player or playerpet or raid or raidpet) then
					def.showLoot = _configFrame.chkLoot:GetChecked()
				end
				
				UnitFrames:ResetUnit(_configFrame._currentUnit)
			else
				myUI.msgbox(_configFrame, L["Invalid value"], _, "OK", "WARNING")
				return
			end
			
		
			
		end
		
		_configFrame.chkEnabled.Event.CheckboxChange = function(chk)
			if (_configFrame._loading) then return end
			
			UnitFrames:SetUnitEnabled(_configFrame._currentUnit, chk:GetChecked())
			_configFrame:LoadUnit(_configFrame._currentUnit)		
		end
		
		
		
		_configFrame:LoadUnit("player")
	end
	
	return _configFrame
end

function UnitFrames:OnUpdate(now, elapsed)
	
	if (_simulating) then		
		__updateSimCastDetails(now, elapsed)
		for _, frame in pairs(self.Frames) do
			
			if (frame.Definition.showCastBar) then 
				
				frame.CastBarBg:SetVisible(true)
			end
		end
	end
	
	for _, frame in pairs(self.Frames) do
		if (frame.CastBarBg) and (frame.CastBarBg:GetVisible()) then
			frame:UpdateCastBar(true, now);
		end		
	end
	local pframe = self.Frames["player"]
	if (pframe) then
		__fadeCombat(pframe, elapsed)
		for _, box in pairs(pframe.LongBuffBoxes) do
			box:UpdateBlink()
		end
	end
end


function UnitFrames:__onAdaptiveUpdatingChanged(enabled)
	adaptiveUpdating = enabled
end

function UnitFrames:OnUnitUpdate(now, elapsed)
	if (adaptiveUpdating) then
		local n = 0
		local maxupdate = math.max(math.ceil(#q / 2), 14) -- anything under #q/2 creates problems due to not penetrating deeply enough into the queue on each update
		local nq = {}
		for _, unitid in ipairs(q) do
			if (n < maxupdate) then
				if (UnitFrames.Frames[unitid]:Update()) then				
					n = n + 1
					table.insert(nq, unitid) -- put on the bottom of the next update queue
				else
					table.insert(nq, 1, unitid)
				end					
			else		
				table.insert(nq, 1, unitid) -- put on the top of the queue for the next update
			end
		end
		
		q = nq
	else
		for _, frame in pairs(UnitFrames.Frames) do
			frame:Update()
		end
	end
end

function UnitFrames.OnCastbar(units)
	if (units) then
		local uid;
		local now = Inspect.Time.Frame()
		for _, frame in pairs(UnitFrames.Frames) do			
			if (units[Inspect.Unit.Lookup(frame.UnitID)]) then
				frame:UpdateCastBar(true, now)
			end
		end
		
		
	end
end





--[[
function UnitFrames:OnBuff(unitID, buffs)
	-- We need to do a general update here because some buffs that effect mana/health don't
	-- raise a Unit.Detail event.

	if (unitID) then
		local unit = Inspect.Unit.Lookup(unitID);
		
		if (unit) then
			if (UnitFrames.Frames[unit]) then
				UnitFrames.Frames[unit]:Update()
			end
		end		
	end
end
]]--


--[[
function UnitFrames.OnUnitDetail(units)
	for _, frame in pairs(UnitFrames.Frames) do 
		frame:Update()
	end
end
]]--
function UnitFrames:OnUnitAvailable(units)
	if (units ~= nil) then
		for _, frame in pairs(UnitFrames.Frames) do
			if (units[Inspect.Unit.Lookup(frame.UnitID)]) then				
				frame._lastDetail = {}
			end
			frame:Update()
		end		
	end
end

function UnitFrames:ToggleMovers()
	for _, uf in pairs(self.Frames) do
		uf:ToggleMover()
	end
end
function UnitFrames:ShowMovers()
	for _, uf in pairs(self.Frames) do
		uf:ShowMover()		
	end
end
function UnitFrames:HideMovers()
	for _, uf in pairs(self.Frames) do
		uf:HideMover()
	end
end


function UnitFrames:OnCombat(inCombat)
	if (self.Frames["player"]) then
		__setInCombat(self.Frames["player"], inCombat)
	end
end





function UnitFrames:BeginSimulation()
	_simulating = true
end

function UnitFrames:EndSimulation()
	_simulating = false
end

function UnitFrames:GetSimulating()
	return _simulating
end

local function __updateCastbar(self, visible, now)
	if (self._layout) then return end
	
	if (self.Definition.showCastBar) then
		if (visible or self._mover or _simulating) then		
			local castdetail
			
			if (_simulating) then
				castdetail = simcdetail
			elseif (self._mover) then
				castdetail = testcdetail	
			else
				castdetail = Inspect.Unit.Castbar(self.UnitID); 
			end
			
			if (castdetail) and (castdetail.remaining > 0) and (castdetail.duration > 0) then
				
				local cx = self.CastBarBg:GetWidth() - (2 * UnitFrames.Border);
				local cast 
				
				if (castdetail.begin and now) then					
					cast = (castdetail.begin + castdetail.duration - now) / castdetail.duration
				else
					cast = castdetail.remaining / castdetail.duration;
				end
				
				cast = 1 - cast
				
				if (castdetail.channeled) then cast = 1 - cast; end
				if (castdetail.uninterruptible) then
					self.CastBarBg:SetBackgroundColor(red:rgba());
				else
					self.CastBarBg:SetBackgroundColor(UnitFrames.bgNormal:rgba());
				end
				
				local cxCast = cx 
				cxCast = minmax(cxCast * cast, 0, cx);
				
				self.CastBar:SetWidth(cxCast);
				if (castdetail.abilityName) then
					self.CastBar.Text:SetText(castdetail.abilityName);
				end
			else
				visible = false	
				self.CastBar:SetWidth(0)
			end	
		else
			visible = false			
			self.CastBar:SetWidth(0)
		end
		
		self.CastBarBg:SetVisible(ifnil(visible, false));
	end
end
local function __resetUF(self)		
	self._lastDetail = {}
	self:Layout(true)
	self:Update()
	self:UpdateCastBar(iif(self._mover, true, false), Inspect.Time.Frame());
end
local function __update(self, detail)
	if (self._layout) then 
		return 0
	elseif (not self.Definition.enabled) and (not self._mover) then 
		self:SetVisible(false); return 0
	elseif (self._mover) then 
		detail = testdetail
	elseif (_simulating) then 
		detail = __updateSimDetails(); 
		Inspect.Unit.Detail(self.UnitID); -- the inspect is to simulate that overhead as well	
	elseif (detail == nil) then 
		detail = Inspect.Unit.Detail(self.UnitID) 
	end
	
	local r = 0
	
	if (detail) then		
		-- check to see if anything has actually changed
		if (self._lastDetail == nil) then self._lastDetail = {} end
		local dirty = false
		for k, v in pairs(detail) do
			if (v ~= self._lastDetail[k]) then
				dirty = true
				break
			end
		end
		local def = self.Definition
		
		
		
		if (dirty) then			
			local cx = self:GetWidth() -- our gauge width excluding portrait width if appropriate
			if (def.showPortrait) then cx = cx - self:GetHeight() end
			
			-- do name and level
			local name = ifnil(detail.name, "??")
			if (name == "#") then name = self.UnitID end
			
			if (not def.showPortrait) and (def.showLevel) then
				name = tostring(ifnil(detail.level, "??")) .. " - " .. name
			end
			self.UnitName:SetText(name); self.UnitName:SetHeight(self.UnitName:GetFullHeight())
			
			if (self.UnitName2) then self.UnitName2:SetText(name); self.UnitName2:SetHeight(self.UnitName2:GetFullHeight()) end
			
			if (def.showPortrait) and (def.showLevel) then
				self.Level:Autosize(tostring(ifnil(detail.level, "??")))
			end
		
			-- do relation
			local clrRel = iif(detail.relation == "hostile", red, iif(detail.relation == "friendly", iif(detail.pvp, lime, dodgerBlue), yellow))	
		
			-- online status + health		
			local astatus = 0.25 --AFK/Offline/Unknown
			local szhealth = " "
			local cxhealth = 0
			local phealth = -1 -- health percentage. = 0 denotes afk/disconnected/offline/dead
			if (detail.offline) then 
				szhealth = L["Offline"]
			elseif (detail.afk) then 
				szhealth = L["AFK"]
			elseif (detail.warfront) then
				szhealth = L["Warfront"]
			elseif (isnil(detail.healthMax) or isnil(detail.health) or (detail.healthMax == 0)) then
				szhealth = L["Unknown"]
			else
				astatus = 1
				if (detail.health == 0) then
					szhealth = L["Dead"]
					cxhealth = 0; phealth = 0
				else
					phealth = detail.health / detail.healthMax
					cxhealth = (minmax(phealth * cx, 0, cx))
					if (def.showHealthText) then				
						szhealth = tostring(detail.health) .. " / " .. tostring(detail.healthMax)
					end
					if (def.showHealthAsPercent) then
						if (def.showHealthText) then
							szhealth = szhealth .. "    " ..  string.format("%02d", phealth * 100) .. "%"
						else				
							szhealth = string.format("%02d", phealth * 100) .. "%"
						end
					end
				end
			end

			
			self.DeadIcon:SetVisible(phealth == 0)
			
			local calling = detail.calling
			if (calling ~= self._lastDetail.calling) then							
				local clrPowerText = white
				local clrCalling = white			
				local callingIcon = 1
				local powerTexture = "images/bg2.jpg"
				
				if (calling == "cleric") then 
					powerTexture = "images/mana.jpg"; clrCalling = lime; callingIcon = 2
				elseif (calling == "mage") then
					powerTexture = "images/mana.jpg"; clrCalling = dodgerBlue; callingIcon = 3
				elseif (calling == "rogue") then
					powerTexture = "images/energy.jpg"; clrCalling = yellow; callingIcon = 4
				elseif (calling == "warrior") then
					powerTexture = "images/power.jpg"; clrCalling = red; callingIcon = 5
					clrPowerText = black
				end
				
				self.UnitPower:SetTexture("myUI", powerTexture)
				self.UnitName:SetFontColor(clrCalling:rgba())
				
				if ((def.showPowerText) or (def.showPowerAsPercent) and (self.UnitPower.Text))then
					self.UnitPower.Text:SetFontColor(clrPowerText:rgba())
				end
				
				if (def.showPortrait) then
					for i, icon in ipairs(self.Icons) do					
						icon:SetVisible(i == callingIcon)
					end
				end			
							
			end
			
			-- power/energy/mana
			local szpower = ""
			local ppower = -1
			local ismana = false
			local cxpower = 0
			
			if (phealth > 0) then
				local pmin, pmax
				if (detail.mana) then			
					pmin = detail.mana; pmax = detail.manaMax; ismana = true
				elseif (detail.energy) then
					pmin = detail.energy; pmax = detail.energyMax
				elseif (detail.power) then
					pmin = detail.power; pmax = 100			
				end
				
				if (pmax) and (pmax > 0) then
					ppower = pmin / pmax
					
					if (def.showPowerText) then
						szpower = tostring(pmin) .. " / " .. tostring(pmax)
					end
					if (def.showPowerAsPercent) then
						if (def.showPowerText) then
							szpower = szpower .. "    " .. string.format("%02d", ppower * 100) .. "%"
						else
							szpower = string.format("%02d", ppower * 100) .. "%"
						end
					end
					
					cxpower = minmax(cx * ppower, 0, cx)
				end		
			end
			
			if (def.showPortrait) then self.IconBg:SetBackgroundColor(clrRel:rgba()) end
			if (def.showLevel) and (self.Level) then self.Level:SetFontColor(clrRel:rgba()) end
			
			self.UnitHealth.Text:Autosize(szhealth)
			self.UnitHealth:SetWidth(cxhealth)
			self.UnitHealth:SetAlpha(astatus)
			self.UnitPower:SetAlpha(astatus)
			
			if (ppower == -1) then
				self.UnitPower:SetWidth(cx) -- no power for this unit so fill with texture
			else
				self.UnitPower:SetWidth(cxpower)
			end
			
			if (self.UnitPower.Text) and (def.showPowerText or def.showPowerAsPercent) then
				self.UnitPower.Text:Autosize(szpower)
			end

			-- charge
			if (def.showCharge and detail.charge and (self.UnitID == "player")) then
				if (detail.charge > 0) then
					self.Charge:SetWidth(minmax((detail.charge/100) * cx, 1, cx))
					self.ChargeBg:SetHeight(8)				
				else
					self.ChargeBg:SetHeight(0)
				end
			elseif (self.ChargeBg) then
				self.ChargeBg:SetHeight(0)
			end
			
			-- role
			local showrole = false
			
			if (self.Role and def.showRole) then
				if (detail.role ~= self._lastDetail.role) then
					if (detail.role == "dps") then
						self.Role:SetTexture("myUI", "images/dps.png")
						showrole = true
					elseif (detail.role == "tank") then
						self.Role:SetTexture("myUI", "images/tank.png")
						showrole = true
					elseif (detail.role == "heal") then
						self.Role:SetTexture("myUI", "images/healer.png")
						showrole = true
					elseif (detail.role == "support") then
						self.Role:SetTexture("myUI", "images/support.png")
						showrole = true
					end
				end
				-- This is messy. this really should be addressed in the layout
				-- method, but that method doesn't know if a role exists to show.
				-- The name can't be pinned to the role either b/c that screws
				-- up vertical positioning on the name
				
				local offx = 2
				if (showrole) then offx = offx + 2 + self.Role.defaultSize end
				
						
				if (def.showTitleBar) then
					self.UnitName:SetPoint("LEFTCENTER", self.TitleBar, "LEFTCENTER", offx, 0)
				else
					if (def.showPortrait) then offx = offx + def.cy end
					self.UnitName:SetPoint("LEFTCENTER", self, "LEFTCENTER", offx, 0)				
				end
			
				self.Role:SetVisible(showrole)
			elseif (self.Role) then
				self.Role:SetVisible(false)
			end
			
			-- warnings
			local apower = 1
			local ahealth = 1
			if (phealth > 0) and (phealth < UnitFrames.lowHealth) then
				self.Border:SetBackgroundColor(UnitFrames.bgLowHealth:rgba())
				ahealth = 0.4
				self.Border:SetAlpha(1)
			elseif (ismana) and (ppower >= 0) and (ppower < UnitFrames.lowMana) then
				self.Border:SetBackgroundColor(UnitFrames.bgLowMana:rgba())
				apower = 0.6
				self.Border:SetAlpha(1)
			else
				self.Border:SetBackgroundColor(black:rgba())
				self.Border:SetAlpha(0.4)
				self.UnitPowerBg:SetAlpha(1)
			end
			
			self.UnitPowerBg:SetAlpha(apower)
			self.UnitHealthBg:SetAlpha(ahealth)
		
			-- planar charge
			if (self.Planar and def.showPlanar) then
				if (detail.planarMax) then
					self.Planar.Text:Autosize(tostring(ifnil(detail.planar,0)) .. "/" .. tostring(detail.planarMax))
					self.Planar:SetVisible(true)
				else
					self.Planar:SetVisible(false)
				end		
			elseif (self.Planar) then
				self.Planar:SetVisible(false)
			end
		
			-- ready check
			if (def.showReady and notnil(detail.ready)) then		
				self.ReadyCheck:SetVisible(detail.ready)
				self.NotReadyCheck:SetVisible(not detail.ready)
			elseif (self.ReadyCheck) then
				self.ReadyCheck:SetVisible(false)
				self.NotReadyCheck:SetVisible(false)
			end
			
			-- ready mark
			if (def.showMark and detail.mark) then
				self.Mark:Autosize(tostring(detail.mark))
				self.Mark:SetVisible(true)
			elseif (self.Mark) then
				self.Mark:SetVisible(false)
			end
			-- various indicators
			local pvp = false
			local vit = false
			
			if (def.showPvP and detail.pvp) then
				pvp = true			
			elseif (self.PvP) then
				self.PvP:SetVisible(false)
			end
		
			if (def.showVitality and detail.vitality and detail.vitality < 100 and (self.UnitID == "player")) then
				vit = true		
			elseif (self.Vitality) then
				self.Vitality:SetVisible(false)
			end
		
			if (self._lastDetail.pvp ~= detail.pvp) or (self._lastDetail.vitality ~= detail.vitality) then		
				if (pvp) then
					self.PvP:ClearAll()
					self.PvP:SetSize(self.PvP.defaultSize, self.PvP.defaultSize)
					self.PvP:SetPoint("RIGHTCENTER", self, "TOPRIGHT", -4, 6) 
					self.PvP:SetVisible(true)
					if (vit) then
						self.Vitality:ClearAll()
						self.Vitality:SetSize(self.Vitality.defaultSize, self.Vitality.defaultSize)
						self.Vitality:SetPoint("RIGHTCENTER", self.PvP, "LEFTCENTER", -4, 0)
						self.Vitality:SetVisible(true)
					end
				elseif (vit) then
					self.Vitality:ClearAll()
					self.Vitality:SetSize(self.Vitality.defaultSize, self.Vitality.defaultSize)
					self.Vitality:SetPoint("RIGHTCENTER", self, "TOPRIGHT", -4, 6)
					self.Vitality:SetVisible(true)
				end
			end
			if (self.Loot and def.showLoot) then
				if (detail.guaranteedLoot) then
					self.Loot:SetVisible(true)
				else
					self.Loot:SetVisible(false)
				end
			end
			
			-- a little more layout stuff needed
			
			if (not self._mover) then
				if (self.UnitID == "mouseover") and def.followMouse then
					local ox = 48
					local m = Inspect.Mouse()
					if (m) then
						self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", m.x + ox, m.y - ox)
					end
				elseif (self.UnitID == "mouseover.target") and def.followMouse then
					if (UnitFrames.Frames["mouseover"]) then
						self:SetPoint("LEFTCENTER", UnitFrames.Frames["mouseover"], "RIGHTCENTER", 32, 0)
					else
						local m = Inspect.Mouse()
						if (m) then
							local ox = 48
							self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", m.x + ox, m.y - ox)
						end
					end		
				end
			end
			
			r = 1 -- something was updated		
		end
		-- purged/expired buff warning				
		
	
		if (def.longBuffs) and (def.longBuffs.enabled) and (self.UnitID == "player") then
			local buffs = Inspect.Buff.List(self.UnitID)
			
			-- check out long buffs
			local buffDetails = {}
			
			local id2 = Inspect.Unit.Lookup(self.UnitID);
			
			if (buffs) then					
				local lngbuff = UnitFrames.longBuffDuration;
				
				for bid, _ in pairs(buffs) do
					local bdetail = Inspect.Buff.Detail(self.UnitID, bid);
					
					if (bdetail) then
						-- see if this is applicable to our long-buff monitoring
						if (bdetail.caster) and ((bdetail.caster == self.UnitID) or (bdetail.caster == id2)) then
							if (ifnil(bdetail.duration,lngbuff) >= lngbuff) then
								if (not bdetail.curse) and (not bdetail.noncancelable) then										
									buffDetails[bdetail.name] = bdetail;
									if (self.LongBuffBoxes[bdetail.name] == nil) then
										-- create a buffbox for this one
										self.LongBuffBoxes[bdetail.name] = BuffBox:new(self, self.Definition.longBuffs, bdetail);
									end
								end
							end
						end						
					end
				end	

				-- handle updates for our long buff boxes
				
				local buffNames = {};
				for k, _ in pairs(self.LongBuffBoxes) do
					table.insert(buffNames, k);
				end
				
				table.sort(buffNames, function(a,b) return a < b end);
				
				
				local lastAnchor = self;
				local anchorFrom = "TOPLEFT";
				local anchorTo = "TOPRIGHT"
				local anchorFromNext = "TOPLEFT";
				local anchorToNext = "TOPRIGHT";
				
				local offx = 0;
				local offy = 0;
				local offxNext = 0;
				local offyNext =  0;
				
				local anchor = self.Definition.longBuffs.anchor;
				
				
				if (anchor == "LEFT") then
					anchorFrom = "TOPRIGHT"
					anchorTo = "TOPLEFT"
					
					anchorFromNext = anchorFrom
					anchorToNext = anchorTo
					
					offx = -2; offxNext = offx
					offy = 0; offyNext = offy
					
					lastAnchor = self.IconBg;
					
				elseif (anchor == "RIGHT") then
					anchorFrom = "TOPLEFT"
					anchorTo = "TOPRIGHT"
					
					anchorFromNext = anchorFrom
					anchorToNext = anchorTo
					
					offx = 2; offxNext = offx
					offy = 0; offyNext = offy
					
				elseif (anchor == "TOP") then
					anchorFrom = "BOTTOMLEFT"
					anchorTo = "TOPLEFT"
					
					anchorFromNext = "BOTTOMLEFT"
					anchorToNext = "BOTTOMRIGHT"
					
					offx = 0; offxNext = 2
					offy = -2; offyNext = 0
					
				elseif (anchor == "BOTTOM") then
					anchorFrom = "TOPLEFT"
					anchorTo = "BOTTOMLEFT"
					
					anchorFromNext = "TOPLEFT"
					anchorToNext = "TOPRIGHT"
					
					offx = 0; offxNext = 2
					offy = 2; offyNext = 0
				end
				
				for _, bname in ipairs(buffNames) do
					local box = self.LongBuffBoxes[bname];
					if (not box.Dismissed) then
						box:ClearAll()
						box:SetPoint(anchorFrom, lastAnchor, anchorTo, offx, offy);
						
						anchorFrom = anchorFromNext;
						anchorTo = anchorToNext;
						offx = offxNext;
						offy = offyNext;
					
						box:Update(buffDetails[box.BuffName]);
						lastAnchor = box;
					end
				end
				
			end
		else
			if (self.LongBuffBoxes) then
				for _, box in pairs(self.LongBuffBoxes) do
					box:SetVisible(false)
				end
			end
		end
		self._lastDetail = detail
		self:SetVisible(true)
		
		
	else
		self._lastDetail = {}
		self:SetVisible(false)
	end
	
	return r
end

local function __ensureFrames(self)
	local def = self.Definition
	
	if (isnil(self.CombatIndicator) and (self.UnitID == "player")) then self.CombatIndicator = myUI:CreateFrame("Frame", self); end
	
	
	if (def.showTitleBar and isnil(self.TitleBar)) then 
		self.TitleBar = myUI:CreateFrame("Texture", self); 
		self.TitleBar:SetTexture("myUI", "images/bg2.jpg");
		self.TitleBar:SetLayer(1)
	end
	if (isnil(self.UnitName)) then self.UnitName = myUI:CreateFrame("Text", self); end
	if isnil(self.UnitName2) then self.UnitName2 = myUI:CreateFrame("Text", self); self.UnitName2:SetFontColor(black:rgba()); end
	
	if (isnil(self.UnitHealth)) then 
		self.UnitHealth = myUI:CreateFrame("Texture", self); self.UnitHealth:SetLayer(2); 
		self.UnitHealth:SetTexture("myUI", "images/health.jpg");		
	end
	-- setup this one even if not showing health info, for AFK, dead, offline, etc
	if (isnil(self.UnitHealth.Text)) then 
		self.UnitHealth.Text = myUI:CreateFrame("Text", self); 
		self.UnitHealth.Text:SetLayer(10)
	end
	if (isnil(self.UnitHealthBg)) then self.UnitHealthBg = myUI:CreateFrame("Frame", self); end
	if (isnil(self.UnitPower)) then self.UnitPower = myUI:CreateFrame("Texture", self); self.UnitPower:SetLayer(2); end
	if (isnil(self.UnitPowerBg)) then self.UnitPowerBg = myUI:CreateFrame("Frame", self); end
	if (self.UnitID == "player") and def.showCharge and isnil(self.ChargeBg) then
		self.ChargeBg = myUI:CreateFrame("Frame", self)
		self.Charge = myUI:CreateFrame("Texture", self.ChargeBg)
		self.Charge:SetTexture("myUI", "images/cast.jpg");
		
		self.ChargeBg:SetLayer(5)
		
	end
	
	self.UnitPowerBg:SetBackgroundColor(black:rgba()); self.UnitPowerBg:SetLayer(1);
	self.UnitHealthBg:SetBackgroundColor(black:rgba()); self.UnitHealthBg:SetLayer(1);
	
	if (isnil(self.DeadIcon)) then
		self.DeadIcon = myUI:CreateFrame("Texture", self)
		self.DeadIcon:SetTexture("myUI", "images/dead.png")
		self.DeadIcon:SetVisible(false)
	end
	
	if (def.showPortrait) then
		if (isnil(self.IconBg)) then self.IconBg = myUI:CreateFrame("Frame", self); end
		
		if (isnil(self.Icons)) then
			self.Icons = {}
			for i=1, 5, 1 do
				table.insert(self.Icons, myUI:CreateFrame("Texture",self))
			end		
			
			self.IconBg:SetBackgroundColor(normal:rgba());
			self.IconBg:SetAlpha(0.60);
			
			self.Icons[1]:SetTexture("myUI", "images/unknown.jpg");
			self.Icons[2]:SetTexture("myUI", "images/cleric.jpg");
			self.Icons[3]:SetTexture("myUI", "images/mage.jpg");
			self.Icons[4]:SetTexture("myUI", "images/rogue.jpg");
			self.Icons[5]:SetTexture("myUI", "images/warrior.jpg");			
		end	
	end
	
	if (def.showPowerText or def.showPowerAsPercent) then
		if (isnil(self.UnitPower.Text)) then 
			self.UnitPower.Text = myUI:CreateFrame("Text", self); 
			self.UnitPower.Text:SetLayer(10); 
		end
	end
	
	if (def.showCastBar) then
		if (isnil(self.CastBarBg)) then self.CastBarBg = myUI:CreateFrame("Frame", self); end
		if (isnil(self.CastBar)) then self.CastBar = myUI:CreateFrame("Texture", self.CastBarBg); end
		if (isnil(self.CastBar.Text)) then self.CastBar.Text = myUI:CreateFrame("Text", self.CastBarBg); end
		
		self.CastBarBg:SetLayer(1);
		self.CastBar:SetTexture("myUI", "images/cast.jpg");
		self.CastBar:SetLayer(2)
		self.CastBar.Text:SetLayer(10);	
	end
	
	if (def.showLevel and def.showPortrait) then
		if (isnil(self.LevelBorder)) then self.LevelBorder = myUI:CreateFrame("Frame", self); end	
		if (isnil(self.LevelBg)) then self.LevelBg = myUI:CreateFrame("Frame", self.LevelBorder); end
		if (isnil(self.Level)) then self.Level = myUI:CreateFrame("Text", self.LevelBg); end
		
		self.LevelBorder:SetBackgroundColor(gray:rgba());
		self.LevelBg:SetBackgroundColor(black:rgba());
		
		self.LevelBorder:SetLayer(10);
		self.LevelBg:SetLayer(11);		
		self.Level:SetLayer(12);
	end

	if (def.showPvP) then
		if (isnil(self.PvP)) then self.PvP = myUI:CreateFrame("Texture", self); end
		
		self.PvP:SetLayer(20);		
		self.PvP.defaultSize = 24
		self.PvP:SetTexture("myUI", "images/pvp.png");
		self.PvP:SetVisible(false)
	end
	
	if (def.showVitality) then
		if (isnil(self.Vitality)) then 
			self.Vitality = myUI:CreateFrame("Texture", self); 
			self.Vitality:SetLayer(20);		
			self.Vitality.defaultSize = 24			
			self.Vitality:SetTexture("myUI", "images/vitality.png");
			self.Vitality:SetVisible(false);			
		end
	end
	
	-- ready check
	if (def.showReady) then
		if isnil(self.ReadyCheck) then
			self.ReadyCheck = myUI:CreateFrame("Texture", self)
			self.ReadyCheck:SetVisible(false)
			self.ReadyCheck:SetTexture("myUI", "images/ready.png")
			self.NotReadyCheck = myUI:CreateFrame("Texture", self)
			self.NotReadyCheck:SetVisible(false)
			self.NotReadyCheck:SetTexture("myUI", "images/error.png")
		end
		
	elseif (self.ReadyCheck) then
		self.ReadyCheck:SetVisible(false)
		self.NotReadyCheck:SetVisible(false)
	end
	
	-- raid mark
	if (def.showMark) then
		if isnil(self.Mark) then self.Mark = myUI:CreateFrame("Text", self) end
		self.Mark:SetLayer(11)
		self.Mark:SetFontColor(white:rgba())
		self.Mark:SetVisible(false)
	elseif (self.Mark) then
		self.Mark:SetVisible(false)
	end
	
	if (def.showLoot) and (self.UnitID ~= "player") and (self.UnitID ~= "player.pet") 
		and (not string.find("group", self.UnitID)) then
		if (isnil(self.Loot)) then
			self.Loot = myUI:CreateFrame("Texture", self)
			self.Loot:SetLayer(20)
			self.Loot.defaultSize = 32
			self.Loot:SetTexture("myUI", "images/loot.png")
			self.Loot:SetVisible(false)
		end
		
	end
	
	-- role indicator
	if (def.showRole) then
		if isnil(self.Role) then self.Role = myUI:CreateFrame("Texture", self) end
		self.Role:SetLayer(4)
		self.Role.defaultSize = 18
		self.Role:SetSize(self.Role.defaultSize, self.Role.defaultSize)
	elseif (self.Role) then
		self.Role:SetVisible(false)
	end
	
	-- planar charge
	if (def.showPlanar and (self.UnitID == "player")) then
		if isnil(self.Planar) then 
			self.Planar = myUI:CreateFrame("Frame", self)
			self.Planar:SetBackgroundColor(white:rgba())
			self.Planar.Text = myUI:CreateFrame("Text", self) 
				
			self.Planar.Text:SetFontColor(dodgerBlue:rgba())
			self.Planar.Text:SetFontSize(12)
			self.Planar.Text:SetBackgroundColor(black:rgba())
			
		end		
		
		
	elseif (self.Planar) then
		self.Planar:SetVisible(false)
	end
end


local function __disposeUF(self)
	if (notnil(self.Icon)) then 
		for _, icon in ipairs(self.Icons) do
			icon:Dispose(); 
		end
		self.Icons = nil; 
	end
	if (notnil(self.IconBg)) then self.IconBg:Dispose(); self.IconBg = nil; end
	if (notnil(self.UnitHealth.Text)) then self.UnitHealth.Text:Dispose(); self.UnitHealth.Text = nil; end
	if (notnil(self.UnitPower.Text)) then self.UnitPower.Text:Dispose(); self.UnitPower.Text = nil; end
	
	if (notnil(self.CastBar)) then
		if (notnil(self.CastBar.Text)) then self.CastBar.Text:Dispose(); self.CastBar.Text = nil; end
		self.CastBar:Dispose(); self.CastBar = nil;				
	end
	if (notnil(self.CastBarBg)) then self.CastBarBg:Dispose(); self.CastBarBg = nil; end
		
	if (notnil(self.LevelBg)) then self.LevelBg:Dispose(); self.LevelBg = nil; end
	if (notnil(self.Level)) then self.Level:Dispose(); self.Level = nil; end
	if (notnil(self.LevelBorder)) then self.LevelBorder:Dispose(); self.LevelBorder = nil; end
	if (notnil(self.PvP)) then self.PvP:Dispose(); self.PvP = nil; end
	if (notnil(self.Vitality)) then self.Vitality:Dispose(); self.Vitality = nil; end

	if (notnil(self.TitleBar)) then self.TitleBar:Dispose(); self.TitleBar = nil; end
	if (notnil(self.UnitName)) then self.UnitName:Dispose(); self.UnitName = nil; end
	if (notnil(self.UnitName2)) then self.UnitName2:Dispose(); self.UnitName2 = nil; end
	if (notnil(self.UnitHealth)) then self.UnitHealth:Dispose(); self.UnitHealth = nil; end
	
	if (notnil(self.UnitHealthBg)) then self.UnitHealthBg:Dispose(); self.UnitHealthBg = nil; end
	if (notnil(self.UnitPower)) then self.UnitPower:Dispose(); self.UnitPower = nil; end
	if (notnil(self.UnitPowerBg)) then self.UnitPowerBg:Dispose(); self.UnitPowerBg = nil; end

	for _, box in pairs(self.LongBuffBoxes) do
		box:Dispose();
	end
	
	if (notnil(self.Mover)) then self.Mover:Dispose(); self.Mover = nil; end
	
	self.LongBuffBoxes = nil;
	
	self:_dispose()
end

local function __layoutUF(self, reset)
	self._layout = true
		
	self:EnsureFrames()

	if (not self.Definition.enabled) and (not self._mover) then 
		self:SetVisible(false)
		self._layout = false
		return
	end

	

	if (reset) then
		self._lastDetail = {}
	end
	
	local def = self.Definition
	local x = def.x
	local y = def.y
	local cx = def.cx
	local cy = def.cy
	local anchorTo = def.anchorTo
	local anchorFrom = def.anchorFrom
	local p
	local unit = self.UnitID
	
	-- handle raid frame syncing 
	if (string.find(unit, "group%d%d")) then
		local gnum = tonumber(string.sub(unit, 6, 8))
		local ispet = string.find(unit, "pet")
		
		if (gnum ~= 1) then
			if (gnum == 6) or (gnum == 11) or (gnum == 16) then
				gnum = gnum - 5
				anchorFrom = "TOPLEFT"
				anchorTo = "TOPRIGHT"
				y = 0; x = def.colspacing
			else
				gnum = gnum - 1
				anchorFrom = "TOPLEFT"
				anchorTo = "BOTTOMLEFT"
				x = 0; y = def.rowspacing
			end
			
			local g = "group" .. lpad(string.format("%d", gnum), 2, "0")
			if (ispet) then g = g .. "." .. string.sub(unit, ispet) end
			
			p = UnitFrames.Frames[g]
			
			if (p == nil) then
				x = def.x
				y = def.y
				anchorTo = def.anchorTo
				anchorFrom = def.anchorFrom
			end
		end	
	end
	
	
	self:ClearAll()
	self:SetWidth(cx)
	self:SetHeight(cy)
	
	p = ifnil(p, UIParent)
	
	
	if (not self._mover) and (unit == "mouseover") and def.followMouse then
		local ox = 48
		local m = Inspect.Mouse()
		self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", m.x + ox, m.y - ox)
	elseif (not self._mover) and (unit == "mouseover.target") and def.followMouse then
		if (UnitFrames.Frames["mouseover"]) then
			self:SetPoint("LEFTCENTER", UnitFrames.Frames["mouseover"], "RIGHTCENTER", 32, 0)
		else
			local m = Inspect.Mouse()
			local ox = 48
			self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", m.x + ox, m.y - ox)
		end
	else
		self:SetPoint(anchorFrom, p, anchorTo, x, y)
	end
	-- so now our base is squared away
	if (self.CombatIndicator) then
		self.CombatIndicator:SetLayer(-2)		
		self.CombatIndicator:SetPoint("TOPLEFT", self, "TOPLEFT", -6, -6)		
		self.CombatIndicator:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 6, 6)
		self.CombatIndicator:SetBackgroundColor(firebrick:rgba())
		self.CombatIndicator:SetVisible(myUI:GetInCombat())
	end
	
	-- handle portrait
	if (def.showPortrait) then
		self.IconBg:ClearAll()		
		self.IconBg:SetSize(cy, cy)
		self.IconBg:SetPoint("TOPLEFT", self, "TOPLEFT")
		
		for i, icon in ipairs(self.Icons) do
			icon:ClearAll()			
			icon:SetLayer(2)
			icon:SetSize(cy-4, cy-4)			
			icon:SetPoint("TOPLEFT", self.IconBg, "TOPLEFT", 2, 2)			
		end		
		
		self.IconBg:SetVisible(true)
		
	else
		if (self.IconBg) then self.IconBg:SetVisible(false) end
		if (self.Icons) then 
			for _, icon in ipairs(self.Icons) do icon:SetVisible(false) end
		end
	end
	
	self.DeadIcon:ClearAll()
	self.DeadIcon:SetSize(cy-4, cy-4)
	self.DeadIcon:SetPoint("CENTER", self, "CENTER")
	self.DeadIcon:SetLayer(20)
	self.DeadIcon:SetVisible(false)
	
	-- ready check
	if (def.showReady) and (self.ReadyCheck) then
		self.ReadyCheck:ClearAll()
		self.NotReadyCheck:ClearAll()
		
		local rcxy = minmax(cy - 4, 24, 48)
		
		self.ReadyCheck:SetSize(rcxy, rcxy)
		self.ReadyCheck:SetLayer(21)
		self.ReadyCheck:SetPoint("CENTER", self, "CENTER")
		
		self.NotReadyCheck:SetSize(rcxy, rcxy)
		self.NotReadyCheck:SetLayer(21)
		self.NotReadyCheck:SetPoint("CENTER", self, "CENTER")
	end
	
	local cy3 = cy / 3
	local cy2 = cy / 2
	local minFont = 10
	local maxValueFont = 15
	local maxCastFont = 14
	local maxPercentFont = 24
	
	-- handle titlebar and name text
	self.UnitName:SetLayer(12); self.UnitName2:SetLayer(9)
	self.UnitName:ClearAll(); self.UnitName2:ClearAll()
	
	if (def.showRole and self.Role) then
		self.Role:ClearAll()
		self.Role:SetSize(self.Role.defaultSize, self.Role.defaultSize)
		self.Role:SetPoint("TOPLEFT", iif(def.TitleBar, self.TitleBar, self), "TOPLEFT", iif(def.showPortrait, cy + 2, 2), 2)		
	end
	
	if (def.showTitleBar) then
		self.TitleBar:ClearAll()
		self.TitleBar:SetHeight(minmax(cy3, 18, 30))
		self.TitleBar:SetPoint("TOPRIGHT", self, "TOPRIGHT")
		self.TitleBar:SetPoint("TOPLEFT", self, "TOPLEFT", iif(def.showPortrait, cy, 0), 0)
		self.TitleBar:SetVisible(true)
		
		cy2 = (cy - self.TitleBar:GetHeight()) / 2 -- this is what's left for our health and power bars
		
		local xoff = 2
		if (def.showRole) then xoff = xoff + self.Role.defaultSize + 4 end
		self.UnitName2:SetVisible(false)
		self.UnitName:SetFontSize(minmax(self.TitleBar:GetHeight() * 0.75, minFont, maxValueFont))
		self.UnitName:SetPoint("LEFTCENTER", self.TitleBar, "LEFTCENTER", xoff, 0)
		self.UnitName:SetPoint("RIGHTCENTER", self.TitleBar, "RIGHTCENTER", 0, 0)
	else
		if (self.TitleBar) then self.TitleBar:SetVisible(false) end
		
		local xoff = 2
		if (def.showPortrait) then 
			xoff = xoff + cy 
		end
		if (def.showRole) then
			xoff = xoff + self.Role.defaultSize + 2
		end
		
		self.UnitName:SetPoint("LEFTCENTER", self, "LEFTCENTER", xoff, 0)
		self.UnitName:SetPoint("RIGHTCENTER", self, "RIGHTCENTER", 0, 0)
		self.UnitName:SetFontSize(minmax(cy3 * 0.75, minFont, maxValueFont))
		self.UnitName2:SetFontSize(self.UnitName:GetFontSize())
		self.UnitName:SetPoint("TOPLEFT", self, "TOPLEFT", xoff, 0)
		self.UnitName2:SetPoint("TOPLEFT", self.UnitName, "TOPLEFT", 2, 0)
		self.UnitName2:SetVisible(true)
		
		self.UnitName:SetBackgroundColor(0,0,0,0.2)
		
	end
	
	
	
	if (def.showCharge and (self.UnitID == "player")) then
		self.ChargeBg:ClearAll()
		self.ChargeBg:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", iif(def.showPortrait, cy, 0), 0)
		self.ChargeBg:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT")
		self.Charge:SetPoint("TOPLEFT", self.ChargeBg, "TOPLEFT")
		self.Charge:SetPoint("BOTTOMLEFT", self.ChargeBg, "BOTTOMLEFT")
		powerAnchor = self.ChargeBg
	elseif self.ChargeBg then
		self.ChargeBg:SetHeight(0)	
	end
	
	
	
	
	-- handle power gauge
	self.UnitPowerBg:ClearAll()	
	self.UnitPowerBg:SetLayer(1)
	
	if (def.showPowerText or def.showPowerAsPercent) then
		-- normal display with text
		self.UnitPowerBg:SetHeight(minmax(cy2, 12, cy2))
		
		self.UnitPower.Text:ClearAll()
		self.UnitPower.Text:SetFontSize(minmax(self.UnitPowerBg:GetHeight() * 0.75, minFont, iif(def.showPowerAsPercent and not def.showPowerText, maxPercentFont, maxValueFont)))
		self.UnitPower.Text:SetPoint("RIGHTCENTER", self.UnitPowerBg, "RIGHTCENTER", -2, 0)
		self.UnitPower.Text:SetVisible(true)
		self.UnitPower.Text:SetLayer(10)
	else
		-- no text, so minimize this bar
		self.UnitPowerBg:SetHeight(6)
		if (self.UnitPower.Text) then self.UnitPower.Text:SetVisible(false) end
	end
	
	if (def.showCharge and (self.UnitID == "player")) then
		self.UnitPowerBg:SetPoint("BOTTOMRIGHT", self.ChargeBg, "TOPRIGHT")
		self.UnitPowerBg:SetPoint("BOTTOMLEFT", self.ChargeBg, "BOTTOMLEFT")
	else
		self.UnitPowerBg:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT")
		self.UnitPowerBg:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", iif(def.showPortrait, cy, 0), 0)
	end
	
	self.UnitPower:SetHeight(self.UnitPowerBg:GetHeight())	
	self.UnitPower:SetPoint("LEFTCENTER", self.UnitPowerBg, "LEFTCENTER")
	
	-- handle health gauge and text
	self.UnitHealthBg:ClearAll(); self.UnitHealth:ClearAll()
	if (def.showTitleBar) then
		self.UnitHealthBg:SetPoint("TOPLEFT", self.TitleBar, "BOTTOMLEFT")		
	else
		self.UnitHealthBg:SetPoint("TOPLEFT", self, "TOPLEFT", iif(def.showPortrait, cy, 0), 0)
	end
	self.UnitHealthBg:SetPoint("BOTTOMRIGHT", self.UnitPowerBg, "TOPRIGHT")
	self.UnitHealth:SetPoint("TOPLEFT", self.UnitHealthBg, "TOPLEFT")
	self.UnitHealth:SetPoint("BOTTOMLEFT", self.UnitHealthBg, "BOTTOMLEFT")
	
	if (def.showHealthText) or (def.showHealthAsPercent) then
		self.UnitHealth.Text:ClearAll()
		self.UnitHealth.Text:SetFontSize(minmax(self.UnitHealthBg:GetHeight() * 0.75, minFont, iif(def.showHealthAsPercent and not def.showHealthText, maxPercentFont, maxValueFont)))		
		self.UnitHealth.Text:SetPoint("RIGHTCENTER", self.UnitHealthBg, "RIGHTCENTER", -2, 0)
		self.UnitHealth.Text:SetVisible(true)
		self.UnitHealth.Text:SetLayer(10)
	else
		if (self.UnitHealth.Text) then self.UnitHealth.Text:SetVisible(false) end
	end
	
	-- mark
	if (def.showMark and self.Mark) then
		self.Mark:ClearAll()
		self.Mark:SetPoint("LEFTCENTER", self.UnitHealthBg, "LEFTCENTER", 2, 0)
		self.Mark:SetFontSize(1.5 * minmax(self.UnitHealthBg:GetHeight() * 0.75, minFont, iif(def.showHealthAsPercent and not def.showHealthText, maxPercentFont, maxValueFont)))
	end
	
	-- handle castbar
	if (def.showCastBar) then
		self.CastBarBg:ClearAll();
		self.CastBar.Text:ClearAll();
		self.CastBar:ClearAll();
		
		local cyCastbar = self.UnitPowerBg:GetHeight()
		
		self.CastBarBg:SetHeight(cyCastbar)
		self.CastBar:SetHeight(cyCastbar-4);		
		self.CastBar.Text:SetHeight(cyCastbar-4);
		self.CastBarBg:SetBackgroundColor(normal:rgba());
		
		local castFont = cyCastbar * 0.65
		if (castFont < minFont) then
			self.CastBar.Text:SetVisible(false)
		else
			self.CastBar.Text:SetFontSize(minmax(castFont, minFont, self.CastBarBg:GetHeight()))
			self.CastBar.Text:SetVisible(true)
		end				
		
		local fontadj = 0
		if (self.CastBarBg:GetHeight() < self.CastBar.Text:GetFullHeight()) then
			fontadj = -(self.CastBar.Text:GetFullHeight() - self.CastBarBg:GetHeight())*2			
		end
		
		self.CastBarBg:SetPoint("TOPLEFT", self, "BOTTOMLEFT", 0, 6)
		self.CastBarBg:SetPoint("TOPRIGHT", self, "BOTTOMRIGHT", 0, 2)
		self.CastBar:SetPoint("LEFTCENTER",self.CastBarBg, "LEFTCENTER", 2, 0)
		self.CastBar.Text:SetPoint("LEFTCENTER", self.CastBarBg, "LEFTCENTER", 4, fontadj)
		self.CastBar.Text:SetPoint("RIGHTCENTER", self.CastBarBg, "RIGHTCENTER", -4, fontadj)
		
	elseif (self.CastBarBg) then 
		self.CastBarBg:SetVisible(false)	
	end
	
	
	-- level box
	if (def.showLevel and def.showPortrait) then -- if no portrait, we show appeneded to the name
		self.LevelBg:ClearAll()
		self.LevelBorder:ClearAll()
		self.Level:ClearAll()
		
		self.Level:SetFontSize(minmax(0.6 * cy3, minFont, maxValueFont))
		self.Level:Autosize("WW")
		
		self.LevelBg:SetSize(self.Level:GetFullWidth(), self.Level:GetFullHeight() * 0.9)
		self.LevelBg:SetPoint("BOTTOMCENTER", self.IconBg, "BOTTOMCENTER", 0, 2)
		
		self.LevelBorder:SetPoint("TOPLEFT", self.LevelBg, "TOPLEFT", -1.5, -1.5)
		self.LevelBorder:SetPoint("BOTTOMRIGHT", self.LevelBg, "BOTTOMRIGHT", 1.5, 1.5)
		self.Level:SetPoint("CENTER", self.LevelBg, "CENTER")
		
		self.LevelBorder:SetVisible(true)
	else
		if (self.LevelBorder) then self.LevelBorder:SetVisible(false) end
	end
	
	-- gaurenteed loot indicator
	if (def.showLoot and self.Loot and not string.find(self.UnitID, "group") 
			and (self.UnitID ~= "player") and (self.UnitID ~= "player.pet")) then
		self.Loot:ClearAll()
		self.Loot:SetSize(self.Loot.defaultSize, self.Loot.defaultSize)
		if (def.showPortrait and self.IconBg) then
			self.Loot:SetPoint("CENTER", self.IconBg, "TOPCENTER", 0, 8)
		else
			self.Loot:SetPoint("CENTER", self, "TOPCENTER")
		end
	elseif (self.Loot) then
		self.Loot:SetVisible(false)
	end
	
	-- planar charge
	if (def.showPlanar and (self.UnitID == "player")) then
		self.Planar.Text:ClearAll()
		self.Planar.Text:SetLayer(8)
		
		self.Planar:ClearAll()
		self.Planar:SetLayer(7)
		self.Planar:SetPoint("TOPLEFT", self.Planar.Text, "TOPLEFT", -1.5, -1.5)
		self.Planar:SetPoint("BOTTOMRIGHT", self.Planar.Text, "BOTTOMRIGHT", 1.5, 1.5)
		
		if (def.showPortrait) then
			self.Planar.Text:SetPoint("LEFTCENTER", self.IconBg, "LEFTCENTER", -6, 0)
		else
			self.Planar.Text:SetPoint("LEFTCENTER", self, "LEFTCENTER")
		end
		
		
	elseif (self.Planar) then
		self.Planar:SetVisible(false)
		self.PlanarBg:SetVisible(false)
	end
	
	self._layout = false
end

local function __toggleMoverUF(self)
	if (self.Definition.enabled) then		
		self._mover = not self._mover
	else
		self._mover = false
	end
	self:Reset()
end
local function __showMoverUF(self)	
	if (self.Definition.enabled) then		
		self._mover = true
		self:Reset()	
	end
	
end
local function __hideMoverUF(self)	
	self._mover = false
	self:Reset()
end

UnitFrame = {
	new = function(self, unitID, def)
		local this = myUI:CreateFrame("Frame");
		this:SetBackgroundColor(black:rgba())
		
		this.UnitID = unitID;
		this.Definition = def;
		this.LongBuffBoxes = {}; -- list of long buff boxes we're tracking
		this.Channeling = false;				
		this._layout = nil;
	
		this._lastDetail = {}
	
		this.Border = myUI:CreateFrame("Frame", this)
		this.Border:SetLayer(-1)
		this.Border:SetBackgroundColor(black:rgba())
		this.Border:SetAlpha(0.6)
		this.Border:SetPoint("TOPLEFT", this, "TOPLEFT",  -2, -2)
		this.Border:SetPoint("BOTTOMRIGHT", this, "BOTTOMRIGHT", 2, 2)
		
		
		this.UpdateCastBar = __updateCastbar
		this.Update = __update
		this.EnsureFrames = __ensureFrames
		
		-- override our framework provided dispose method to dispose of all our child frames as well
		this._dispose = this.Dispose
		this.Dispose = __disposeUF		
		this.Layout = __layoutUF
		this.Reset = __resetUF
		this.ToggleMover = __toggleMoverUF
		this.ShowMover = __showMoverUF
		this.HideMover = __hideMoverUF
		this:Layout();
		this:Update();
		this:UpdateCastBar(nil, false);
		
		return this;
	end
	
}

--[[
local function __setXYTextUF(self)
	local sz = string.format("%d", self:GetLeft()) .. ", " .. string.format("%d", self:GetTop()) 
		.. "  (" .. string.format("%d", self:GetWidth()) .. " x " .. string.format("%d", self:GetHeight()) .. ")"
	
	self.XY:SetText(sz)
	self.XY:SetWidth(self.XY:GetFullWidth())
end

local function __disposeUFM(self)
	self.Text:Dispose(); self.Text = nil;

	self.UnitFrame = nil;
	self.Event.LeftDown = nil;
	self.Event.LeftUp = nil;
	self.Event.LeftUpoutside = nil;
	self.Event.MouseMove = nil;
	self.Event.WheelForward = nil;
	self.Event.WheelBack = nil;
	
	self._dispose();
end

local function __onLeftDownUFM(self)
	self.dragging = true;
	
	local m = Inspect.Mouse()
	self.lastx = m.x
	self.lasty = m.y
	self.offx = m.x - self:GetLeft() 
	self.offy = m.y - self:GetTop()			
end

local function __onLeftUpUFM(self)
	self.dragging = false;
	
	self.UnitFrame.Definition.x = math.floor(self:GetLeft());
	self.UnitFrame.Definition.y = math.floor(self:GetTop());
	
	self.UnitFrame:Layout(true);
	self.UnitFrame:Update();
	self.UnitFrame._layout = true;
	self.UnitFrame:SetAlpha(0.25);
	self.UnitFrame:SetVisible(true);
	
	self:setXYText();
end

local function __onMouseMoveUFM(self,x, y)
	if (self.dragging) then
		local m = Inspect.Mouse()
		
		if (math.abs(m.x - self.lastx) > 1) or (math.abs(m.y - self.lasty) > 1) then				
			local x = m.x - self.offx
			local y = m.y - self.offy
			
			self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", x, y);
	
			self.lastx = m.x
			self.lasty = m.y
			
			self:setXYText();
		end
	end
end

local function __onWheelForwardUFM(self)			
	local cx = self:GetWidth()
	local cy = self:GetHeight()
	
	local r = cx/cy
	
	cx = (cx + 2)
	cy = (cx / r)
	
	if (cx < 800) and (cy < 800) then
		self:SetWidth(cx)
		self:SetHeight(cy)
		
		self.UnitFrame.Definition.cx = cx
		self.UnitFrame.Definition.cy = cy
		
		self.UnitFrame:Layout(true)
		self.UnitFrame:Update()
		self.UnitFrame:SetAlpha(0.25)
		self.UnitFrame:SetVisible(true)
		
		self:setXYText();
	end
end

local function __onWheelBackUFM(self)
	local cx = self:GetWidth()
	local cy = self:GetHeight()
	
	local r = cx/cy
	
	cx = (cx - 2)
	cy = (cx / r)
	
	if (cx > 20) and (cy >30) then
		self:SetWidth(cx)
		self:SetHeight(cy)
		
		self.UnitFrame.Definition.cx = cx
		self.UnitFrame.Definition.cy = cy
		
		self.UnitFrame:Layout(true)
		self.UnitFrame:Update()
		self.UnitFrame:SetAlpha(0.25)
		self.UnitFrame:SetVisible(true)
		
		self:setXYText();
	end
end



UnitFrameMover = {
	new = function(self, uframe)
		local this = myUI:CreateFrame("Frame");
		
		this.UnitFrame = uframe
		
		this:SetWidth(uframe.Definition.cx);
		this:SetHeight(uframe.Definition.cy);
		this:SetPoint("TOPLEFT", UIParent, "TOPLEFT", uframe.Definition.x, uframe.Definition.y);
		
		this:SetBackgroundColor(.6,0,0,0.4);
		
		this:SetVisible(false);
		
		this.Text = myUI:CreateFrame("Text", this);
		this.Text:SetFontSize(12);
		this.Text:SetFontColor(white:rgba());
		this.Text:SetPoint("TOPLEFT", this, "TOPLEFT", 4, 1);
		this.Text:SetPoint("TOPRIGHT", this, "TOPRIGHT", -4, 1);
		this.Text:SetHeight(14);
		this.Text:SetText(uframe.UnitID);
		this.Text:SetWordwrap(false);
		
		this.XY = myUI:CreateFrame("Text", this)
		this.XY:SetFontSize(12)
		this.XY:SetHeight(20)
		this.XY:SetFontColor(white:rgba());
		this.XY:SetPoint("TOPLEFT", this, "TOPLEFT", -16, -24);
		this.XY:SetWordwrap(false)
		
		this.setXYText = __setXYTextUF	
		this:setXYText();
		
		this.Instruct = myUI:CreateFrame("Text", this)
		this.Instruct:SetFontSize(12)
		this.Instruct:SetFontColor(white:rgba())
		this.Instruct:SetPoint("TOPLEFT", this.Text, "BOTTOMLEFT", 0, 4)
		this.Instruct:SetPoint("BOTTOMRIGHT", this, "BOTTOMRIGHT", -1, -1)
		this.Instruct:SetWordwrap(true)
		
		this.Instruct:SetText("Drag to move, Mouse Wheel to resize.")
		
		this:SetLayer(100);
		
		this._dispose = this.Dispose
		this.Dispose = __disposeUFM
	
		this.dragging = false;
		this.offx = 0;
		this.offy = 0;
		this.lastx = 0;
		this.lasty = 0;
		
		this.Event.LeftDown = __onLeftDownUFM
		this.Event.LeftUp = __onLeftUpUFM
		this.Event.LeftUpoutside = __onLeftUpUFM
		this.Event.MouseMove = __onMouseMoveUFM
		this.Event.WheelForward = __onWheelForwardUFM
		this.Event.WheelBack = __onWheelBackUFM
		
		return this;
	
	end
}

]]--

BuffBox = {

	buffBoxCount = 0,
	
	new = function(self, parent, def, detail)
		self.buffBoxCount = self.buffBoxCount + 1;
		
		local this = myUI:CreateFrame("Frame", parent);
		
		this.BuffName = detail.name;
		
		this.Definition = def;
		
		this:SetBackgroundColor(UnitFrames.bgNormal:rgba());
	
		this.Icon = myUI:CreateFrame("Texture", this);
		local border = UnitFrames.Border;
		
		this.Icon:SetPoint("TOPLEFT", this, "TOPLEFT", border, border);
		this.Icon:SetPoint("BOTTOMRIGHT", this, "BOTTOMRIGHT", -border, - border);
		this.Icon:SetTexture("Rift", detail.icon);
		this.Icon:SetLayer(1);
		
		this.Blink = true;
		this.Warning = false;
		
		this._dispose = this.Dispose
		this.Dispose = function(self)
			self.Icon:Dispose();
			self.Event.RightClick = nil;
			self._dispose();
		end
		
		this.Event.RightClick = function(self)				
				self.Dismissed = true;
				self:SetVisible(false);
			end
		
		this.Update =  function(self, detail)
			if (detail ~= nil) then this.BuffName = detail.name; end
			
			if (self.Dismissed) then
				self:SetVisible(false);
			else
				if (detail ~= nil) and ((detail.duration == nil) or (ifnil(detail.remaining, 0) > 0)) then
					self.Warning = false;
					
					self:SetAlpha(0.25);
					self:SetWidth(self.Definition.cxyNormal);
					self:SetHeight(self.Definition.cxyNormal);
					self:SetBackgroundColor(UnitFrames.bgNormal:rgba());
				else
					self.Warning = true;					
					self:SetBackgroundColor(UnitFrames.expiredBuff:rgba());
					--self:SetAlpha(1);
					self:SetWidth(self.Definition.cxyExpired);
					self:SetHeight(self.Definition.cxyExpired);
				end			
				self:SetVisible(true)
			end			
		end
		
		this.UpdateBlink = function(self)
			if (self.Warning) then
				if (self:GetVisible()) then
					local a = self:GetAlpha();
					
					if (self.Blink) then
						a = a + 0.1;
					else
						a = a - 0.1;
					end
					
					self:SetAlpha(minmax(a, 0.25, 1)); 
					if (a >= 1) then
						self.Blink = false;
					elseif (a <= 0.29) then
						self.Blink = true;
					end
					
				end
			else
				self.Blink = false;
			end
		end
	
		this:Update(detail);
		return this;
	end

}






