
local me = {}

local ifnil = myUI.global.ifnil
local isnil = myUI.global.isnil
local notnil = myUI.global.notnil
local nilstr = myUI.string.nilstr
local minmax = myUI.global.minmax
local iif = myUI.global.iif

local white = myUI.colors.White
local black = myUI.colors.Black
local transparent = myUI.color:new(0,0,0,0)
local activebg = myUI.color:clone(black, 100)	
local active = myUI.colors.DeepSkyBlue

local mbox, __recycleMsgBox, __getMsgBox
local _recycledMsgBox = {}
local ibox,  __recycleInputBox, __getInputBox
local _recycledInputBox = {}

--[[ 
Recognized message box buttons
	"OK",
	"CANCEL"
	"OKCANCEL"
	"YESNO"
	"YESNOCANCEL"
	
results ..
	"OK"
	"CANCEL"
	"YES"
	"NO"
	"ERROR"   -- incorrect buttons passed in
	"NONE"   -- message box has not been closed
	
Recognized icons ...
	"INFO"
	"ERROR"
	"WARNING"
	"QUESTION"
]]--


function myUI.msgbox(parent, message, title, buttons, icon, callback, state, cx, cy, layer)
	-- callback is a function that will be called when a button is clicked
	--	callback(result, state)
	--
	-- if specified, this function must return True or nil to allow the button click to continue
	local box = __getMsgBox()
	box:Initialize(parent, message, title, buttons, icon, callback, state, cx, cy, layer)
	box:SetVisible(true)
end


__recycleMsgBox = function(box)
	if (box) then
		table.insert(_recycledMsgBox, box)
	end
end


__getMsgBox = function()
	if (#_recycledMsgBox > 0) then
		return table.remove(_recycledMsgBox)
	else
		return mbox:new()
	end
end

mbox = {
	new = function(self) --, parent, message, title, buttons, icon, callback, cx, cy, layer)
		-- callback is a function that will be called when a button is clicked
		--	callback(result)
		--
		-- if specified, this function must return True or nil to allow the button click to continue
		
		local this = myUI:CreateFrame("RiftWindow")
		this:SetVisible(false)
		this:SetController("content")
		
		this.Icon = myUI:CreateFrame("Texture", this)
		this.Icon:SetPoint("TOPLEFT", this, "TOPLEFT", 8, 0)
		this.Icon:SetWidth(0)
		this.Icon:SetHeight(32)
		
		this.Button1 = myUI:CreateFrame("RiftButton", this)
		this.Button2 = myUI:CreateFrame("RiftButton", this)
		this.Button3 = myUI:CreateFrame("RiftButton", this)
		
		this.onButton = function(btn)
			local p = btn:GetParent()
			p:SetVisible(false)
			
			if (btn.callback) then
				btn.callback(btn.Result, btn:GetParent().State)				
			end
			
			__recycleMsgBox(p)
		end
		
		this.Button1.Event.LeftPress = this.onButton
		this.Button2.Event.LeftPress = this.onButton
		this.Button3.Event.LeftPress = this.onButton
		
		this.Message = myUI:CreateFrame("Text", this)
		--this.Message:SetFontSize(16)
		this.Message:SetPoint("TOPLEFT", this.Icon, "TOPRIGHT", 8, 0);
		this.Message:SetPoint("BOTTOMRIGHT", this, "BOTTOMRIGHT", -8, -40);
		this.Message:SetFontColor(myUI.colors.White:rgba())
		this.Message:SetWordwrap(true)
		
		this.Initialize = function(self, parent, message, title, buttons, icon, callback, state, cx, cy, layer)
			self:SetTitle(ifnil(title, "myUI"))						
			self.Result = "NONE"
			self.State = state
			local L = myUI._L
			
			if (notnil(buttons)) then buttons = string.upper(buttons) end
			if (notnil(icon)) then icon = string.upper(icon) end
			
			self.Message:SetText(ifnil(message,""))
			
			this.Icon:SetWidth(this.Icon:GetHeight()+2)
			if (icon == "WARNING") then
				this.Icon:SetTexture("myUI", "images/warning.png")
				this.Icon:SetWidth(this.Icon:GetHeight())
			elseif (icon == "ERROR") then	
				this.Icon:SetTexture("myUI", "images/error.png")
				this.Icon:SetWidth(this.Icon:GetHeight())
			elseif (icon == "INFO") then
				this.Icon:SetTexture("myUI", "images/info.png")
				this.Icon:SetWidth(this.Icon:GetHeight())
			elseif (icon == "QUESTION") then
				this.Icon:SetTexture("myUI", "images/question.png")
				this.Icon:SetWidth(this.Icon:GetHeight())
			else
				this.Icon:SetWidth(0)
			end
			
			this.Button1.callback = callback
			this.Button2.callback = callback
			this.Button3.callback = callback
			
			buttons = string.upper(ifnil(buttons, "OKCANCEL"))
			local btns = 1
			
			if (buttons == "OK") then
				this.Button2:SetText(L["OK"]); this.Button2.Result = "OK"
			elseif (buttons == "CANCEL") then
				this.Button2:SetText(L["Cancel"]); this.Button2.Result = "CANCEL"
			elseif (buttons == "OKCANCEL") then
				this.Button2:SetText(L["OK"]); this.Button2.Result = "OK"
				this.Button3:SetText(L["Cancel"]); this.Button3.Result = "CANCEL"
				btns = 2
			elseif (buttons == "YESNO") then
				this.Button2:SetText(L["Yes"]); this.Button2.Result = "YES"
				this.Button3:SetText(L["No"]); this.Button3.Result = "NO"
				btns = 2
			elseif (buttons == "YESNOCANCEL") then
				this.Button1:SetText(L["Yes"]); this.Button1.Result = "YES"
				this.Button2:SetText(L["No"]); this.Button2.Result = "NO"
				this.Button3:SetText(L["Cancel"]) this.Button3.Result = "CANCEL"
				btns = 3
			else
				this.Button2:SetText(""); this.Button2.Result = "ERROR"				
			end
			
			self.Button1:ClearAll()
			self.Button2:ClearAll()
			self.Button3:ClearAll()
			
			
			
			if (btns == 2) then
				local cxb = self.Button1:GetWidth() * 2
				local xb = (self:GetWidth() - cxb) / 2 
				
				self.Button2:SetPoint("BOTTOMLEFT", this, "BOTTOMLEFT", xb, 0)
				self.Button3:SetPoint("BOTTOMLEFT", this.Button2, "BOTTOMRIGHT")
			
			else
				this.Button2:SetPoint("BOTTOMCENTER", this, "BOTTOMCENTER")
				this.Button1:SetPoint("RIGHTCENTER", this.Button2, "LEFTCENTER")
				this.Button3:SetPoint("LEFTCENTER", this.Button2, "RIGHTCENTER")					
			end
			
			self.Button1:SetVisible(btns == 3)
			self.Button2:SetVisible(true)
			self.Button3:SetVisible(btns > 1)
			
			local bottom, left, right, top = self:GetTrimDimensions()
						
			local mincx = self.Button1:GetWidth() * 3 + 24
			local mincy = self.Button1:GetHeight() + top + self.Message:GetFullHeight() + 20
			
			cx = ifnil(cx, mincx)
			cy = ifnil(cy, mincy)

			parent = ifnil(parent, UIParent)
			local x = (parent:GetWidth() - cx) / 2
			local y = (parent:GetHeight() - cy) / 2
			
			self:SetWidth(cx)
			self:SetHeight(cy)
			self:SetPoint("TOPLEFT", parent, "TOPLEFT", x, y)
			
			layer = ifnil(layer, 5000)
			
			self:SetLayer(layer)
			self.Icon:SetLayer(layer+1)
			self.Message:SetLayer(layer+1)
			self.Button1:SetLayer(layer+2)
			self.Button2:SetLayer(layer+2)
			self.Button3:SetLayer(layer+2)
		
		
		end
	
		--this:Initialize(parent, message, title, buttons, icon, callback, cx, cy, layer)
		return this
	end
	
}

-- callback is a function that will be called when a button is clicked
--	callback(result, inputtext, state)
--
-- if specified, this function must return True or nil to allow the button click to continue

function myUI.inputbox(parent, message, title, callback, state, cx, cy, layer)
	local box = __getInputBox()
	box:Initialize(parent, message, title, callback, state, cx, cy, layer)
	box:SetVisible(true)
end

local ibox = {
	new = function(self)
		local this = myUI:CreateFrame("RiftWindow")
		this:SetVisible(false)
		this:SetController("content")
		
		local L = myUI._L
		
		this.Message = myUI:CreateFrame("Text", this)
		this.Message:SetPoint("TOPLEFT", this, "TOPLEFT", 8, 0)
		this.Message:SetPoint("TOPRIGHT", this, "TOPRIGHT", -8, 0)
		this.Message:SetFontColor(myUI.colors.White:rgba())
		this.Message:SetWordwrap(true)
		
		this.btnCancel = myUI:CreateFrame("RiftButton", this)
		this.btnCancel:SetText(L["Cancel"])
		this.btnCancel:SetPoint("BOTTOMRIGHT", this, "BOTTOMRIGHT")
		this.btnCancel.Result = "CANCEL"
		
		this.btnOK = myUI:CreateFrame("RiftButton", this)
		this.btnOK:SetText(L["OK"])
		this.btnOK:SetPoint("BOTTOMRIGHT", this.btnCancel, "BOTTOMLEFT")
		this.btnOK.Result = "OK"
		
		this.txtInput = myUI:CreateFrame("RiftTextfield", this)
		this.txtInput:SetPoint("BOTTOMLEFT", this, "BOTTOMLEFT", 8, -this.btnCancel:GetHeight())
		this.txtInput:SetPoint("BOTTOMRIGHT", this, "BOTTOMRIGHT", -8, - this.btnCancel:GetHeight())
		this.txtInput:SetBackgroundColor(activebg:rgba())
		
		this.onButton = function(btn)
			local p = btn:GetParent()
			p:SetVisible(false)
			
			if (btn.callback) then
				btn.callback(btn.Result, p.txtInput:GetText(), p.State)							
			end
			
			__recycleInputBox(p)
		end
		
		this.btnOK.Event.LeftPress = this.onButton
		this.btnCancel.Event.LeftPress = this.onButton
		
		this.Initialize = function(self, parent, message, title, callback, state, cy, cy, layer)
			self:SetTitle(ifnil(title, "myUI"))						
			self.Result = "NONE"
			self.State = state
			
			self.txtInput:SetText("")
			
			local L = myUI._L
			
			self.Message:SetText(ifnil(message,""))
			
			
			this.btnOK.callback = callback
			this.btnCancel.callback = callback			
			
			local bottom, left, right, top = self:GetTrimDimensions()
						
			local mincx = self.btnCancel:GetWidth() * 3 + 24
			local mincy = self.btnCancel:GetHeight() + top + self.Message:GetFullHeight() + 20
			
			cx = ifnil(cx, mincx)
			cy = ifnil(cy, mincy)

			parent = ifnil(parent, UIParent)
			local x = (parent:GetWidth() - cx) / 2
			local y = (parent:GetHeight() - cy) / 2
			
			self:SetWidth(cx)
			self:SetHeight(cy)
			self:SetPoint("TOPLEFT", parent, "TOPLEFT", x, y)
			
			layer = ifnil(layer, 5000)
			
			self:SetLayer(layer)
			self.Message:SetLayer(layer+1)
			self.btnCancel:SetLayer(layer+2)
			self.btnOK:SetLayer(layer+2)			
			self.txtInput:SetLayer(layer+2)
			
			self.txtInput:SetKeyFocus(true)
		end
		
		return this
	end
}


__recycleInputBox = function(box)
	if (box) then
		table.insert(_recycledInputBox, box)
	end
end

__getInputBox = function()	
	if (#_recycledInputBox > 0) then
		return table.remove(_recycledInputBox)
	else
		return ibox:new()
	end
end

-- A drag box is a pre-defined frame that can be used for drag/drop on any other frame
-- It's assumed that 'target' is a top-level frame aligned topleft to uiparent

local dragbox = {
	new = function(self, target) 
		if (target) then
			local this = myUI:CreateFrame("Frame", target)
	
			local layer = target:GetLayer()
			if (notnil(layer)) then
				this:SetLayer(layer+1)
			end
	
			this:SetBackgroundColor(0,0,0,0)
			
			this.dragging = false
			this.offx = 0
			this.offy = 0
			this.lastx = 0
			this.lasty = 0
			
			
			this.Event.LeftDown = function(self)
				self.dragging = true
				
				local m = Inspect.Mouse()
				self.lastx = m.x
				self.lasty = m.y
				
				local p = self:GetParent()
				self.offx = m.x - p:GetLeft()
				self.offy = m.y - p:GetTop()
			end
			
			this.onLeftUp = function(self)
				self.lastx = 0
				self.lasty = 0
				self.dragging = false
			end
			this.Event.LeftUp = this.onLeftUp
			this.Event.LeftUpoutside = this.onLeftUp
			
			this.Event.MouseMove = function(self, x, y)
				if (self.dragging) then
					local m = Inspect.Mouse()
					if (m.x and m.y) then
						if (math.abs(m.x - self.lastx) > 1) or (math.abs(m.y - self.lasty) > 1) then
							local p = self:GetParent()
							local x = m.x - self.offx
							local y = m.y - self.offy
						
							p:SetPoint("TOPLEFT", UIParent, "TOPLEFT", x, y)
						
							self.lastx = m.x
							self.lasty = m.y											
						end
					end
				end
			end
	
			this._baseDispose = this.Dispose
			this.Dispose = function(self)
				this.Event.LeftDown = nil
				this.Event.LeftUpoutside = nil
				this.Event.LeftUp = nil
				this.Event.MouseMove = nil
				
				this:_baseDispose()
			end
	
			return this
		end
	end
}


local bbox = {
	new = function(self, parent)
		local this = myUI:CreateFrame("Frame", parent)
		
		local bleft = myUI:CreateFrame("Frame", this)
		local btop = myUI:CreateFrame("Frame", this)
		local bright = myUI:CreateFrame("Frame", this)
		local bbottom = myUI:CreateFrame("Frame", this)
		
		this._borders = {bleft, btop, bright, bbottom}
		
		local cxy = 2
		
		bleft:SetPoint("TOPLEFT", this, "TOPLEFT", -1, 0)
		bleft:SetPoint("BOTTOMLEFT", this, "BOTTOMLEFT", -1, 0)
		bleft:SetWidth(cxy)
		
		btop:SetPoint("TOPLEFT", this, "TOPLEFT", 0, -1)
		btop:SetPoint("TOPRIGHT", this, "TOPRIGHT", 0, -1)
		btop:SetHeight(cxy)
		
		bright:SetPoint("TOPRIGHT", this, "TOPRIGHT", 1, 0)
		bright:SetPoint("BOTTOMRIGHT", this, "BOTTOMRIGHT", 1, 0)
		bright:SetWidth(cxy)
		
		bbottom:SetPoint("BOTTOMLEFT", this, "BOTTOMLEFT", 0, 1)
		bbottom:SetPoint("BOTTOMRIGHT", this, "BOTTOMRIGHT", 0, 1)
		bbottom:SetHeight(cxy)
		
		local white = myUI.color:clone(myUI.colors.White, 100);
		
		for _, b in ipairs(this._borders) do
			b:SetBackgroundColor(white:rgba())			
		end
		
		
		this.GetBorderColor = function(self)
			return self._border[1]:GetBackgroundColor()
		end
		
		this.SetBorderColor = function(self, r, g, b, a)
			for _, b in ipairs(self._borders) do b:SetBackgroundColor(r,g,b,a) end
		end
	
		
		this._baseSetVisible = this.SetVisible
		this.SetVisible = function(self, value)
			for _, b in ipairs(self._borders) do b:SetVisible(value) end
			this:_baseSetVisible(value)
		end
		
		this._baseSetLayer = this.SetLayer
		this.SetLayer = function(self, value)
			if (value) then
				for _, b in ipairs(self._borders) do b:SetLayer(value + 1) end
				this:_baseSetLayer(value)
			end
		end
		
		
		this._baseDispose = this.Dispose
		this.Dispose = function(self)
			for _, b in ipairs(self._borders) do b:Dispose() end
			self:_baseDispose()
		end
		return this
	end
}


local function __checkAnchorABX(chk)
	local p = chk:GetParent()
	
	if (chk:GetChecked()) then
		for k, a in pairs(p._anchorChecks) do
			if (k ~= chk.anchor) then a:SetChecked(false) end
		end
		
		if (p.AnchorChanged) then
			p:AnchorChanged(chk.anchor)
		end	
	end
	
	
end

local function __setAnchorABX(self, anchor)
	self._anchorChecks[anchor]:SetChecked(true)
end

local function __getAnchorABX(self)
	for _, a in pairs(self._anchorChecks) do
		if (a:GetChecked()) then return a.anchor end
	end
end


local function __disposeABX(self)

	for _, chk in pairs(self._anchorChecks) do
		chk.Event.CheckboxChange = nil
		chk:Dispose()
	end

	self._anchorChecks = nil
	self:_dispose()
end
local function __setScaleABX(self, scale)
	if (scale) then
		scale = minmax(scale, 0.5, 2)
		
		local r = self:GetHeight() / self:GetWidth()
		self:SetWidth(200 * scale)
		self:SetHeight(200 * scale * r)
		
		for _, c in pairs(self._anchorChecks) do
			c:SetHeight(24 * scale)
			c:SetWidth(24 * scale)
		end
		
	end
end
local function __setEnabledABX(self, enabled)
	for _, c in pairs(self._anchorChecks) do
		c:SetEnabled(enabled)
	end
end

local anchorbox = {
	new = function(self, p)
		-- Events
		--	AnchorChanged(chkbox)
		
		if (nil == p) then p = UIParent end
	
		local this = bbox:new(p)
		
		local cxscreen = UIParent:GetWidth()
		local cyscreen = UIParent:GetHeight()
		
		this:SetWidth(200)
		this:SetHeight(200 * cyscreen/cxscreen)
		
		local atext = {"TOPLEFT", "TOPCENTER", "TOPRIGHT", "RIGHTCENTER", "BOTTOMRIGHT", "BOTTOMCENTER", "BOTTOMLEFT", "LEFTCENTER", "CENTER"}
		local aoff = {4, 0, 0, 0, 0, 0, 4, 4, 0}
		local anchors = {}
		
		for i = 1, 9, 1 do
			local c = myUI:CreateFrame("RiftCheckbox", this)
			anchors[atext[i]] = c
			
			c:SetLayer(p:GetLayer() + 5)
			c:SetWidth(24)
			c:SetHeight(24)			
			c:SetPoint(atext[i], this, atext[i], aoff[i], 0)
			c.anchor = atext[i]
			
			--if (ComboTracker.Definition.anchor == atext[i]) then
			--	c:SetChecked(true)
			--end
			
			c.Event.CheckboxChange = __checkAnchorABX
			
		end
		
		this._anchorChecks = anchors
		this._dispose = this.Dispose
		this.Dispose = __disposeABX
		
		this.SetAnchor = __setAnchorABX
		this.GetAnchor = __getAnchorABX
		this.SetScale = __setScaleABX
		this.SetEnabled = __setEnabledABX
		return this
	end


}

local function __diposeTS(self)
	for _, t in ipairs(self._tabs) do
		t:Dispose()
	end
	
	self._tabs = nil
	self:_dispose()
end
local function __getTabTS(self, index)
	return self._tabs[index]
end

local function __clickTAB(tab)
	local p = tab:GetParent()
	if (not tab._selected) then
		for _, t in ipairs(p._tabs) do
			if t._selected then
				t._selected = false
				t:SetBackgroundColor(transparent:rgba())
				t:SetFontColor(white:rgba())
			end
		end
		
		tab:SetBackgroundColor(activebg:rgba())
		tab:SetFontColor(active:rgba())
		tab._selected = true
		
		if (p.SelectedIndexChanged) then
			p:SelectedIndexChanged(tab.index)
		end
	end
end
local function __getSelectedTAB(tab)
	return tab._selected
end

local function __getSelectedIndexTS(self)
	for i, v in ipairs(self._tabs) do
		if (v:GetSelected()) then return v.index end
	end
	
	return -1
end
local function __setSelectedIndexTS(self, index)
	if (index > 0) and (index <= #self._tabs) then
		__clickTAB(self._tabs[index])
	end
end
local function __getTabCountTS(self) 
	return #self._tabs
end
local function __getActiveBgColorTS(self)
	return activebg
end
local function __getActiveFgColorTS(self)
	return active
end

local tabstrip = {
	-- Events
	--	SelectedIndexChanged(tabstrip, index)
	
	new = function(self, tabs, orientation, parent)		
		local this = myUI:CreateFrame("Frame", parent)
		-- orientation = "VERTICAL", "HORIZONTAL"
		
		local isnil = myUI.global.isnil
		local ifnil = myUI.global.ifnil
		local iif = myUI.global.iif
		
		orientation = ifnil(orientation, "HORIZONTAL")
		
		this:SetHeight(20)
		this:SetBackgroundColor(transparent:rgba())
		
		local layer = parent:GetLayer() + 1
		this:SetLayer(layer)
		
		this._tabs = {}
		local cx = 0
		local cxtab = 32		
		local cxmax = cxtab
		local vert = (orientation == "VERTICAL")
		local cytab = iif(vert, 30, 20)
		if (tabs) then
			local anchor = this
			local anchorFrom = "TOPLEFT"
			
			for i, v in ipairs(tabs) do
				local t = myUI:CreateFrame("Text", this)
				
				t:SetFontColor(white:rgba())
				t:SetFontSize(14)
				t:SetText(" " .. ifnil(v, ""))
				t:SetHeight(cytab)
				t:SetLayer(layer+1)
				local w = t:GetFullWidth() + 4
				w = minmax(w, cxtab, w)
				if (w > cxmax) then cxmax = w end
				
				t:SetWidth(w)
						
				cx = iif(vert, cxmax, cx + w)
				
				
				t:SetPoint("TOPLEFT", anchor, anchorFrom)
				t:SetVisible(true)
				
				t.index = i
				t.Event.LeftClick = __clickTAB
				t.IsSelected = __getSelectedTAB
				
				
				anchor = t
				anchorFrom = iif(vert, "BOTTOMLEFT", "TOPRIGHT")
				
				table.insert(this._tabs, t)
			end
			
			if (vert) then
				for _, v in ipairs(this._tabs) do
					v:SetWidth(cxmax)
				end
				
				if (#this._tabs > 0) then
					this:SetHeight(#this._tabs * cytab)
				end
			end
		end
		
		if (vert) then
			cxmax = cxmax + 20
			for _, t in ipairs(this._tabs) do
				t:SetWidth(cxmax)
			end
			this:SetWidth(cxmax)
		else
			this:SetWidth(minmax(cx, 100, cx))
		end
		
		this.GetTab = __getTabTS
		this.GetSelectedIndex = __getSelectedIndexTS
		this.SetSelectedIndex = __setSelectedIndexTS
		this.GetTabCount = __getTabCountTS
		this.GetActiveBgColor = __getActiveBgColorTS
		this.GetActiveFgColor = __getActiveFgColorTS
		
		this._dispose = this.Dispose
		this.Dispose = __disposeTS
	
		this:SetVisible(true)
		return this
	end
}





local MIN_SLIDER_THUMB = 32

local function __setItemSelectedLB(self, index)
	if (index > 0) and (index <= #self._items) then
		for i, label in ipairs(self._labels) do
			label.isSelected = (i == index)
			if (label.isSelected) then
				label:SetFontColor(active:rgba())
				label:SetBackgroundColor(activebg:rgba())				
			else
				label:SetFontColor(white:rgba())
				label:SetBackgroundColor(transparent:rgba())
			end
		end		
	end
	if (self.SelectedIndexChanged) then 
		self:SelectedIndexChanged(index) 
	end
end
local function __getItemSelectedLB(self, index)
	if (index > 0) and (index <= #self._items) then
		return self._labels[index].isSelected
	end
end
local function __getSelectedIndexLB(self)

	local r = -1
	for i, label in ipairs(self._labels) do
		if (label.isSelected) then r = i; break end
	end
	return r
end
local function __getSelectedItemLB(self)
	local r = nil
	for i, label in ipairs(self._labels) do
		if (label.isSelected) then 
			r = label.Value
			break
		end
	end
	return r
end
local function __onLeftClickLBI(label)
	__setItemSelectedLB(label:GetParent():GetParent(), label.Index)
end

local function __getItemHeightLB(self)
	if (#self._labels > 0) then
		return math.floor(self._labels[1]:GetHeight() + 0.5)
	else
		return 0
	end
end

local function __getVirtualHeightLB(self)
	return math.floor((#self._items * __getItemHeightLB(self)) + 0.5)
end
local function __getRealHeightLB(self)
	local _, top, _, bottom = self.Viewport:GetBounds()
	
	return math.floor((bottom - top) + 0.5)
end

local function __layoutLB(self)
	local realcy = __getRealHeightLB(self)
	local virtcy = __getVirtualHeightLB(self)
	
	if (#self._items > 0) then	
		if (realcy > MIN_SLIDER_THUMB) then
			if (realcy < virtcy) then
				self.SliderThumb:SetHeight(minmax(realcy * (realcy/virtcy), MIN_SLIDER_THUMB, realcy))
			else
				self.SliderThumb:SetHeight(realcy)
			end
			
			if (virtcy <= realcy) then
				self.SliderThumb:SetPoint("TOPCENTER", self.Slider, "TOPCENTER", 0, 0)
				self._labels[1]:SetPoint("TOPLEFT", self.Viewport, "TOPLEFT", 0, 0)
				self._labels[1]:SetPoint("TOPRIGHT", self.Viewport, "TOPRIGHT", 0, 0)
			else
				local _, topthumb, _, bottomthumb = self.SliderThumb:GetBounds()
				local _, topslider, _, bottomslider = self.Slider:GetBounds()
				
				if (bottomthumb > bottomslider) then
					local overhang = bottomthumb - bottomslider
					local offset = topthumb - topslider - overhang
					
					self.SliderThumb:SetPoint("TOPCENTER", self.Slider, "TOPCENTER", 0, offset)
					
					local r = offset / (bottomslider - topslider)
					local cyvirt = __getVirtualHeightLB(self)
					local cyvirt = minmax(cyvirt, __getRealHeightLB(self), cyvirt)
					
					offset = cyvirt * r
					self._labels[1]:SetPoint("TOPLEFT", self.Viewport, "TOPLEFT", 0, -offset)
					self._labels[1]:SetPoint("TOPRIGHT", self.Viewport, "TOPRIGHT", 0, -offset)
				end
			end
		else
			self.SliderThumb:SetHeight(MIN_SLIDER_THUMB)
		end
		self.SliderThumb:SetVisible(true)
	else
		self.SliderThumb:SetVisible(false)
	end
end


local function __ensureLabelsLB(self)
	local nlabels = #self._labels
	local label
	
	while(#self._labels < #self._items) do
		label = myUI:CreateFrame("Text", self.Viewport)
		if (#self._labels == 0) then
			label:SetPoint("TOPLEFT", self.Viewport, "TOPLEFT")
			label:SetPoint("TOPRIGHT", self.Viewport, "TOPRIGHT")
		else
			label:SetPoint("TOPLEFT", self._labels[#self._labels], "BOTTOMLEFT")
			label:SetPoint("TOPRIGHT", self._labels[#self._labels], "BOTTOMRIGHT")
		end
		label.Event.LeftClick = __onLeftClickLBI
		
		label:SetWordwrap(false)
		label._index = #self._labels + 1
		table.insert(self._labels, label)
	end
	
	for i = 1, #self._items, 1 do
		self._labels[i]:SetText(tostring(self._items[i]))
		self._labels[i]:SetVisible(true)
	end
	for i = #self._items + 1, #self._labels, 1 do
		self._labels[i]:SetVisible(false)
		self._labels[i]:SetText("")
		self._labels[i].Value = nil
	end
end

local function __addItemLB(self, item, nolayout)
	table.insert(self._items, item)	
	
	__ensureLabelsLB(self)
	
	local label = self._labels[#self._items]
	label.Value = item
	label.Index = #self._items
	label:SetText(tostring(item))	
	label:SetHeight(label:GetFullHeight())
	
	if (not (nolayout == true)) then
		self:Layout()
	end
	
	return #self._items -- new item index
end

local function __getItemLB(self, index)
	return self._items[index]
end

local function __getItemCountLB(self)
	return #self._items
end

local function __removeItemLB(self, index, nolayout)	
	if (#self._items > 0) then
		local r = table.remove(self._items, index)
		
		__ensureLabelsLB(self)

		if (not (nolayout == true)) then
			self:Layout()
		end
		
		if (self.SelectedIndexChanged) then 
			self:SelectedIndexChanged(self:GetSelectedIndex()) 
		end
		return r
	end
end

local function __clearItemsLB(self, nolayout)
	self._items = {}
	
	for _, label in ipairs(self._labels) do
		label:SetText("")
		label:SetVisible(false)
		label.isSelected = false
	end
	
	if (not (nolayout == true)) then
		self:Layout()
	end
	
	if (self.SelectedIndexChanged) then 
		self:SelectedIndexChanged(-1) 
	end
end

local function __itemPairsLB(self)
	return ipairs(self._items)
end





local listbox = {
	new = function(self, parent)
		local this = bbox:new(parent)
		local mask = myUI:CreateFrame("Mask", this)
		local slider = myUI:CreateFrame("Frame", this)
		local sbgtop = myUI:CreateFrame("Texture", slider) -- 16x10
		local sbgbottom = myUI:CreateFrame("Texture", slider) -- 16x10
		local sbgmid = myUI:CreateFrame("Texture", slider) -- 16x4
		
		sbgtop:SetTexture("myUI", "images/vscroll-top-bg.png")
		sbgbottom:SetTexture("myUI", "images/vscroll-bottom-bg.png")
		sbgmid:SetTexture("myUI", "images/vscroll-mid-bg.png")
		
		sbgtop:SetLayer(2)
		sbgbottom:SetLayer(1)
		sbgmid:SetLayer(2)
		
		sbgtop:SetSize(16, 10)
		sbgbottom:SetSize(16,10)
		sbgmid:SetWidth(16)
		
		sbgtop:SetPoint("TOPRIGHT", slider, "TOPRIGHT")
		sbgbottom:SetPoint("BOTTOMRIGHT", slider, "BOTTOMRIGHT")
		sbgmid:SetPoint("TOPLEFT", sbgtop, "BOTTOMLEFT", 0, -1)
		sbgmid:SetPoint("BOTTOMRIGHT", sbgbottom, "TOPRIGHT", 0, 1)
		
		local sthumb = myUI:CreateFrame("Frame", slider)
		sthumb:SetWidth(16)
		sthumb:SetHeight(32)
		sthumb:SetLayer(5)
		
		local sfgtop = myUI:CreateFrame("Texture", sthumb) -- 16x10
		local sfgbottom = myUI:CreateFrame("Texture", sthumb) -- 16x10
		local sfgmid = myUI:CreateFrame("Texture", sthumb) -- 15,x4
		
		sfgtop:SetLayer(7)
		sfgbottom:SetLayer(7)
		sfgmid:SetLayer(6)
		
		sfgtop:SetTexture("myUI", "images/vscroll-top-fg.png")
		sfgbottom:SetTexture("myUI", "images/vscroll-bottom-fg.png")
		sfgmid:SetTexture("myUI", "images/vscroll-mid-fg.png")
		
		sfgtop:SetSize(16,10)
		sfgbottom:SetSize(16,10)
		sfgmid:SetWidth(16)
		
		sfgtop:SetPoint("TOPLEFT", sthumb, "TOPLEFT")
		sfgbottom:SetPoint("BOTTOMRIGHT", sthumb, "BOTTOMRIGHT")
		sfgmid:SetPoint("TOPLEFT", sfgtop, "BOTTOMLEFT", 0, -1)
		sfgmid:SetPoint("BOTTOMRIGHT", sfgbottom, "TOPRIGHT", 0, 1)
		
		sthumb:SetPoint("TOPCENTER", slider, "TOPCENTER")
		
		mask:SetPoint("TOPLEFT", this, "TOPLEFT", 8, 8)
		mask:SetPoint("BOTTOMRIGHT", this, "BOTTOMRIGHT", -24, -8)
		
		slider:SetPoint("TOPLEFT", mask, "TOPRIGHT", 4, 0)
		slider:SetPoint("BOTTOMLEFT", mask, "BOTTOMRIGHT", 4, 0)
		slider:SetWidth(16)
		
		--mask:SetBackgroundColor(1,1,1,0.1)
		
		this._items = {}
		this._labels = {}		
		this._topoffset = 0
		
		this.Slider = slider
		this.SliderThumb = sthumb
		this.Viewport = mask
		
		this.Layout = __layoutLB
		this.ItemPairs = __itemPairsLB
		this.AddItem = __addItemLB
		this.RemoveItem = __removeItemLB
		this.ClearItems = __clearItemsLB
		this.GetItemCount = __getItemCountLB
		this.GetItem = __getItemLB
		this.GetItemSelected = __getItemSelectedLB
		this.SetItemSelected = __setItemSelectedLB
		this.GetSelectedIndex = __getSelectedIndexLB
		this.GetSelectedItem = __getSelectedItemLB
		
		this.Event.Size = function(self)
			self:Layout()
		end
		
		sthumb.Event.LeftDown = function(self)
			self._dragging = true
			
			local m = Inspect.Mouse()			
			self._dragy = m.y
		
		end
		
		sthumb.Event.MouseMove = function(self, x, y)
			if (self._dragging) then
				local slider = self:GetParent()
				local miny = 0
				local maxy = slider:GetHeight() - self:GetHeight()
					
				local deltay = y - self._dragy
				
				local yoffset = self:GetTop() - slider:GetTop()
				local newy = minmax(yoffset + deltay, 0, maxy)
				
				self:SetPoint("TOPCENTER", slider, "TOPCENTER", 0, newy)
				self._dragy = y
				
				local lbox = slider:GetParent()
				local vp = lbox.Viewport
				
				if (#lbox._items > 0) then
					local offratio = (self:GetTop() - slider:GetTop()) / slider:GetHeight()
					
					yoffset = offratio * __getVirtualHeightLB(lbox)
					
					lbox._labels[1]:SetPoint("TOPLEFT", vp, "TOPLEFT", 0, -yoffset)
					lbox._labels[1]:SetPoint("TOPRIGHT", vp, "TOPRIGHT", 0, -yoffset)
				
				end
				
			end
		end
		
		sthumb.Event.LeftUp = function(self)
			self._dragging = false
		end
		
		sthumb.Event.LeftUpoutside = function(self)
			self._dragging = false
		end
		
		return this
	end
}

local ttip
local ttipid = 0

function myUI:ShowTooltip(caller, heading, content, x, y)
	if (not caller) then return end
	
	if (not ttip) then
		ttip = bbox:new()
		ttip:SetBackgroundColor(0, 0, 0, 0.8)
		ttip:SetLayer(1100)
		ttip.header = myUI:CreateFrame("Text", ttip)
		ttip.header:SetFontSize(ttip.header:GetFontSize() * 1.2)
		ttip.header:SetFontColor(myUI.colors.LightSkyBlue:rgba())
		ttip.header:Autosize("WWW")
		ttip.body = myUI:CreateFrame("Text", ttip)
		ttip.body:SetWordwrap(true)
		
		ttip:SetPoint("TOPLEFT", ttip.header, "TOPLEFT", -8, -6)
		ttip:SetPoint("TOPRIGHT", ttip.header, "TOPRIGHT", 8, -6)
		ttip:SetPoint("BOTTOMLEFT", ttip.body, "BOTTOMLEFT", -8, 8)
		ttip:SetPoint("BOTTOMRIGHT", ttip.body, "BOTTOMRIGHT", 8, 8)
		
		ttip.body:SetPoint("TOPLEFT", ttip.header, "BOTTOMLEFT", 0, 6)
		ttip.body:SetPoint("TOPRIGHT", ttip.header, "BOTTOMRIGHT", 0, 6)
		
		ttip:SetVisible(false)
	end
	local atmouse = false
	if (x == nil) or (y == nil) then
		atmouse = true
		local m = Inspect.Mouse()
		if (x == nil) then x = m.x end
		if (y == nil) then y = m.y end	
	end

	ttipid = ttipid + 1
	caller.__ttipid = ttipid
	
	if (isnil(content)) then
		ttip.header:SetText("")
		ttip.body:SetText(heading)
		ttip.header:SetWidth(minmax(ttip.body:GetFullWidth(), 80, 280))
		ttip.body:SetHeight(ttip.body:GetFullHeight())
		ttip.header:SetHeight(0)
	else
		ttip.header:SetText(ifnil(heading, ""))
		ttip.header:SetHeight(ttip.header:GetFullHeight())
		ttip.header:SetWidth(minmax(ttip.header:GetFullWidth(), 200, 280))		
		ttip.body:SetText(ifnil(content, ""))
		ttip.body:SetHeight(ttip.body:GetFullHeight())
	end
	-- todo: more complicated positioning logic to assure we don't go off the screen
	
	local cy = UIParent:GetHeight()
	ttip.header:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", x + iif(atmouse, 40, 0), -cy + y - iif(atmouse, 40, 0) - ttip.body:GetHeight())
	
	ttip:SetVisible(true)
	
end

function myUI:HideTooltip(caller)
	if (not caller) or (not caller.__ttipid) then return end
	
	if (caller.__ttipid >= ttipid) then
		caller.__ttipid = nil
		ttip:SetVisible(false)
	end
end





me.Tabstrip = tabstrip
me.AnchorBox = anchorbox
me.BorderBox = bbox
me.Dragbox = dragbox
me.Listbox = listbox
myUI.Widgets = me